/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>

#ifdef __arm__
#include <X11/extensions/Xrandr.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include "alchemy.h"
#include "utils.h"
#include "basic_window.h"
#include "normal_window.h"
#include "dialog_window.h"
#include "toolbar_window.h"
#include "splash_window.h"
#include "winlist_window.h"
#include "ichoose_window.h"
#include "star_window.h"
#include "keyboard_glue.h"
#include "clock_glue.h"
#include "ac.h"
#include "acctrl.h"
#include "grab.h"
#include "ewmh.h"
#include <eventloop.h>
#include <display.h>

#ifdef __arm__
// For rotation handling
static Bool rotation_initialized = False;
static Rotation rotation_table[4];
#endif

// Adopt all these poor windows who had nobody to manage them
// before we came
void ac_adopt_existing(wm *wm)
{
	Window dum_r;
	Window *winlist;
	int i, n;
	XWindowAttributes wattr;
	
	XQueryTree(wm->display, wm->root, &dum_r, &dum_r, &winlist, &n);
	for(i=0;i<n;i++) {
		XGetWindowAttributes(wm->display, winlist[i], &wattr);
		if(!wattr.override_redirect && wattr.map_state == IsViewable) {
			window *w=ac_create_new_window (wm, winlist[i]);
			if(w)
				w->ignore_next_unmap=True;
		}
	}
	if(n)
		XFree(winlist);
}

// Check whether a printer is connected and return true or false
static Bool ac_printer_status()
{
	int fd;
	Bool ret=False;

	if((fd=open("/tmp/printerconnected",O_RDONLY)) >0) {
		char s;

		if((read(fd,&s,1)==1) && s=='1')
			ret = True;
		close(fd);
	}
	return ret;
}

// A toolbar (w) appeared or disappeared
// This affects other toolbars and normal windows
// Resize/Reposition them
void ac_retile(wm *wm, window *w, Bool got_mapped)
{
	window *win;
	int multiplier=0;

	if(got_mapped)
		multiplier=-1;
	else
		multiplier=1;

	For_All_Windows_Do(wm, win);
		if(win->wtype == NORMAL_WIN) {
			win->set_size(win, win->width,win->height +(multiplier * w->height) );
		} else if(win->wtype == TOOLBAR_WIN && win != w) {
			if(window_get_state(win) == NormalState)
				if(win->y <= w->y)
					win->set_position(win, win->x,win->y + (multiplier * w->height) );
		} else if(win->wtype == MENU_WIN) {
			// Move the unlucky ichoose menu window
			// if it is around
			win->y += multiplier * w->height;
			if(win->appid == AC_KEYBOARD_MENU_APPID) {
				XMoveWindow(wm->display, win->app_window, win->x, win->y);
			}
		}
	For_All_Windows_Done(wm, win);
}

// Tell me how much space 
// the toolbars have eaten away
unsigned int ac_get_toolbar_displacement(wm *wm)
{
	unsigned int total_displacement=0;
	window *w;

	For_All_Mappable_Do(wm, w);
		if(w->wtype == TOOLBAR_WIN)
			if(window_get_state(w) == NormalState)
				total_displacement += w->height;
	For_All_Mappable_Done(wm, w);
	return total_displacement;
}

// Create a window structure around a XWindow, and adopt it
static window *ac_create_new_window (wm *wm, Window map_window)
{
	window *w;
	Atom win_type;
	Window prop_window;

	XGrabServer(wm->display);

	xerror_variable = 0;
	win_type = xwindow_get_atom_property(map_window, wm, _NET_WM_WINDOW_TYPE);
	if(xerror_variable) {
		// If window has already disappeared
		XUngrabServer(wm->display);
		return NULL;
	}
	
	if(XGetTransientForHint(wm->display, map_window, &prop_window)) {
		DBG(debug_info("Creating new dialog window map_window=%x", map_window);)
		w = new_dialog_window (wm, map_window);
	} else if(win_type == wm->atomlist[_NET_WM_WINDOW_TYPE_TOOLBAR]) {
		DBG(debug_info("Creating new toolbar window map_window=%x", map_window);)
		w = new_toolbar_window (wm, map_window);
	} else if(win_type ==  wm->atomlist[_NET_WM_WINDOW_TYPE_DIALOG]) {
		DBG(debug_info("Creating new dialog window map_window=%x", map_window);)
		w = new_dialog_window (wm, map_window);
	} else if(win_type ==  wm->atomlist[_NET_WM_WINDOW_TYPE_SPLASH]) {
		DBG(debug_info("Creating new splash window map_window=%x", map_window);)
		w = new_splash_window (wm, map_window);
	} else {
		DBG(debug_info("Creating new normal window map_window=%x", map_window);)
		w = new_normal_window (wm, map_window);
	}
	
	w->show(w);

	XSync(wm->display, False);
	XUngrabServer(wm->display);

	return w;
}

// Somebody wants to be mapped, if already managed, just show it
// else first adopt it
static void ac_handle_map_request(wm *wm, XMapEvent *xev, window *w)
{
	if(w) {
		window *m = wm->mapped_window;

		if(m && m!=w) {
			w->show(w);

			// Do not hide if only a dialog wants to be mapped
			if(w->wtype == NORMAL_WIN)
				m->hide(m);
		} else
			w->show(w);

		return;
	}

	w = ac_create_new_window (wm, xev->window);

	// TODO : ewmh update
}

// Unless it has to be ignored, destroy the window
static void ac_handle_unmap_notify(wm *wm, XUnmapEvent *xev, window *w)
{
	if(!w)
		return;

	DBG(debug_info("%s(), wtype=%d, name=%s, title=%s, win=%x",__func__, w->wtype, w->name, w->title, w->app_window);)
	if(w->ignore_next_unmap) {
		w->ignore_next_unmap = False;
	} else {
		XEvent ev;

		XGrabServer(wm->display);
		if (XCheckTypedWindowEvent(w->wm->display, w->frame_window, DestroyNotify, &ev)) {
			w->destroy(w);
		} else {
			window_set_state(w, WithdrawnState);
			XReparentWindow(w->wm->display, w->app_window, w->wm->root,
					w->x, w->y); 
			w->destroy(w);
		}
		XUngrabServer(wm->display);
	}
}

// Someone wants to be destroyed, Very well, de-allocate his structures
static void	ac_handle_destroy_notify(wm *wm, XEvent *xev, window *w)
{
	if(w) {
		XGrabServer(wm->display);
		DBG(debug_info("%s(), wtype=%d, name=%s, title=%s, win=%x",__func__, w->wtype, w->name, w->title, w->app_window);)
		w->destroy(w);
		XUngrabServer(wm->display);
	}
}

// Handle button press event
static void ac_handle_button_press(wm *wm, XEvent *xev, window *w, event_area where)
{
	XEvent ev;
	window *tw;
	window *to_destroy;

	XAllowEvents(wm->display, ReplayPointer, CurrentTime);

	if(w && w->wtype != MENU_WIN) {

		if (XCheckTypedWindowEvent(wm->display, w->frame_window, DestroyNotify, &ev)) {
			DBG(debug_info("!!!!!Putting back frame_window event!!!!!");)
			XPutBackEvent(wm->display, &ev);
			return;
		}

		window_handle_focus(w);
	}

	if(wm->is_menu_mapped) {
		if(w) {
			px_event pxe;
			px_eventloop* eventloop = eventloop_get_eventloop();

			pxe.xev = xev;
			pxe.private_data = w;
			pxe.type = NR_CALLBACKS;


			if(w && w->wtype == MENU_WIN) {
				eventloop->handle_event(eventloop, &pxe, w->topbar_frame);
			}

			to_destroy = NULL;
			For_All_Mappable_Do(wm, tw);
				if(tw->wtype == MENU_WIN && tw != w) {
					to_destroy = tw;
					break;
				}
			For_All_Mappable_Done(wm, tw);

			XGrabServer(wm->display);
			if(to_destroy)
				to_destroy->destroy(to_destroy);
			XUngrabServer(wm->display);
		} else {
			to_destroy = NULL;
			For_All_Mappable_Do(wm, tw);
				if(tw->wtype == MENU_WIN) {
					to_destroy = tw;
					break;
				}
			For_All_Mappable_Done(wm, tw);
			XGrabServer(wm->display);
			if(to_destroy)
				to_destroy->destroy(to_destroy);
			XUngrabServer(wm->display);
		}
		return;
	}

	ac_handle_widget_event(wm, xev, w, where);
}

//static void ac_handle_key_press(wm *wm, XKeyEvent *xev)
static void ac_handle_key_press(wm *wm, XEvent *xev)
{
	char kbuf[TMP_BUF_SIZE];
	KeySym ksym;
	window *win = wm->mapped_window;
	XKeyEvent *xkey=&xev->xkey;
	XEvent my_xev=*xev;
	
	if(win) {
		my_xev.xkey.window = win->app_window;
		my_xev.xkey.subwindow = 0;

		XLookupString(xkey, kbuf, sizeof(kbuf), &ksym, NULL);
		window_handle_focus(win);
		switch (ksym) {
			case XK_F21:
				DBG(debug_info("XK_F21 pressed");)
				ac_keypress_list[0] = win;
				if(xkey->state & ShiftMask) {
					// Close button
					if((win != wm->launcher_window) &&
						(win->appid != AC_LOCK_APPID)) {
						if(win->flags.bits.close) {
							if(win->fast_load) {
								// Do not delete, just hide
								if(win->wm->launcher_window) {
									window *launch_top=win->wm->launcher_window;
										
									/* Navigate through the layers till the top */
									while(launch_top->layer_above)
										launch_top=launch_top->layer_above;
									launch_top->show(launch_top);

									win->hide(win);
									win->fast_visible=False;
								}
							} else
								window_handle_deletion(win);
						} else if(win->flags.bits.tick)
							window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_TICK]);
						else
							window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_UP]);
					} else if(win && win==wm->launcher_window) {
						window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_CLOSE]);
					}
				} else if(xkey->state & Mod1Mask) {
					// Accel right flick
					XSendEvent(wm->display, win->app_window,
						False, KeyPressMask, &my_xev);
					window_send_protocol_event(win, wm->atomlist[_AC_ACCEL_RIGHT_FLICK]);
				}

				break;
			case XK_F22:
				DBG(debug_info("XK_F22 pressed");)
				ac_keypress_list[1] = win;
				XSendEvent(wm->display, win->app_window, False,
						//KeyPressMask, xev);
						KeyPressMask, &my_xev);
				if(xkey->state & ShiftMask) {
					// Rocker up
					window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_ROCKER_UP]);
				} else if(xkey->state & Mod1Mask) {
					// Accel left
					window_send_protocol_event(win, wm->atomlist[_AC_ACCEL_LEFT_FLICK]);
				}
				break;
			case XK_F23:
				DBG(debug_info("XK_F23 pressed");)
				ac_keypress_list[2] = win;
				XSendEvent(wm->display, win->app_window, False,
						KeyPressMask, &my_xev);
				if(xkey->state & ShiftMask)
					// Rocker down
					window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_ROCKER_DOWN]);
				else if(xkey->state & Mod1Mask)
					// Accel top
					window_send_protocol_event(win, wm->atomlist[_AC_ACCEL_TOP_FLICK]);
				break;
			case XK_F24:
				DBG(debug_info("XK_F24 pressed");)
				ac_keypress_list[3] = win;
				XSendEvent(wm->display, win->app_window, False,
						KeyPressMask, &my_xev);
				if(xkey->state & ShiftMask)
					// Rocker select
					window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_ROCKER_SELECT]);
				else if(xkey->state & Mod1Mask)
					// Accel bottom
					window_send_protocol_event(win, wm->atomlist[_AC_ACCEL_BOTTOM_FLICK]);
				break;
			case XK_F25:
				DBG(debug_info("XK_F25 pressed");)
				ac_keypress_list[4] = win;
				if(xkey->state & ShiftMask) {
					// Voice button
					if(win->flags.bits.voice || win->appid == AC_LAUNCHER_APPID)
						window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_VOICE]);
				} else if(xkey->state & Mod1Mask) {
					// Accel right extreme
					XSendEvent(wm->display, win->app_window,
						False, KeyPressMask, &my_xev);
					window_send_protocol_event(win, wm->atomlist[_AC_ACCEL_RIGHT_EXTREME]);
				}
				break;
			case XK_F26:
				DBG(debug_info("XK_F26 pressed");)
				ac_keypress_list[5] = win;
				if(win->appid != AC_LOCK_APPID && (win->appid!=AC_LAUNCHER_APPID || win->layer_below)) {
					if(xkey->state & ShiftMask) {
						// Force kill (Accel shift + close)
						if(win->pid > 0)
							kill(win->pid, SIGKILL);

						XKillClient(wm->display, win->app_window);
					} else if(xkey->state & Mod1Mask) {
						// Accel left extreme
						XSendEvent(wm->display, win->app_window, False,
							KeyPressMask, &my_xev);
						window_send_protocol_event(win, wm->atomlist[_AC_ACCEL_LEFT_EXTREME]);
					}
				}
				break;
			case XK_F27:
				DBG(debug_info("XK_F27 pressed");)
				ac_keypress_list[6] = win;
				XSendEvent(wm->display, win->app_window, False,
						KeyPressMask, &my_xev);
				window_send_protocol_event(win, wm->atomlist[_AC_ACCEL_TOP_EXTREME]);
				break;
			case XK_F28:
				DBG(debug_info("XK_F28 pressed");)
				ac_keypress_list[7] = win;
				XSendEvent(wm->display, win->app_window, False,
						KeyPressMask, &my_xev);
				window_send_protocol_event(win, wm->atomlist[_AC_ACCEL_BOTTOM_EXTREME]);
				break;
			default:
				debug_info("Unhandled key %s", kbuf);
				break;
		}
	}
	XFlush(wm->display);
}

//static void ac_handle_key_release(wm *wm, XKeyEvent *xev)
static void ac_handle_key_release(wm *wm, XEvent *xev)
{
	char kbuf[TMP_BUF_SIZE];
	KeySym ksym;
	window *win;
	XKeyEvent *xkey=&xev->xkey;
	XEvent my_xev=*xev;


	XLookupString(xkey, kbuf, sizeof(kbuf), &ksym, NULL);
	switch (ksym) {
		case XK_F21:
			DBG(debug_info("XK_F21 released");)
			win = ac_keypress_list[0];
			if(win && window_exists(wm, win)) {
				if (xkey->state & ShiftMask) {
					if(win != wm->launcher_window) {
						if(win->flags.bits.up)
							window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_UP_REL]);
						else if(win->flags.bits.tick)
							window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_TICK_REL]);
					} else
						window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_CLOSE_REL]);
				} else if (xkey->state & Mod1Mask){
					my_xev.xkey.window = win->app_window;
					XSendEvent(wm->display, win->app_window,
						False, KeyReleaseMask, &my_xev);
				}
			}
			ac_keypress_list[0]=(window *)NULL;
			break;

		case XK_F22:
			DBG(debug_info("XK_F22 released");)
			win = ac_keypress_list[1];
			if(win && window_exists(wm, win)) {
				my_xev.xkey.window = win->app_window;
				XSendEvent(wm->display, win->app_window,
						False, NoEventMask, &my_xev);
				if(xkey->state & ShiftMask)
					// Rocker up
					window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_ROCKER_UP_REL]);
			}
			ac_keypress_list[1]=(window *)NULL;
			break;
		case XK_F23:
			DBG(debug_info("XK_F23 released");)
			win = ac_keypress_list[2];
			if(win && window_exists(wm, win)) {
				my_xev.xkey.window = win->app_window;
				XSendEvent(wm->display, win->app_window, False,
					KeyReleaseMask, &my_xev);
				if(xkey->state & ShiftMask)
					// Rocker down
					window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_ROCKER_DOWN_REL]);
			}
			ac_keypress_list[2]=(window *)NULL;
			break;
		case XK_F24:
			DBG(debug_info("XK_F24 released");)
			win = ac_keypress_list[3];
			if(win && window_exists(wm, win)) {
				my_xev.xkey.window = win->app_window;
				XSendEvent(wm->display, win->app_window, False,
					KeyReleaseMask, &my_xev);
				if(xkey->state & ShiftMask)
					// Rocker select
					window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_ROCKER_SELECT_REL]);
			}
			ac_keypress_list[3]=(window *)NULL;
			break;
		case XK_F25:
			DBG(debug_info("XK_F25 released");)
			win = ac_keypress_list[4];
			if(win && window_exists(wm, win)) {
				if(xkey->state & ShiftMask) {
					if(win->flags.bits.voice || win->appid == AC_LAUNCHER_APPID)
						window_send_protocol_event(win, wm->atomlist[_AC_BUTTON_VOICE_REL]);
					else if(win->flags.bits.needs_annotation && !win->flags.bits.full_screen)
						// Start Annotate
						window_launch_annotate(win);
				} else if(xkey->state & Mod1Mask) {
					my_xev.xkey.window = win->app_window;
					// Accel right extreme
					XSendEvent(wm->display, win->app_window,
						False, KeyReleaseMask, &my_xev);
				}
			}
			ac_keypress_list[4]=(window *)NULL;
			break;
		case XK_F26:
			DBG(debug_info("XK_F26 released");)
			// Accel left extreme
			win = ac_keypress_list[5];
			if(win && window_exists(wm, win)) {
				my_xev.xkey.window = win->app_window;
				XSendEvent(wm->display, win->app_window, False,
					KeyReleaseMask, &my_xev);
			}
			ac_keypress_list[5]=(window *)NULL;
			break;
		case XK_F27:
			DBG(debug_info("XK_F27 released");)
			// Accel top extreme
			win = ac_keypress_list[6];
			if(win && window_exists(wm, win)) {
				my_xev.xkey.window = win->app_window;
				XSendEvent(wm->display, win->app_window, False,
					KeyReleaseMask, &my_xev);
			}
			ac_keypress_list[6]=(window *)NULL;
			break;
		case XK_F28:
			DBG(debug_info("XK_F28 released");)
			// Accel bottom extreme
			win = ac_keypress_list[7];
			if(win && window_exists(wm, win)) {
				my_xev.xkey.window = win->app_window;
				XSendEvent(wm->display, win->app_window, False,
					KeyReleaseMask, &my_xev);
			}
			ac_keypress_list[7]=(window *)NULL;
			break;
		default:
			DBG(debug_info("Unhandled key %s", kbuf);)
			break;
	}
	XFlush(wm->display);
}

// Grab all the keys that need to be handled by the window manager
static void ac_init_keys(wm *wm)
{
	Display *dpy = wm->display;
	int i;

	for(i=0;i<MAX_KEYS;i++)
		ac_keypress_list[i]=(window *)NULL;

	// Close/Up button (ShiftMask) and Accel right flick (Mod1Mask)
	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_F21), AnyModifier , wm->root, True, GrabModeAsync, GrabModeAsync);
	// Rocker up (ShiftMask) and Accel left flick (Mod1Mask)
	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_F22), AnyModifier, wm->root, True, GrabModeAsync, GrabModeAsync);
	// Rocker down (ShiftMask) and Accel top flick (Mod1Mask)
	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_F23), AnyModifier , wm->root, True, GrabModeAsync, GrabModeAsync);
	// Rocker select (ShiftMask) and Accel bottom flick (Mod1Mask)
	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_F24), AnyModifier , wm->root, True, GrabModeAsync, GrabModeAsync);
	// Annotate/Voice key (ShiftMask) and Accel right extreme (Mod1Mask)
	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_F25), AnyModifier, wm->root, True, GrabModeAsync, GrabModeAsync);
	// Application kill (ShiftMask) and Accel left extreme (Mod1Mask)
	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_F26), AnyModifier, wm->root, True, GrabModeAsync, GrabModeAsync);
	// Accel up extreme
	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_F27), Mod1Mask , wm->root, True, GrabModeAsync, GrabModeAsync);
	// Accel bottom extreme
	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_F28), Mod1Mask , wm->root, True, GrabModeAsync, GrabModeAsync);
}

// configure a window .. not sure who uses this
// Now I know :)
static void ac_handle_configure_request(wm *wm, XConfigureRequestEvent *xev, window *w, event_area where)
{
	XWindowChanges xwc;

	if(w) {
		if(xev->above != None && xev->detail == Above) {
			w->show(w);
		}

		if(xev->above !=None && xev->detail == Below) {
			window *wn;

			if(wm->launcher_window)
				wn=wm->launcher_window;
			else
				wn=window_get_next_mappable(w);
			if(wn) {
				wn->show(wn);
				w->hide(w);
			}
		}
	} 
	
	// Vivek : another highly experimental change..
	/*
	else {
	*/
		xwc.x = xev->x;
		xwc.y = xev->y;
		if(w && w->wtype == NORMAL_WIN && !w->flags.bits.full_screen) {
			xwc.width = w->width;
			xwc.height = w->height;
		} else {
			xwc.width = xev->width;
			xwc.height = xev->height;
		}
		xwc.sibling = xev->above;
		xwc.stack_mode = xev->detail;
		XConfigureWindow(wm->display, xev->window, xev->value_mask, &xwc);
		/*
		return;
	}
	*/
}

// If rotation happenned resize all windows
static void ac_handle_configure_notify(wm *wm, XConfigureEvent *xev)
{
	if(xev->window == wm->root) { // This happens during rotation

		if((xev->height != wm->display_height) || (xev->width != wm->display_width)) {
			window *w;
			int height_change, width_change;
			
#ifdef __arm__
			XRRScreenConfiguration *sc;
			Rotation current_rotation;
#endif

			height_change = xev->height - wm->display_height;
			width_change = xev->width - wm->display_width;

			wm->display_height = xev->height;
			wm->display_width = xev->width;

			XGrabServer(wm->display);
#ifdef __arm__
			sc = XRRGetScreenInfo(wm->display, wm->root);
			XRRConfigRotations(sc, &current_rotation);
			if(current_rotation & rotation_table[ROT_0]){
				wm->current_rot = ROT_0;
			} else if(current_rotation & rotation_table[ROT_90]){
				wm->current_rot = ROT_90;
			} else if(current_rotation & rotation_table[ROT_180]){
				wm->current_rot = ROT_180;
			} else if(current_rotation & rotation_table[ROT_270]){
				wm->current_rot = ROT_270;
			}

			if((w=wm->mapped_window)) {
				w->rotation = wm->current_rot;
				while(w->layer_below) {
					w = w->layer_below;
					w->rotation = wm->current_rot;
				}
			}
#endif
			For_All_Windows_Do(wm, w);
				if(w->wtype == NORMAL_WIN) {
					w->set_size(w, w->width + width_change, 
							w->height + height_change);
					// Draw the top and bottom bars
					w->redraw(w);
					window_send_configure_event(w);
				}
			For_All_Windows_Done(wm, w);
			XUngrabServer(wm->display);
		}
	}
}

// Handle any message sent from the client window
static void ac_handle_client_message(wm *wm, XClientMessageEvent *xev)
{
	window *clientwin=NULL;
	
	if(xev->data.l[1] == wm->root)
		clientwin = wm->mapped_window;
	else
		clientwin = window_find_windowid(wm, xev->data.l[1]);
	DBG(debug_info("clientwin=%x\n",clientwin);)

	if(xev->message_type == wm->atomlist[_AC_WM_COMMAND]) {

		if(xev->data.l[0] != AC_CMD_LAUNCH_APP && !clientwin)
			return;

		DBG(debug_info("got %d eventn", xev->data.l[0]);)
		
		switch(xev->data.l[0]) {

			case AC_CMD_EXIT:
				wm->exit(wm);
				break;

			case AC_CMD_SET_THEME: {
				char data[MAX_PROPERTY_LENGTH];
				
				xwindow_get_text_property(wm, wm->root, _AC_WM_THEME, data);

				if(data[0] == 0) {
					debug_info("Unable to extract the new theme information\n");
				} else {
					ac_theme *cur_theme=wm->current_theme;
					window *win;
					FILE *fp;

					// If new theme is different
					// load and re-render all windows
					if(strcmp(data,cur_theme->theme_path)) {
						For_All_Windows_Do(wm, win);
							cur_theme->cleanup_theme(cur_theme, win);
						For_All_Windows_Done(wm, win);

						cur_theme->free_theme(cur_theme);
						cur_theme->load_theme(cur_theme, data, wm->root);
						For_All_Windows_Do(wm, win);
							win->redraw(win);
							win->refresh_widgets(win);
						For_All_Windows_Done(wm, win);

						// Restart clock with the new
						// theme
						if(wm->clock_window) {
							window_handle_deletion(wm->clock_window);
							clock_start_clock(data);
						}

						// Save the theme to be used
						// from next restarts
						fp=fopen(THEME_CONFIG_FILE,"w");
						if(fp) {
							fprintf(fp,"%s\n", data);
							fclose(fp);
						}
					}
				}
			}
			break;

			case AC_CMD_LAUNCH_APP: {
				char data[MAX_PROPERTY_LENGTH];
				XEvent xcm;
				
				xwindow_get_text_property(wm, wm->root, _AC_WM_LAUNCH_CMD, data);
				if(xev->window != wm->root) {
					memset(&xcm,0,sizeof(xcm));
					xcm.xclient.type = ClientMessage;
					xcm.xclient.window = wm->root;
					xcm.xclient.message_type = wm->atomlist[_AC_WM_COMMAND];
					xcm.xclient.format = 8;
					xcm.xclient.data.l[0] = AC_CMD_LAUNCH_APP;
				
					XSendEvent(wm->display, (Window) xev->data.l[1], False, NoEventMask, &xcm);
					XFlush(wm->display);
				}

				if(data[0] == 0) {
					debug_info("Unable to extract the commandline");
				} else {
					char *cmd, *appid_str, *appname;
					unsigned long appid;
					window *win;

					char env1[TMP_BUF_SIZE];
					char env2[TMP_BUF_SIZE];
					char env3[TMP_BUF_SIZE];

					appid_str = strtok(data, ",");
					if(!appid_str)
						break;
					appid = strtoul(appid_str, NULL, 10);

					appname = strtok(NULL, ",");
					if(!appname)
						break;

					cmd = strtok(NULL, ",");
					if(!cmd)
						break;

					// First search if window with the specified
					// appid is already running
					win = window_find_appid(wm, appid);

					if(win) {
						// If so, show the window
						window *old_mapped_win;

						old_mapped_win = wm->mapped_window;
						// Set visibility to on 
						// for fast load apps now
						if(win->fast_load)
							win->fast_visible = True;

						win->show(win);
						if(old_mapped_win && old_mapped_win != wm->mapped_window)
							old_mapped_win->hide(old_mapped_win);

					} else {
						// launch the application
						sprintf(env1, "_AC_W_APPID=%lu",appid);
						sprintf(env2, "_NET_WM_NAME=%s",appname);
						if(clientwin)
							sprintf(env3, "_AC_W_LAUNCHER=%lu",clientwin->appid);
						else
							sprintf(env3, "_AC_W_LAUNCHER=1");

						spawn_child(cmd, env1, env2, env3, NULL);
					}
				}
			}
			break;

			case AC_CMD_SET_FLAGS:
				clientwin->set_flags(clientwin, window_get_int_property(clientwin, _AC_W_FLAGS));
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SET_EMAIL_FILE: 
				{
					char data[MAX_PROPERTY_LENGTH];
					
					window_get_text_property(clientwin, _AC_W_EMAIL_FILE, data);

					if(data[0] == 0) {
						debug_info("Unable to extract the email file\n");
					} else {
						char env1[TMP_BUF_SIZE],env2[TMP_BUF_SIZE],env3[TMP_BUF_SIZE];
						char cmd[2*TMP_BUF_SIZE];

						DBG(debug_info("data = %s \n", data);)

						if(!strncmp(data,"to,",3))
							sprintf(cmd,EMAIL_CMDLINE,"-to",&data[3]);
						else if(!strncmp(data,"file,",5))
							sprintf(cmd,EMAIL_CMDLINE,"-content",&data[5]);

						if(clientwin->onetap) {
							FILE *fp;

							clientwin->onetap = False;
							strcat(cmd, " -onetap");
							fp=fopen("/tmp/onetap_bg.png","wb");
							save_image(wm->display, clientwin->app_window, fp,
									0,
									(clientwin->height/2)-ONE_TAP_STRIP_HEIGHT,
									0,
									(clientwin->height/2)-ONE_TAP_STRIP_HEIGHT);
							fclose(fp);
						}

						sprintf(env1,"_NET_WM_NAME=%s",clientwin->name);
						sprintf(env2,"_AC_W_APPID=%lu",clientwin->appid);
						sprintf(env3,"_AC_W_LAYER_WINDOWID=%lu",clientwin->app_window);
						spawn_child(cmd, env1, env2, env3, NULL);
					}
				}
				break;

			case AC_CMD_SET_STATUS_MSG:
				AC_HANDLE_VAR_CHANGE_CMD(_AC_W_STATUS_MSG, status_msg, 1);
				break;

			case AC_CMD_SET_STATUS_FONT:
				AC_HANDLE_VAR_CHANGE_CMD(_AC_W_STATUS_FONT, sfont, 1);
				break;

			case AC_CMD_SET_TITLE_FONT:
				AC_HANDLE_VAR_CHANGE_CMD(_AC_W_TITLE_FONT, tfont, 1);
				break;


			case AC_CMD_SET_STATUS_JUSTIFY: 
				{
					char data[MAX_PROPERTY_LENGTH];
					
					window_get_text_property(clientwin, _AC_W_STATUS_JUSTIFY, data);

					if(data[0] == 0) {
						debug_info("Unable to extract status justify\n");
					} else {
						DBG(debug_info("data = %s \n", data);)
						switch(data[0]) {
							case 'l':
								if(clientwin->sjustify != JUSTIFY_LEFT) {
									clientwin->sjustify = JUSTIFY_LEFT;
									clientwin->redraw(clientwin);
								}
								break;

							case 'r':
								if(clientwin->sjustify != JUSTIFY_RIGHT) {
									clientwin->sjustify = JUSTIFY_RIGHT;
									clientwin->redraw(clientwin);
								}
								break;

							case 'c':
								if(clientwin->sjustify != JUSTIFY_CENTER) {
									clientwin->sjustify = JUSTIFY_CENTER;
									clientwin->redraw(clientwin);
								}
								break;

							default:
								debug_info("Unknown justification requested");
						}
					}
				}
				break;

			case AC_CMD_SET_PAGE_NUM:
				AC_HANDLE_VAR_CHANGE_CMD(_AC_W_PAGE_NUM, page_num, 1);
				break;

			case AC_CMD_SET_PAGE_NUM_FONT:
				AC_HANDLE_VAR_CHANGE_CMD(_AC_W_PAGE_NUM_FONT, pfont, 1);
				break;

			case AC_CMD_SET_PAGE_NUM_JUSTIFY:
				{
					char data[MAX_PROPERTY_LENGTH];
				
					window_get_text_property(clientwin, _AC_W_PAGE_NUM_JUSTIFY, data);
					if(data[0] == 0) {
						debug_info("Unable to extract status justify\n");
					} else {
						DBG(debug_info("data = %s \n", data);)

						switch(data[0]) {
							case 'l':
								if(clientwin->pjustify != JUSTIFY_LEFT) {
									clientwin->pjustify = JUSTIFY_LEFT;
									clientwin->redraw(clientwin);
								}
								break;
							case 'r':
								if(clientwin->pjustify != JUSTIFY_RIGHT) {
									clientwin->pjustify = JUSTIFY_RIGHT;
									clientwin->redraw(clientwin);
								}
								break;
							case 'c':
								if(clientwin->pjustify != JUSTIFY_CENTER) {
									clientwin->pjustify = JUSTIFY_CENTER;
									clientwin->redraw(clientwin);
								}
								break;
						}
					}
				}
				break;


			case AC_CMD_SET_KEYBOARD_LAYOUT:
				AC_HANDLE_VAR_CHANGE_CMD(_AC_W_KEYBOARD_LAYOUT, keyboard_layout, 0);
				if(wm->mapped_window)
					window_handle_keyboard(wm->mapped_window);
				break;

			case AC_CMD_SET_KEYBOARD_CATEGORY:
				AC_HANDLE_VAR_CHANGE_CMD(_AC_W_KEYBOARD_CATEGORY, keyboard_category, 0);
				if(wm->mapped_window)
					window_handle_keyboard(wm->mapped_window);
				break;

			case AC_CMD_SET_ANNOTATE_KEY:
				AC_HANDLE_VAR_CHANGE_CMD(_AC_W_ANNOTATE_KEY, annotation_key, 0);
				window_handle_annotate_icon(clientwin);
				break;

			case AC_CMD_SET_HELP_KEY:
				AC_HANDLE_VAR_CHANGE_CMD(_AC_W_HELP_KEY, help_key, 0);
				break;

#ifdef __arm__
			case AC_CMD_SET_ROTATION:
				{
					char data[MAX_PROPERTY_LENGTH];
					
					window_get_text_property(clientwin, _AC_W_ROTATION_STATE, data);

					if(data[0] == 0) {
						debug_info("Unable to extract deletion rotation state\n");
					} else {
						int i;

						debug_info("data = %s", data);
						i=data[0]-'0';
						if(i>=0 && i<=3) {
							debug_info("i=%d",i);
							clientwin->rotation = i;
							window_handle_rotation(clientwin);
						}
					}
				}
				break;
#endif


			case AC_CMD_DEL_ANNOTATION:
				/* 
				 * Since deletions are done at client,
				 * a mere refresh is done here now
				 */
				window_handle_annotate_icon(clientwin);
				break;

			case AC_CMD_SHOW_INPUT :
				clientwin->set_flag(clientwin, NEEDS_INPUT, True);
				clientwin->set_flag(clientwin, INPUT_CHOOSE, True);
				if(!strcmp(clientwin->keyboard_layout,"xstroke"))
					clientwin->set_flag(clientwin, XSTROKE, True);
				else
					clientwin->set_flag(clientwin, INPUT, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_HIDE_SPEAKER :
				clientwin->set_flag(clientwin, SPEAKER, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_SPEAKER :
				if(!clientwin->app_speaks) {
					if(access(FLITE_CMDLINE, R_OK | X_OK))
						return;
					if(access(XCLIP_CMDLINE, R_OK | X_OK))
						return;
				}
				clientwin->set_flag(clientwin, SPEAKER, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_HIDE_VOICE :
				clientwin->set_flag(clientwin, VOICE, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_VOICE :
				clientwin->set_flag(clientwin, VOICE, True);
				clientwin->set_flag(clientwin, NEEDS_ANNOTATION, False);
				clientwin->set_flag(clientwin, ANNOTATE, False);
				clientwin->set_flag(clientwin, ANNOTATION_PRESENT, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_APP_SPEAKS :
				clientwin->app_speaks = True;
				break;

			case AC_CMD_AC_SPEAKS :
				clientwin->app_speaks = False;
				break;

			case AC_CMD_HIDE_INPUT :
				clientwin->set_flag(clientwin, NEEDS_INPUT, False);
				clientwin->set_flag(clientwin, INPUT, False);
				clientwin->set_flag(clientwin, XSTROKE, False);
				clientwin->set_flag(clientwin, INPUT_CHOOSE, False);
				clientwin->set_flag(clientwin, NEEDS_KEYBOARD, False);
				clientwin->redraw(clientwin);
				window_handle_keyboard(clientwin);
				xstroke_stop_recognition(wm);	// Stop xstroke
				break;

			case AC_CMD_SHOW_ANNOTATE :
				clientwin->set_flag(clientwin, VOICE, False);
				clientwin->set_flag(clientwin, NEEDS_ANNOTATION, True);
				window_handle_annotate_icon(clientwin);
				break;

			case AC_CMD_HIDE_ANNOTATE :
				clientwin->set_flag(clientwin, NEEDS_ANNOTATION, False);
				clientwin->set_flag(clientwin, ANNOTATE, False);
				clientwin->set_flag(clientwin, ANNOTATION_PRESENT, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_EMAIL :
				clientwin->set_flag(clientwin, EMAIL, True);
				clientwin->set_flag(clientwin, ONETAP, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_HIDE_EMAIL :
				clientwin->set_flag(clientwin, EMAIL, False);
				clientwin->set_flag(clientwin, ONETAP, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_PRINTER :
				clientwin->set_flag(clientwin, NEEDS_PRINTER, True);
				if(!clientwin->flags.bits.printer && wm->is_printer_availible) {
					clientwin->set_flag(clientwin, PRINTER, True);
					clientwin->redraw(clientwin);
				}
				break;

			case AC_CMD_HIDE_PRINTER :
				clientwin->set_flag(clientwin, NEEDS_PRINTER, False);
				if(clientwin->flags.bits.printer) {
					clientwin->set_flag(clientwin, PRINTER, False);
					clientwin->redraw(clientwin);
				}
				break;

			case AC_CMD_SHOW_PAGE_NAV :
				clientwin->set_flag(clientwin, PAGE_NAV, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_HIDE_PAGE_NAV :
				clientwin->set_flag(clientwin, PAGE_NAV, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_ENABLE_PAGE_NEXT :
				clientwin->set_flag(clientwin, NEXT_PAGE_ENABLED, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_DISABLE_PAGE_NEXT :
				clientwin->set_flag(clientwin, NEXT_PAGE_ENABLED, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_ENABLE_PAGE_PREV :
				clientwin->set_flag(clientwin, PREV_PAGE_ENABLED, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_DISABLE_PAGE_PREV :
				clientwin->set_flag(clientwin, PREV_PAGE_ENABLED, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_RETRACTBAR :
				clientwin->set_flag(clientwin, RETRACTBAR, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_HIDE_RETRACTBAR :
				clientwin->set_flag(clientwin, RETRACTBAR, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_KEYBOARD :
				clientwin->set_flag(clientwin, NEEDS_KEYBOARD, True);
				window_handle_keyboard(clientwin);
				break;

			case AC_CMD_HIDE_KEYBOARD :
				clientwin->set_flag(clientwin, NEEDS_KEYBOARD, False);
				window_handle_keyboard(clientwin);
				xstroke_stop_recognition(wm);	// Stop xstroke
				break;

			case AC_CMD_SHOW_STATUS :
				clientwin->set_flag(clientwin, STATUS_MSG, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_HIDE_STATUS :
				clientwin->set_flag(clientwin, STATUS_MSG, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SWAP_CLOSE_UP:
				if(clientwin->flags.bits.close) {
					clientwin->set_flag(clientwin, UP, True);
					clientwin->set_flag(clientwin, CLOSE, False);
					clientwin->set_flag(clientwin, TICK, False);
				} else {
					clientwin->set_flag(clientwin, UP, False);
					clientwin->set_flag(clientwin, CLOSE, True);
					clientwin->set_flag(clientwin, TICK, False);
				}
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_CLOSE:
				clientwin->set_flag(clientwin, CLOSE, True);
				clientwin->set_flag(clientwin, UP, False);
				clientwin->set_flag(clientwin, TICK, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_UP:
				clientwin->set_flag(clientwin, UP, True);
				clientwin->set_flag(clientwin, CLOSE, False);
				clientwin->set_flag(clientwin, TICK, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_TICK:
				clientwin->set_flag(clientwin, UP, False);
				clientwin->set_flag(clientwin, CLOSE, False);
				clientwin->set_flag(clientwin, TICK, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_SHOW_HELP :
				clientwin->set_flag(clientwin, HELP, True);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_HIDE_HELP :
				clientwin->set_flag(clientwin, HELP, False);
				clientwin->redraw(clientwin);
				break;

			case AC_CMD_GRAB_ON :
				clientwin->needs_grab = True;
				break;

			case AC_CMD_GRAB_OFF :
				clientwin->needs_grab = False;
				break;

			case AC_CMD_PRINTER_STATUS:
				{
					window *w;

					wm->is_printer_availible = ac_printer_status();
					For_All_Windows_Do(wm, w);
						if(w->wtype == NORMAL_WIN) {
							if(wm->is_printer_availible && w->flags.bits.needs_printer) {
								if(!w->flags.bits.printer) {
									w->set_flag(w, PRINTER, True);
									w->redraw(w);
								}
							} else {
								if(w->flags.bits.printer) {
									w->set_flag(w, PRINTER, False);
									w->redraw(w);
								}
							}
						}
					For_All_Windows_Done(wm, w);
				}
				break;

			default :
				printf("_AC_COMMAND client message not handled \n");
				break;
		}
	} else if(xev->message_type == wm->atomlist[WM_CHANGE_STATE]) {

		if(clientwin) {
			if(xev->format == 32 && xev->data.l[0] == IconicState)
				clientwin->iconify(clientwin);
		}
	} else if(xev->message_type == wm->atomlist[XSTROKE_RECOGNITION_CMD]) {

		if(wm->xstroke_window) {
			if(xev->format == 32 && xev->data.l[0] == XSTROKE_STOP_RECOGNITION)
				wm->xstroke_window->ignore_next_unmap=True;
		}
	} else {
		ac_ewmh_handle_clientmessage(wm, xev, clientwin);
	}
}

// Handle changes like window title, name
static void ac_handle_property_notify(
		wm *wm,
		XPropertyEvent *xev,
		window *w)
{
	if(w) {
		if(xev->atom == XA_WM_NAME) {
			if(w->title)
				XFree(w->title);
			XFetchName(wm->display, w->app_window, &w->title);
			w->redraw(w);
			// TODO : handle duplicate window names
		} else if(xev->atom == wm->atomlist[_NET_WM_NAME]) {
			unsigned char tsval[MAX_PROPERTY_LENGTH];

			if(w->name)
				free(w->name);
			w->name = NULL;

			tsval[0]=0;
			window_get_text_property(w, _NET_WM_NAME, tsval);
			if(tsval[0]!=0)
				w->name = strdup(tsval);
			else {
				XClassHint xch;
				// Try WM_CLASS  instead ...
				if(XGetClassHint(wm->display, w->app_window, &xch)) {
					w->name = strdup(xch.res_name);
					XFree(xch.res_name);
					XFree(xch.res_class);
				} else {
					// If nothing is possible, use default :)
					w->name = strdup("default");
				}
			}
		} else if(xev->atom == wm->atomlist[_NET_WM_VISIBLE_NAME]) {
			unsigned char tsval[MAX_PROPERTY_LENGTH];
			if(w->title)
				XFree(w->title);
			w->title=NULL;

			tsval[0]=0;
			window_get_text_property(w, _NET_WM_VISIBLE_NAME, tsval);
			if(tsval[0]!=0)
				w->title = strdup(tsval);
			else {
				XFetchName(wm->display, w->app_window, &w->title);
			}

			if(!w->title && w->name) {
				// Finally just settle on the name itself
				w->title = strdup(w->name);
				XStoreName(wm->display, w->app_window, w->title);
			}
			w->redraw(w);
		} else {
			DBG(debug_info("Not handling property change");)
		}
	}
}

// Initialize a list of all atoms
static void ac_init_atoms(wm *wm)
{
	wm->atomlist[WM_CLASS]=XInternAtom(wm->display,"WM_CLASS",False);
	wm->atomlist[WM_STATE]=XInternAtom(wm->display,"WM_STATE",False);
	wm->atomlist[WM_SAVE_YOURSELF]=XInternAtom(wm->display,"WM_SAVE_YOURSELF",False);
	wm->atomlist[WM_DELETE_WINDOW]=XInternAtom(wm->display,"WM_DELETE_WINDOW",False);
	wm->atomlist[WM_CHANGE_STATE]=XInternAtom(wm->display,"WM_CHANGE_STATE",False);
	wm->atomlist[WM_PROTOCOLS]=XInternAtom(wm->display,"WM_PROTOCOLS",False);
	wm->atomlist[_NET_WM_STATE]=XInternAtom(wm->display,"_NET_WM_STATE",False);
	wm->atomlist[_NET_WM_STATE_FULLSCREEN]=XInternAtom(wm->display,"_NET_WM_STATE_FULLSCREEN",False);
	wm->atomlist[_NET_WM_WINDOW_TYPE_TOOLBAR]=XInternAtom(wm->display,"_NET_WM_WINDOW_TYPE_TOOLBAR",False);
	wm->atomlist[_NET_WM_WINDOW_TYPE_DIALOG]=XInternAtom(wm->display,"_NET_WM_WINDOW_TYPE_DIALOG",False);
	wm->atomlist[_NET_WM_WINDOW_TYPE_SPLASH]=XInternAtom(wm->display,"_NET_WM_WINDOW_TYPE_SPLASH",False);
	wm->atomlist[_NET_ACTIVE_WINDOW]=XInternAtom(wm->display,"_NET_ACTIVE_WINDOW",False);
	wm->atomlist[_NET_CLOSE_WINDOW]=XInternAtom(wm->display,"_NET_CLOSE_WINDOW",False);
	wm->atomlist[_NET_SUPPORTED]=XInternAtom(wm->display,"_NET_SUPPORTED",False);
	wm->atomlist[_NET_NUMBER_OF_DESKTOPS]=XInternAtom(wm->display,"_NET_NUMBER_OF_DESKTOPS",False);
	wm->atomlist[_NET_CURRENT_DESKTOP]=XInternAtom(wm->display,"_NET_CURRENT_DESKTOP",False);
	wm->atomlist[_NET_CLIENT_LIST]=XInternAtom(wm->display,"_NET_CLIENT_LIST",False);
	wm->atomlist[_NET_SUPPORTING_WM_CHECK]=XInternAtom(wm->display,"_NET_SUPPORTING_WM_CHECK",False);
	wm->atomlist[_NET_WM_NAME]=XInternAtom(wm->display,"_NET_WM_NAME",False);
	wm->atomlist[_NET_WM_VISIBLE_NAME]=XInternAtom(wm->display,"_NET_WM_VISIBLE_NAME",False);
	wm->atomlist[_NET_WM_WINDOW_TYPE]=XInternAtom(wm->display,"_NET_WM_WINDOW_TYPE",False);
	wm->atomlist[_NET_WM_PID]=XInternAtom(wm->display,"_NET_WM_PID",False);
	wm->atomlist[_NET_WM_PING]=XInternAtom(wm->display,"_NET_WM_PING",False);
	wm->atomlist[_NET_WM_STATE_FULLSCREEN]=XInternAtom(wm->display,"_NET_WM_STATE_FULLSCREEN",False);
	wm->atomlist[_AC_WM_THEME]=XInternAtom(wm->display,"_AC_WM_THEME",False);
	wm->atomlist[_AC_WM_LAUNCH_CMD]=XInternAtom(wm->display,"_AC_WM_LAUNCH_CMD",False);
	wm->atomlist[_AC_WM_PID_MAP]=XInternAtom(wm->display,"_AC_WM_PID_MAP",False);
	wm->atomlist[_AC_WM_LAYER_MAP]=XInternAtom(wm->display,"_AC_WM_LAYER_MAP",False);
	wm->atomlist[_AC_WM_COMMAND]=XInternAtom(wm->display,"_AC_WM_COMMAND",False);
	wm->atomlist[_AC_W_FLAGS]=XInternAtom(wm->display,"_AC_W_FLAGS",False);
	wm->atomlist[_AC_W_APPID]=XInternAtom(wm->display,"_AC_W_APPID",False);
	wm->atomlist[_AC_W_LAUNCHER]=XInternAtom(wm->display,"_AC_W_LAUNCHER",False);
	wm->atomlist[_AC_W_CMDLINE]=XInternAtom(wm->display,"_AC_W_CMDLINE",False);
	wm->atomlist[_AC_W_TITLE_FONT]=XInternAtom(wm->display,"_AC_W_TITLE_FONT",False);
	wm->atomlist[_AC_W_TITLE_JUSTIFY]=XInternAtom(wm->display,"_AC_W_TITLE_JUSTIFY",False);
	wm->atomlist[_AC_W_STATUS_MSG]=XInternAtom(wm->display,"_AC_W_STATUS_MSG",False);
	wm->atomlist[_AC_W_STATUS_FONT]=XInternAtom(wm->display,"_AC_W_STATUS_FONT",False);
	wm->atomlist[_AC_W_STATUS_JUSTIFY]=XInternAtom(wm->display,"_AC_W_STATUS_JUSTIFY",False);
	wm->atomlist[_AC_W_PAGE_NUM]=XInternAtom(wm->display,"_AC_W_PAGE_NUM",False);
	wm->atomlist[_AC_W_PAGE_NUM_FONT]=XInternAtom(wm->display,"_AC_W_PAGE_NUM_FONT",False);
	wm->atomlist[_AC_W_PAGE_NUM_JUSTIFY]=XInternAtom(wm->display,"_AC_W_PAGE_NUM_JUSTIFY",False);
	wm->atomlist[_AC_W_KEYBOARD_LAYOUT]=XInternAtom(wm->display,"_AC_W_KEYBOARD_LAYOUT",False);
	wm->atomlist[_AC_W_KEYBOARD_CATEGORY]=XInternAtom(wm->display,"_AC_W_KEYBOARD_CATEGORY",False);
	wm->atomlist[_AC_W_ANNOTATE_KEY]=XInternAtom(wm->display,"_AC_W_ANNOTATE_KEY",False);
	wm->atomlist[_AC_W_ANNOTATE_STATE]=XInternAtom(wm->display,"_AC_W_ANNOTATE_STATE",False);
	wm->atomlist[_AC_W_FAST_LOAD]=XInternAtom(wm->display,"_AC_W_FAST_LOAD",False);
	wm->atomlist[_AC_W_CHIKKI]=XInternAtom(wm->display,"_AC_W_CHIKKI",False);
#ifdef __arm__
	wm->atomlist[_AC_W_ROTATION_STATE]=XInternAtom(wm->display,"_AC_W_ROTATION_STATE",False);
#endif
	wm->atomlist[_AC_W_HELP_KEY]=XInternAtom(wm->display,"_AC_W_HELP_KEY",False);
	wm->atomlist[_AC_W_EMAIL_FILE]=XInternAtom(wm->display,"_AC_W_EMAIL_FILE",False);
	wm->atomlist[_AC_W_DEL_ANNOTATION]=XInternAtom(wm->display,"_AC_W_DEL_ANNOTATION",False);
	wm->atomlist[_AC_W_LAYER_WINDOWID]=XInternAtom(wm->display,"_AC_W_LAYER_WINDOWID",False);
	wm->atomlist[_AC_BUTTON_UP]=XInternAtom(wm->display,"_AC_BUTTON_UP",False);
	wm->atomlist[_AC_BUTTON_UP_REL]=XInternAtom(wm->display,"_AC_BUTTON_UP_REL",False);
	wm->atomlist[_AC_BUTTON_EMAIL]=XInternAtom(wm->display,"_AC_BUTTON_EMAIL",False);
	wm->atomlist[_AC_BUTTON_EMAIL_REL]=XInternAtom(wm->display,"_AC_BUTTON_EMAIL_REL",False);
	wm->atomlist[_AC_BUTTON_NEXTPAGE]=XInternAtom(wm->display,"_AC_BUTTON_NEXTPAGE",False);
	wm->atomlist[_AC_BUTTON_NEXTPAGE_REL]=XInternAtom(wm->display,"_AC_BUTTON_NEXTPAGE_REL",False);
	wm->atomlist[_AC_BUTTON_PREVPAGE]=XInternAtom(wm->display,"_AC_BUTTON_PREVPAGE",False);
	wm->atomlist[_AC_BUTTON_PREVPAGE_REL]=XInternAtom(wm->display,"_AC_BUTTON_PREVPAGE_REL",False);
	wm->atomlist[_AC_BUTTON_PRINTER]=XInternAtom(wm->display,"_AC_BUTTON_PRINTER",False);
	wm->atomlist[_AC_BUTTON_PRINTER_REL]=XInternAtom(wm->display,"_AC_BUTTON_PRINTER_REL",False);
	wm->atomlist[_AC_BUTTON_PREF]=XInternAtom(wm->display,"_AC_BUTTON_PREF",False);
	wm->atomlist[_AC_BUTTON_PREF_REL]=XInternAtom(wm->display,"_AC_BUTTON_PREF_REL",False);
	wm->atomlist[_AC_BUTTON_RETRACTBAR]=XInternAtom(wm->display,"_AC_BUTTON_RETRACTBAR",False);
	wm->atomlist[_AC_BUTTON_RETRACTBAR_REL]=XInternAtom(wm->display,"_AC_BUTTON_RETRACTBAR_REL",False);
	wm->atomlist[_AC_BUTTON_VOICE]=XInternAtom(wm->display,"_AC_BUTTON_VOICE",False);
	wm->atomlist[_AC_BUTTON_VOICE_REL]=XInternAtom(wm->display,"_AC_BUTTON_VOICE_REL",False);
	wm->atomlist[_AC_BUTTON_SPEAKER]=XInternAtom(wm->display,"_AC_BUTTON_SPEAKER",False);
	wm->atomlist[_AC_BUTTON_SPEAKER_REL]=XInternAtom(wm->display,"_AC_BUTTON_SPEAKER_REL",False);
	wm->atomlist[_AC_BUTTON_TICK]=XInternAtom(wm->display,"_AC_BUTTON_TICK",False);
	wm->atomlist[_AC_BUTTON_TICK_REL]=XInternAtom(wm->display,"_AC_BUTTON_TICK_REL",False);
	wm->atomlist[_AC_BUTTON_CLOSE]=XInternAtom(wm->display,"_AC_BUTTON_CLOSE",False);
	wm->atomlist[_AC_BUTTON_CLOSE_REL]=XInternAtom(wm->display,"_AC_BUTTON_CLOSE_REL",False);
	wm->atomlist[_AC_BUTTON_ROCKER_UP]=XInternAtom(wm->display,"_AC_BUTTON_ROCKER_UP",False);
	wm->atomlist[_AC_BUTTON_ROCKER_UP_REL]=XInternAtom(wm->display,"_AC_BUTTON_ROCKER_UP_REL",False);
	wm->atomlist[_AC_BUTTON_ROCKER_DOWN]=XInternAtom(wm->display,"_AC_BUTTON_ROCKER_DOWN",False);
	wm->atomlist[_AC_BUTTON_ROCKER_DOWN_REL]=XInternAtom(wm->display,"_AC_BUTTON_ROCKER_DOWN_REL",False);
	wm->atomlist[_AC_BUTTON_ROCKER_SELECT]=XInternAtom(wm->display,"_AC_BUTTON_ROCKER_SELECT",False);
	wm->atomlist[_AC_BUTTON_ROCKER_SELECT_REL]=XInternAtom(wm->display,"_AC_BUTTON_ROCKER_SELECT_REL",False);
	wm->atomlist[_AC_ACCEL_RIGHT_FLICK]=XInternAtom(wm->display,"_AC_ACCEL_RIGHT_FLICK",False);
	wm->atomlist[_AC_ACCEL_LEFT_FLICK]=XInternAtom(wm->display,"_AC_ACCEL_LEFT_FLICK",False);
	wm->atomlist[_AC_ACCEL_TOP_FLICK]=XInternAtom(wm->display,"_AC_ACCEL_TOP_FLICK",False);
	wm->atomlist[_AC_ACCEL_BOTTOM_FLICK]=XInternAtom(wm->display,"_AC_ACCEL_BOTTOM_FLICK",False);
	wm->atomlist[_AC_ACCEL_RIGHT_EXTREME]=XInternAtom(wm->display,"_AC_ACCEL_RIGHT_EXTREME",False);
	wm->atomlist[_AC_ACCEL_LEFT_EXTREME]=XInternAtom(wm->display,"_AC_ACCEL_LEFT_EXTREME",False);
	wm->atomlist[_AC_ACCEL_TOP_EXTREME]=XInternAtom(wm->display,"_AC_ACCEL_TOP_EXTREME",False);
	wm->atomlist[_AC_ACCEL_BOTTOM_EXTREME]=XInternAtom(wm->display,"_AC_ACCEL_BOTTOM_EXTREME",False);
	wm->atomlist[XSTROKE_RECOGNITION_CMD]=XInternAtom(wm->display,"XSTROKE_RECOGNITION_CMD",False);
}

// Get frame and window pointer from the X Event
static window *ac_get_event_xwindow(wm *wm, Window xwin, event_area *where)
{
	window *w, *ps;

	*where = NOWHERE;
	if(wm->mapped_window)
		ps = wm->mapped_window;
	else
		ps = wm->window_list;

	if(!ps)
		return NULL;

	//For_All_Mappable_PS_Do(wm, ps, w);
	For_All_Windows_Do(wm, w);
		if(xwin == w->app_window) {
			*where = APP_FRAME;
			return w;
		} else if(xwin == w->frame_window) {
			*where = TOPLEVEL_FRAME;
			return w;
		} else if(w->topbar_frame && xwin == w->topbar_frame->win){
			*where = TOPBAR_FRAME;
			return w;
		} else if(w->bottombar_frame && xwin == w->bottombar_frame->win){
			*where = BOTTOMBAR_FRAME;
			return w;
		}
	//For_All_Mappable_PS_Done(wm, ps, w);
	For_All_Windows_Done(wm, w);
	return NULL;
}

// Get frame and window pointer from the X Event
static window *ac_get_event_window(wm *wm, XEvent *xev, event_area *where)
{
	window *w, *ps;
	Window xwin=xev->xany.window;

	*where = NOWHERE;
	if(wm->mapped_window)
		ps = wm->mapped_window;
	else
		ps = wm->window_list;

	if(!ps)
		return NULL;

	//For_All_Mappable_PS_Do(wm, ps, w);
	For_All_Windows_Do(wm, w);
		if(xwin == w->app_window) {
			*where = APP_FRAME;
			return w;
		} else if(xwin == w->frame_window) {
			*where = TOPLEVEL_FRAME;
			return w;
		} else if(w->topbar_frame && xwin == w->topbar_frame->win){
			*where = TOPBAR_FRAME;
			return w;
		} else if(w->bottombar_frame && xwin == w->bottombar_frame->win){
			*where = BOTTOMBAR_FRAME;
			return w;
		}
	//For_All_Mappable_PS_Done(wm, ps, w);
	For_All_Windows_Done(wm, w);
	return NULL;
}

// Get window pointer for the specified X window
window *ac_get_window(wm *wm, Window win)
{
	window *w;

	For_All_Windows_Do(wm, w);
		if(win == w->app_window)
			return w;
	For_All_Windows_Done(wm, w);
	return NULL;
}

// Figure out which widget needs to handle it and pass it accordingly
static void ac_handle_widget_event(wm *wm, XEvent *xev, window *w, event_area where)
{
	if(w) {
		px_event pxe;
		px_eventloop* eventloop = eventloop_get_eventloop();
		px_widget* toplevel = NULL;
		
		pxe.xev = xev;
		pxe.private_data = w;
		pxe.type = NR_CALLBACKS;

		if(w->wtype == NORMAL_WIN) {
			if(where == TOPBAR_FRAME)
				toplevel = w->topbar_frame;
			else if(where == BOTTOMBAR_FRAME)
				toplevel = w->bottombar_frame;
		} else if(w->wtype == DIALOG_WIN) {
			if(where == TOPBAR_FRAME)
				toplevel = w->topbar_frame;
		} else if(w->wtype == MENU_WIN) {
			toplevel = w->topbar_frame;
		}

		eventloop->handle_event(eventloop, &pxe, toplevel);

		if (w->wtype == MENU_WIN)
		{
			// The needs_keyboard bit is a hack to ensure that widgets
			// are all destroyed __after__ the event is handled
			if(w->flags.bits.needs_keyboard) {
				XGrabServer(wm->display);
				w->destroy(w);
				XUngrabServer(wm->display);
			}
		}
	}
}

#ifdef __arm__
// Initialize rotation tables for various LCDs 
static void initialize_rotation_table()
{
	int fd;

	fd=open("/proc/simputer",O_RDONLY);
	if(fd<0) {
		/* V3 Mono Display : h/w orientation landscape rotated 90 */
		rotation_table[ROT_0]=2;
		rotation_table[ROT_90]=4;
		rotation_table[ROT_180]=8;
		rotation_table[ROT_270]=1;
	} else {
		char buf[15];

		read(fd,buf,14);
		if(!strncmp(buf,"v3",2) || !strncmp(buf,"v4_mono",7)) {
			/* V3/V4 Mono Display : h/w orientation landscape rotated 90 */
			rotation_table[ROT_0]=2;
			rotation_table[ROT_90]=4;
			rotation_table[ROT_180]=8;
			rotation_table[ROT_270]=1;
		} else if(!strncmp(buf,"v4_stn",6)) {
			/* V4 STN Display : h/w orientation inverted */
			rotation_table[ROT_0]=4;
			rotation_table[ROT_90]=8;
			rotation_table[ROT_180]=1;
			rotation_table[ROT_270]=2;
		} else {
			/* V4 TFT Display : h/w orientation normal */
			rotation_table[ROT_0]=1;
			rotation_table[ROT_90]=2;
			rotation_table[ROT_180]=4;
			rotation_table[ROT_270]=8;
		}
		close(fd);
	}

	rotation_initialized = True;
}

// Rotate the X Server display to the specified orientation
static Bool ac_rotate_to(wm *wm, rotation_type rot)
{
	Rotation lcd_rotation, current_rotation;

	XRRScreenConfiguration *sc;
	if(!rotation_initialized)
		initialize_rotation_table();
	
	lcd_rotation = rotation_table[rot];

	/* 
	 * Get current rotation and determine
	 *  if rotation is required
	 */
	sc = XRRGetScreenInfo(wm->display, wm->root);
	XRRConfigCurrentConfiguration(sc, &current_rotation);

	if (current_rotation == lcd_rotation)
		return False;
	
	XGrabServer(wm->display);
	XRRSetScreenConfig (wm->display, sc, wm->root,
         (SizeID) 0, lcd_rotation, CurrentTime);
	XSync(wm->display, False);
	XUngrabServer(wm->display);
	wm->current_rot = rot;
	return True;
}

static Bool is_rotation_event(Display *dpy, XEvent *xev, char *wm_arg)
{
	wm *wm = (struct _wm *) wm_arg;

	if(xev->type == Expose && xev->xexpose.window == wm->root)
		return True;
}

#endif

// Allocate the basic window manager structures
wm *ac_init(const char *theme)
{
	wm *ac;
	FILE *fp;

	XSetWindowAttributes root_attr;
	px_display* display;
#ifdef __arm__
	XEvent xev;
#endif

	// SIGCHLD handler so that zombies are not created
	init_signals();

	ac=(wm *)malloc(sizeof (wm));
	if(!ac)
		err_exit("Could not allocate wm structure");
	if(!getenv("DISPLAY"))
		ac->display = XOpenDisplay(":0.0");
	else
		ac->display = XOpenDisplay(NULL);
	if(!ac->display)
		err_exit("Could not open display");

	
	display = get_display(); // For Picowidget a global variable!!
	display->set_display(display, ac->display);

#ifdef __XSYNC__
	XSynchronize(ac->display, True);
#endif

	ac->screen = DefaultScreen(ac->display);
	ac->root = DefaultRootWindow(ac->display);

	root_attr.event_mask = SubstructureRedirectMask | SubstructureNotifyMask |
				ExposureMask | StructureNotifyMask | PropertyChangeMask;

	// Whee: I am a WM !!
	// Surprisingly , all it takes is a small change of root window attributes
	XChangeWindowAttributes(ac->display, ac->root, CWEventMask, &root_attr);

#ifdef __arm__
	// Rotate the screen first
	if ( ac_rotate_to(ac, ROT_0) ) {
		XEvent rex;

		XIfEvent(ac->display, &rex, is_rotation_event, (char *)ac);
		ac->display_width = rex.xexpose.width;
		ac->display_height = rex.xexpose.height;
	} else {
		ac->display_width = DisplayWidth(ac->display, ac->screen);
		ac->display_height = DisplayHeight(ac->display, ac->screen);
	}
#else
	ac->display_width = DisplayWidth(ac->display, ac->screen);
	ac->display_height = DisplayHeight(ac->display, ac->screen);
#endif
	ac->black_pixel = BlackPixel(ac->display, ac->screen);
	ac->white_pixel = WhitePixel(ac->display, ac->screen);
	
  	XSetErrorHandler(xerror_handler);
	ac_init_atoms(ac);
	ac_init_keys(ac);
	
	ac->window_list = NULL;
	ac->mapped_window = NULL;
	ac->focused_window = NULL;
	ac->launcher_window = NULL;
	ac->keyboard_window = NULL;
	ac->xstroke_window = NULL;
	ac->clock_window = NULL;
	ac->on_top_list = NULL;
	ac->identity_save_window = NULL;

	ac->is_keyboard_mapped = False;
	ac->is_menu_mapped = False;
	ac->is_printer_availible = ac_printer_status();
	ac->current_kbd_category = strdup("english");
	ac->current_kbd_layout = strdup("qwerty");
	ac->num_windows = 0;

	// Allocate and load the theme, then set the atom on the root window
	ac->current_theme = new_theme(ac);
	ac->current_theme->load_theme(ac->current_theme, theme, ac->root);
	xwindow_set_text_property(ac, ac->root, _AC_WM_THEME, theme);

	ac->mainloop=ac_mainloop;
	ac->handle_event=ac_handle_event;
	ac->get_toolbar_displacement=ac_get_toolbar_displacement;
	ac->retile=ac_retile;

#ifdef __arm__
	ac->rotate_to=ac_rotate_to;
	{
#include "blank_cursor.xbm"
		XColor col;
		Pixmap pix = XCreatePixmapFromBitmapData (ac->display, ac->root, cur_bits, cur_width, cur_height, 0, 0, 1);

		memset (&col, 0, sizeof (col));
		ac->normal_cursor = XCreatePixmapCursor (ac->display, pix, pix, &col, &col, 1, 1);
	}
#else
	ac->normal_cursor = XCreateFontCursor(ac->display, XC_left_ptr);
#endif
	XDefineCursor(ac->display, ac->root, ac->normal_cursor);

	ac->exit=ac_exit;

	spawn_child(IDENTITY_CMDLINE,NULL);

	ac_adopt_existing(ac);	// Take care of already existing windows

	if(!ac->keyboard_window)
		keyboard_start_keyboard("english","qwerty");	// Pre-launch keyboard

	if(!ac->clock_window)
		clock_start_clock(theme);	// Start the launcher clock

	fp = fopen(PID_FILE,"w");
    
	if(fp) {
		fprintf(fp, "%d", getpid());
		fclose(fp);
	}

	return ac;
}

// This is for client side acctrl
// to have a shared memory map of all pids/appids
void ac_make_pid_appid_map(wm *wm)
{
	int bufsize, len;
	int lbufsize, llen;
	char *ac_apid_map,*ac_layer_map;
	unsigned long tmplen;
	char tmpbuf[TMP_BUF_SIZE];
	window *w;

	// Go through all windows, make a list and set the atom
	lbufsize=bufsize=1024;
	llen=len=0;
	ac_apid_map = (char *)malloc(bufsize);
	if(!ac_apid_map)
		err_exit("Unable to allocate memory");
	ac_layer_map = (char *)malloc(lbufsize);
	if(!ac_layer_map)
		err_exit("Unable to allocate memory");
	ac_apid_map[0] = 0;
	ac_layer_map[0] = 0;

	For_All_Mappable_Do(wm, w);
		sprintf(tmpbuf, "%lu=%d,",w->appid, w->pid);
		tmplen=mystrlen(tmpbuf);
		if(tmplen + len >= bufsize) {
			bufsize += 1024;
			ac_apid_map = (char *)realloc(ac_apid_map, bufsize);
			if(!ac_apid_map)
				err_exit("Unable to allocate memory");
		}
		strcat(ac_apid_map, tmpbuf);
		len+=tmplen;
		{
			// Update layer map
			window *lw;

			for(lw=w;lw->layer_below;lw=lw->layer_below) {
				sprintf(tmpbuf, "%lu=%d,",lw->layer_below->app_window, lw->pid);
				tmplen=mystrlen(tmpbuf);
				if(tmplen + llen >= lbufsize) {
					lbufsize += 1024;
					ac_layer_map = (char *)realloc(ac_layer_map, lbufsize);
					if(!ac_layer_map)
						err_exit("Unable to allocate memory");
				}
				strcat(ac_layer_map, tmpbuf);
				llen+=tmplen;
			}
		}
	For_All_Mappable_Done(wm, w);

	xwindow_set_text_property(wm, wm->root, _AC_WM_PID_MAP, ac_apid_map);
	xwindow_set_text_property(wm, wm->root, _AC_WM_LAYER_MAP, ac_layer_map);
	free(ac_apid_map);
	free(ac_layer_map);
}

// somewhere some event happenned.. do something about it
void ac_handle_event(wm *wm, XEvent *xev)
{
	event_area where;

	//window *w = ac_get_event_window(wm, xev, &where);
	window *w = NULL;

	DBG(debug_info("Event type %d received", xev->type);)

	switch(xev->type) {
		case MapRequest:
			w = ac_get_event_xwindow(wm, xev->xmap.window, &where);
			ac_handle_map_request(wm, &xev->xmap, w);
			break;
		case UnmapNotify:
			w = ac_get_event_xwindow(wm, xev->xunmap.window, &where);
			if(where != TOPBAR_FRAME && where != BOTTOMBAR_FRAME)
				ac_handle_unmap_notify(wm, &xev->xunmap, w);
			break;
		case DestroyNotify:
			w = ac_get_event_xwindow(wm, xev->xdestroywindow.window, &where);
			ac_handle_destroy_notify(wm, xev, w);
			break;
		case ButtonPress:
			w = ac_get_event_window(wm, xev, &where);
			ac_handle_button_press(wm, xev, w, where);
			break;
		case KeyPress:
			//ac_handle_key_press(wm, &xev->xkey);
			ac_handle_key_press(wm, xev);
			break;
		case KeyRelease:
			//ac_handle_key_release(wm, &xev->xkey);
			ac_handle_key_release(wm, xev);
			break;
		case ConfigureRequest:
			w = ac_get_event_xwindow(wm, xev->xconfigurerequest.window, &where);
			ac_handle_configure_request(wm, &xev->xconfigurerequest, w, where);
			break;
		case ConfigureNotify:
			ac_handle_configure_notify(wm, &xev->xconfigure);
			break;
		case ClientMessage:
			w = ac_get_event_window(wm, xev, &where);
			ac_handle_client_message(wm, &xev->xclient);
			break;
		case PropertyNotify:
			w = ac_get_event_window(wm, xev, &where);
			ac_handle_property_notify(wm, &xev->xproperty, w);
			break;
		case Expose:
		case LeaveNotify:
		case MotionNotify:
		case EnterNotify:
		case ButtonRelease:
			w = ac_get_event_window(wm, xev, &where);
			ac_handle_widget_event(wm, xev, w, where);
			break;
		default:
			DBG(debug_info("Event type %d unhandled!", xev->type);)
			break;
	}
}


// The never ending loop .. unless someone kills me
void ac_mainloop(wm *wm)
{
	XEvent xev;

	while(1) {
		XNextEvent(wm->display, &xev);
		wm->handle_event(wm, &xev);
	}
}

// Au Revior .. init will start me again....
// I may die.. but I will respawn
void ac_exit(wm *wm)
{
	XCloseDisplay(wm->display);
	if(wm->current_theme)
		free(wm->current_theme);
	if(wm->current_kbd_category)
		free(wm->current_kbd_category);
	if(wm->current_kbd_layout)
		free(wm->current_kbd_layout);
	free(wm);
	unlink(PID_FILE);

	exit(0);
}
