/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>

#include "alchemy.h"
#include "acctrl.h"
#include "annotation_glue.h"

char *atom_string[] = {
	"",
	"_AC_WM_THEME",
	"_AC_WM_COMMAND",
	"3",
	"4",
	"_AC_W_FLAGS",
	"_AC_WM_LAUNCH_CMD",
	"_AC_W_TITLE_FONT",
	"_AC_W_TITLE_JUSTIFY",
	"_AC_W_STATUS_MSG",
	"_AC_W_STATUS_FONT",
	"_AC_W_STATUS_JUSTIFY",
	"_AC_W_PAGE_NUM",
	"_AC_W_PAGE_NUM_FONT",
	"_AC_W_PAGE_NUM_JUSTIFY",
	"_AC_W_KEYBOARD_LAYOUT",
	"_AC_W_KEYBOARD_CATEGORY",
	"_AC_W_ANNOTATE_KEY",
	"_AC_W_HELP_KEY",
	"_AC_W_EMAIL_FILE",
	"_AC_W_DEL_ANNOTATION",
#ifdef __arm__
	"_AC_W_ROTATION_STATE",
#endif
};

static void xwindow_get_text_property(Display *dpy, Window win, const char *atom_name, unsigned char *string_ret)
{
	unsigned char *prop_ret;
	Atom at, actual_type;
	int status, actual_format;
	unsigned long n,b;

	at=XInternAtom(dpy, atom_name, False);

	status = XGetWindowProperty(dpy, win, at, 0,
			MAX_PROPERTY_LENGTH, False, AnyPropertyType, &actual_type,
			&actual_format, &n, &b, &prop_ret);

	if( (status!=Success) || (!n) || (!prop_ret) )
		string_ret[0] = 0;
	else {
		strcpy(string_ret, prop_ret);
		XFree(prop_ret);
	}
}

// Set a X window property atom as text
static void xwindow_set_text_property(Display *dpy, Window win, const char *atom_name, const char *property_value)
{
	Atom at;
	int status;

	at=XInternAtom(dpy, atom_name, False);

	status = XChangeProperty(dpy, win, at, XA_STRING, 8,
			PropModeReplace, property_value, strlen(property_value));
	XFlush(dpy);
}

// Returns true is a process with the specified pid exists
static Bool process_exists(pid_t pid)
{
	struct stat sbuf;
	char filename[TMP_BUF_SIZE];

	sprintf(filename, "/proc/%d", pid);
	if(stat(filename, &sbuf)<0)
		return False;
	return True;
}

// Get the appid of the specified window
unsigned long getWindowAppid(Window win, Display *dpy)
{
	Atom prop;
	Atom realType;
	unsigned long n;
	unsigned long extra;
	int format;
	int status;
	char * value;

	prop = XInternAtom(dpy, "_AC_W_APPID", True);
	if (prop == None) {
		fprintf(stderr, "No such property '_AC_W_APPID'\n");
		return 0;
	}

	status = XGetWindowProperty(dpy, win, prop, 0L, 512L, False, AnyPropertyType, 
			&realType, &format, &n, &extra, (unsigned char **) &value);

	if (status != Success  || value == 0 ) {
		fprintf(stderr, "couldn't read property on window\n");
		return 0;
	}

	return ((unsigned long) (*((unsigned long *) (value))));
}

// This api allows one to figure out whether the given appid is
// running as a process or not 
Bool ac_is_appid_running(Display *dpy, unsigned long appid)
{
	char data[MAX_PROPERTY_LENGTH];

	data[0]=0;

	// Get the appid to pid map from the server
	xwindow_get_text_property(dpy, DefaultRootWindow(dpy),
			"_AC_WM_PID_MAP", data);

	if(strlen(data)) {
		char *tok_begin, *tok_sep;

		tok_begin = data;
		
		while(*tok_begin) {
			unsigned long app_id;
			char *eq_ptr;

			if(!(tok_sep = strchr(tok_begin, ',')))
				break;

			app_id = strtoul(tok_begin, &eq_ptr, 10);

			if(app_id == appid) {
				unsigned long mapped_pid;

				mapped_pid = strtoul(eq_ptr + 1, NULL, 10);
				if(process_exists(mapped_pid)) {
					return True;
				}
			}
			
			tok_begin = tok_sep + 1;
		}
	}
	return False;
}

// This is the code executed by all windows on startup before
// the window is mapped. This will help in eliminating multiple startup
// instances of the same program and also set various atoms required by
// alchemy.
void ac_init_app(Display *dpy, Window win)
{
	unsigned long my_appid=0;
	pid_t my_pid=0;
	char *env_var=NULL;

	if(!dpy) {
		return;
	}

	XSync(dpy, False); // Make sure that X Events are all done

	env_var = getenv("_AC_W_APPID");

	if(env_var) {
		Atom app_atom;
		unsigned long appid[1];
		my_appid = appid[0] = strtoul(env_var, NULL, 10);
		app_atom = XInternAtom(dpy, "_AC_W_APPID", False);
		if (app_atom != None) {
			XChangeProperty(dpy, win, app_atom, XA_CARDINAL, 32, 
					PropModeReplace, (unsigned char *)appid, 1);
		}
		unsetenv("_AC_W_APPID");
	}

	env_var = getenv("_AC_W_FAST_LOAD");

	if(env_var) {
		Atom app_atom;
		unsigned long fast_load[1];
		fast_load[0] = strtoul(env_var, NULL, 10);
		app_atom = XInternAtom(dpy, "_AC_W_FAST_LOAD", False);
		if (app_atom != None) {
			XChangeProperty(dpy, win, app_atom, XA_CARDINAL, 32, 
					PropModeReplace, (unsigned char *)fast_load, 1);
		}
		unsetenv("_AC_W_FAST_LOAD");
	}

	env_var = getenv("_NET_WM_NAME");
	if(env_var) {
		Atom wname_atom;
		wname_atom = XInternAtom(dpy, "_NET_WM_NAME", False);
		if (wname_atom != None) {
			XChangeProperty(dpy, win, wname_atom, XA_STRING, 8, 
					PropModeReplace, env_var, mystrlen(env_var));
		}
		unsetenv("_NET_WM_NAME");
	}

	{
		Atom pid_atom;
		unsigned long pidstr[1];

		my_pid = pidstr[0] = getpid();
		pid_atom = XInternAtom(dpy, "_NET_WM_PID", False);
		if (pid_atom != None) {
			XChangeProperty(dpy, win, pid_atom, XA_CARDINAL, 32, 
					PropModeReplace, (unsigned char *)pidstr, 1);
		}
	}

	env_var = getenv("_AC_W_LAYER_WINDOWID");

	if(env_var) {
		// Hmmm.. a layered window, check if someone has already layered
		// the target window or I am THE ONE
		Atom layer_atom;
		unsigned long layer[1];
		char data[MAX_PROPERTY_LENGTH];
		char tmpbuf[100];
		int lock_fd;


		layer[0] = strtoul(env_var, NULL, 10);
		layer_atom = XInternAtom(dpy, "_AC_W_LAYER_WINDOWID", False);
		if (layer_atom != None) {
			XChangeProperty(dpy, win, layer_atom, XA_CARDINAL, 32, 
					PropModeReplace, (unsigned char *)layer, 1);
		}

		if((lock_fd = open(PID_FILE, O_RDWR | O_CREAT)) <0) {
			perror("Cannot open /tmp/alchemy.pid!!");
			exit(1);
		};

		// Get an exclusive lock .. wait till you get one
		// This has to be serialized
		if(flock(lock_fd, LOCK_EX)<0) {
			perror("Error obtaining lock");
			exit(1);
		}

		data[0]=0;
		// Get the appid to pid map from the server
		xwindow_get_text_property(dpy, DefaultRootWindow(dpy),
				"_AC_WM_LAYER_MAP", data);

		if(strlen(data)) {
			char *tok_begin, *tok_sep;

			tok_begin = data;
			
			while(*tok_begin) {
				unsigned long window_layerid;
				char *eq_ptr;

				if(!(tok_sep = strchr(tok_begin, ',')))
					break;

				window_layerid = strtoul(tok_begin, &eq_ptr, 10);

				if(window_layerid == layer[0]) {
					// Check if pid is running
					unsigned long mapped_pid;

					mapped_pid = strtoul(eq_ptr + 1, NULL, 10);
					if(process_exists(mapped_pid)) {
   						exit(1); // Too bad .. some one has already started
					}
				}
				
				tok_begin = tok_sep + 1;
			}
		}

		// If we are still here, then we can add ourselves to the list
		sprintf(tmpbuf,"%lu=%d,",layer[0], my_pid);
		strcat(data,tmpbuf);
		xwindow_set_text_property(dpy, DefaultRootWindow(dpy), "_AC_WM_LAYER_MAP",
				data);

		close(lock_fd);
		unsetenv("_AC_W_LAYER_WINDOWID");
	} else if(my_appid) {
		// Not a layer .. check for appid
		char data[MAX_PROPERTY_LENGTH];
		char tmpbuf[100];
		int lock_fd;

		if((lock_fd = open(PID_FILE, O_RDWR | O_CREAT)) <0) {
			perror("Cannot open /tmp/alchemy.pid!!");
			exit(1);
		};

		// Get an exclusive lock .. wait till you get one
		// This has to be serialized
		if(flock(lock_fd, LOCK_EX)<0) {
			perror("Error obtaining lock");
			exit(1);
		}

		data[0]=0;
		// Get the appid to pid map from the server
		xwindow_get_text_property(dpy, DefaultRootWindow(dpy),
				"_AC_WM_PID_MAP", data);

		if(strlen(data)) {
			char *tok_begin, *tok_sep;

			tok_begin = data;
			
			while(*tok_begin) {
				unsigned long appid;
				char *eq_ptr;

				if(!(tok_sep = strchr(tok_begin, ',')))
					break;

				appid = strtoul(tok_begin, &eq_ptr, 10);

				if(appid == my_appid) {
					unsigned long mapped_pid;

					mapped_pid = strtoul(eq_ptr + 1, NULL, 10);
					if(process_exists(mapped_pid)) {
						close(lock_fd);
						exit(1); // Too bad .. some one has already started
					}
				}
				
				tok_begin = tok_sep + 1;
			}
		}

		// If we are still here, then we can add ourselves to the list
		sprintf(tmpbuf,"%lu=%d,",my_appid, my_pid);
		strcat(data,tmpbuf);
		xwindow_set_text_property(dpy, DefaultRootWindow(dpy), "_AC_WM_PID_MAP",
				data);
		close(lock_fd);
	}

	env_var = getenv("WM_TRANSIENT_FOR");

	if(env_var) {
		Window prop_win;
		prop_win = (Window) strtoul(env_var, NULL, 10);
		XSetTransientForHint(dpy, win, prop_win);
		unsetenv("WM_TRANSIENT_FOR");
	}	

	env_var = getenv("_AC_W_LAUNCHER");

	if(env_var) {
		Atom app_atom;
		unsigned long launcher_appid[1];
		launcher_appid[0] = strtoul(env_var, NULL, 10);
		app_atom = XInternAtom(dpy, "_AC_W_LAUNCHER", False);
		if (app_atom != None) {
			XChangeProperty(dpy, win, app_atom, XA_CARDINAL, 32, 
				PropModeReplace, (unsigned char *)launcher_appid, 1);
		}
		unsetenv("_AC_W_LAUNCHER");
	}	

	env_var = getenv("_AC_W_CHIKKI");

	if(env_var) {
		Atom app_atom;
		unsigned char a=1;

		app_atom = XInternAtom(dpy, "_AC_W_CHIKKI", False);
		if (app_atom != None) {
			XChangeProperty(dpy, win, app_atom, XA_CARDINAL, 32, 
				PropModeReplace, (unsigned char *)&a, 1);
		}
		unsetenv("_AC_W_CHIKKI");
	}	
}

static Atom atom_array[100];

int xwindow_get_int_property(Display *dpy, Window w, const char *atom_name)
{
	Atom at, actual_type;
	int status, actual_format, int_ret;
	int *property_value=0;
	unsigned long n,b;

	at=XInternAtom(dpy, atom_name, False);

	status = XGetWindowProperty(dpy, w, at, 0,
			MAX_PROPERTY_LENGTH, False, AnyPropertyType, &actual_type,
			&actual_format, &n, &b, (unsigned char **)&property_value);
	if( (status!=Success) || (!n) || (!property_value) )
		return -1;
	
	int_ret = property_value[0];
	XFree(property_value);
	return int_ret;
}

int xwindow_set_int_property(Display *dpy, Window win, const char *atom_name, const unsigned int property_value)
{
	Atom at;
	int status;

	at=XInternAtom(dpy, atom_name, False);

	status = XChangeProperty(dpy, win, at, XA_CARDINAL, 32, PropModeReplace,
			(unsigned char *)&property_value, 1);
	return status;
}

Bool is_client_message(Display *dpy, XEvent *xev, char *win);

void ac_control_win(int cmd_id, char *data, Display *dpy, Window win)
{

	XEvent	ev;

	//printf("cmd_id = %d \n",cmd_id);

	// msg_type will be in [2]
	if(!atom_array[2]) {
		atom_array[2]=XInternAtom(dpy, atom_string[2], False);

		if (atom_array[2] == None) {
			fprintf(stderr, "No such property '%s'\n",atom_string[cmd_id]);
			return;
		}
	}

	switch(cmd_id) {
		case AC_CMD_LAYER_NEXT_APP: {
				char tmp[TMP_BUF_SIZE];
				char data[MAX_PROPERTY_LENGTH];

				sprintf(tmp,"_AC_W_APPID=%lu", getWindowAppid(win, dpy));
				putenv(tmp);
				sprintf(tmp,"_AC_W_LAYER_WINDOWID=%lu", win);
				putenv(tmp);
				xwindow_get_text_property(dpy, win, "_NET_WM_NAME", data);
				sprintf(tmp,"_NET_WM_NAME=%s", data);
				putenv(tmp);
			}
			return;

		case AC_CMD_LAYER_THIS_WIN: {
				char tmp_data[MAX_PROPERTY_LENGTH];
				int tmp_num;
				Window new_win;

				new_win = strtoul(data, NULL, 0);
				xwindow_get_text_property(dpy, win, "_NET_WM_NAME", tmp_data);
				xwindow_set_text_property(dpy, new_win, "_NET_WM_NAME", tmp_data);
				tmp_num = xwindow_get_int_property(dpy, win, "_AC_W_APPID");
				if(tmp_num != -1)
					xwindow_set_int_property(dpy, new_win, "_AC_W_APPID", tmp_num);
				xwindow_set_int_property(dpy, new_win, "_AC_W_LAYER_WINDOWID", win);
			}
			return;

		case AC_CMD_SET_THEME :
		case AC_CMD_SET_FLAGS :
		case AC_CMD_LAUNCH_APP :
		case AC_CMD_SET_TITLE_FONT :
		case AC_CMD_SET_TITLE_JUSTIFY :
		case AC_CMD_SET_STATUS_MSG :
		case AC_CMD_SET_STATUS_FONT :
		case AC_CMD_SET_STATUS_JUSTIFY :
		case AC_CMD_SET_PAGE_NUM :
		case AC_CMD_SET_PAGE_NUM_FONT :
		case AC_CMD_SET_PAGE_NUM_JUSTIFY :
		case AC_CMD_SET_KEYBOARD_LAYOUT :
		case AC_CMD_SET_KEYBOARD_CATEGORY :
		case AC_CMD_SET_ANNOTATE_KEY :
		case AC_CMD_SET_HELP_KEY :
#ifdef __arm__
		case AC_CMD_SET_ROTATION :
#endif
		case AC_CMD_SET_EMAIL_FILE : {
			//printf("sending %s data = %s \n",atom_string[cmd_id], data);

			if(!atom_array[cmd_id])  {
				atom_array[cmd_id] = XInternAtom(dpy, atom_string[cmd_id], 
									 False);

				if (atom_array[cmd_id] == None) {
					fprintf(stderr, "No such property '%s'\n",atom_string[cmd_id]);
					return;
				}
			}

			if(cmd_id == AC_CMD_LAUNCH_APP || cmd_id == AC_CMD_SET_THEME)
				XChangeProperty(dpy, DefaultRootWindow(dpy),
					atom_array[cmd_id], XA_STRING, 8,
					PropModeReplace, data, mystrlen(data));
			else
				XChangeProperty(dpy, win, atom_array[cmd_id],
					XA_STRING, 8, PropModeReplace, data,
					mystrlen(data));

			break;

		 }

		case AC_CMD_DEL_ANNOTATION :
			{
				char app_name[MAX_PROPERTY_LENGTH];


				app_name[0]=0;
				xwindow_get_text_property(dpy, win, "_NET_WM_NAME", app_name);
				if (app_name[0]) {
					delete_annotation(app_name, data);
				}
			}
			break;

		case AC_CMD_SHOW_INPUT:
		case AC_CMD_HIDE_INPUT:

		case AC_CMD_SHOW_ANNOTATE:
		case AC_CMD_HIDE_ANNOTATE:

		case AC_CMD_SHOW_EMAIL:
		case AC_CMD_HIDE_EMAIL:

		case AC_CMD_SHOW_PRINTER:
		case AC_CMD_HIDE_PRINTER:

		case AC_CMD_SHOW_PAGE_NAV:
		case AC_CMD_HIDE_PAGE_NAV:

		case AC_CMD_ENABLE_PAGE_NEXT:
		case AC_CMD_DISABLE_PAGE_NEXT:

		case AC_CMD_ENABLE_PAGE_PREV:
		case AC_CMD_DISABLE_PAGE_PREV:

		case AC_CMD_SHOW_RETRACTBAR:
		case AC_CMD_HIDE_RETRACTBAR:

		case AC_CMD_SHOW_STATUS:
		case AC_CMD_HIDE_STATUS:

		case AC_CMD_SHOW_KEYBOARD:
		case AC_CMD_HIDE_KEYBOARD:

		case AC_CMD_SWAP_CLOSE_UP:
		case AC_CMD_SHOW_CLOSE:
		case AC_CMD_SHOW_UP:

		case AC_CMD_PRINTER_STATUS:

		case AC_CMD_SHOW_SPEAKER:
		case AC_CMD_HIDE_SPEAKER:

		case AC_CMD_SHOW_VOICE:
		case AC_CMD_HIDE_VOICE:

		case AC_CMD_APP_SPEAKS:
		case AC_CMD_AC_SPEAKS:

		case AC_CMD_SHOW_TICK:

		case AC_CMD_SHOW_HELP:
		case AC_CMD_HIDE_HELP:

		case AC_CMD_GRAB_ON:
		case AC_CMD_GRAB_OFF:
			break;

		default:
			printf("event not handled \n");
			break;
	}


	memset(&ev, '\0', sizeof ev);
	ev.xclient.type = ClientMessage;
	ev.xclient.window = win;
	ev.xclient.message_type = atom_array[2];
	ev.xclient.format = 8;

	ev.xclient.data.l[0] = cmd_id;
	ev.xclient.data.l[1] = win;

	XSendEvent(dpy, DefaultRootWindow(dpy), False, SubstructureRedirectMask|SubstructureNotifyMask, &ev);
	XFlush(dpy);

	if (cmd_id == AC_CMD_LAUNCH_APP && win != DefaultRootWindow(dpy)) {
		XEvent rex;
		/* Wait for alchemy to read the atom */
		XIfEvent(dpy, &rex, is_client_message, (char *)win);
	}
}

Bool is_client_message(Display *dpy, XEvent *xev, char *win)
{
	if(xev->type == ClientMessage && 
			xev->xclient.window == DefaultRootWindow(dpy) &&
			xev->xclient.message_type == atom_array[2] &&
			xev->xclient.format == 8 &&
			xev->xclient.data.l[0] == AC_CMD_LAUNCH_APP)
		return True;
	return False;
}

void ewmh_command(char *cmd_atom_str, int sub_cmd, Display *dpy)
{
	ewmh_command_win(cmd_atom_str, sub_cmd, dpy, DefaultRootWindow(dpy));
}

void ewmh_command_win(char *cmd_atom_str, int sub_cmd, Display *dpy, Window w)
{
	XEvent	ev;
	Window	root;
	Atom cmd_prop, cmd_atom;

	root = DefaultRootWindow(dpy);

	cmd_prop = XInternAtom(dpy, "_NET_WM_STATE", False);

	if (cmd_prop == None) {
		fprintf(stderr, "No such property 'WINDOW_STATE'\n");
		return;
	}

	cmd_atom = XInternAtom(dpy, cmd_atom_str, False);
	if (cmd_atom == None) {
		fprintf(stderr, "No such property '%s'\n", cmd_atom_str);
		return;
	}
	      
	memset(&ev, '\0', sizeof ev);
	ev.xclient.type = ClientMessage;
	ev.xclient.window = root; 	/* we send it _from_ root as we have no win  */
	ev.xclient.message_type = cmd_prop;
	ev.xclient.format = 8;

	ev.xclient.data.l[0] = cmd_atom;
	ev.xclient.data.l[1] = w;
	ev.xclient.data.l[2] = sub_cmd;

	XSendEvent(dpy, root, False, SubstructureRedirectMask|SubstructureNotifyMask, &ev);
}

