/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

/* Copyright 2003 PicoPeta Simputers Pvt. Ltd.  */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "alchemy.h"
#include "acctrl.h"

static void xwindow_get_text_property(Display *dpy, Window win, const char *atom_name, unsigned char *string_ret)
{
	unsigned char *prop_ret;
	Atom at, actual_type;
	int status, actual_format;
	unsigned long n,b;

	at=XInternAtom(dpy, atom_name, False);

	status = XGetWindowProperty(dpy, win, at, 0,
			MAX_PROPERTY_LENGTH, False, AnyPropertyType, &actual_type,
			&actual_format, &n, &b, &prop_ret);

	if( (status!=Success) || (!n) || (!prop_ret) )
		string_ret[0] = 0;
	else {
		strcpy(string_ret, prop_ret);
		XFree(prop_ret);
	}
}

static void
usage(char *progname)
{
	printf("Usage: %s <alchemy command> [command options if any]\n", progname);
	puts("Options:");
	puts("  -F       <flags>  set application flags"); 
	puts("  -L       <appid> <app-name> <file-name>  Launch application"); 
	puts("  -T       <title bar font>  set title bar font"); 
	puts("  -s       <status bar message>  set status bar message"); 
	puts("  -S       <status bar font>  set status bar font"); 
	puts("  -P       <page number>  set page number"); 
	puts("  -k       <category> set keyboard category");
	puts("  -l       <layout> set keyboard layout");
#ifdef __arm__
	puts("  -o       <orientation> set application's rotation orientation");
#endif
	puts("  -A       <annotate-key> set annotate key");
	puts("  -H       <help-key> set help key");
	puts("  -f       <file-name>  set emailable file name"); 
	puts("  -D       <annotate-key>  Delete the specified annotation key"); 
	puts("  -n       Switch to next application window");
	puts("  -p       Switch to previous application window");

	puts("  -h       <on|off>  show/hide help icon");
	puts("  -i       <on|off>  show/hide Input method/chooser"); 
	puts("  -a       <on|off>  show/hide annotate icon"); 
	puts("  -e       <on|off>  show/hide email icon"); 
	puts("  -r       <on|off>  show/hide printer icon"); 
	puts("  -g       <on|off>  show/hide page navigation"); 
	puts("  -G       <pen|pdis|nen|ndis>  Enable/Disable prev/next page buttons"); 
	puts("  -b       <on|off>  show/hide retractable bar launcher"); 
	puts("  -K       <on|off>  show/hide keyboard"); 
	puts("  -y       <on|off>  show/hide status message"); 
	puts("  -R       <on|off>  show/hide speaker icon"); 
	puts("  -v       <on|off>  show/hide voice icon"); 
	puts("  -w       <swap|up|close|tick> swap icons/show up/show close/show tick"); 
	puts("  -B       <on|off> enable/disable background grabbing for annotation");
	puts("  -C       Check Printer status");

	puts("  -t       <theme>  Set theme to be used by alchemy"); 
	puts("  -E       Print current alchemy theme to stdout");
	puts("  -I       <windowid> Get window appid");
	puts("  -u       Toggle full screen mode");
	puts("  -z       <on|off> Turn on/off everything");
	puts("  -q       Request alchemy to quit");
	puts("  -?       this help");
	exit(1);
}

int main(int argc, char* argv[])
{
	Window dest_window;
	Display *dpy;
	char *display_name = (char *)getenv("DISPLAY");
	int i;
	Atom at, actual_type;
	int status, actual_format;
	int *property_value=0;
	unsigned long n,b;

	if(!display_name)
	display_name=strdup(":0.0");

	dpy = XOpenDisplay(display_name);
	if (dpy == NULL) {
		printf("Cant connect to display: %s\n", display_name);
		exit(1);
	}

	at = XInternAtom(dpy, "_NET_ACTIVE_WINDOW", False);
	status = XGetWindowProperty(dpy, DefaultRootWindow(dpy), at, 0,
			MAX_PROPERTY_LENGTH, False, AnyPropertyType, &actual_type,
			&actual_format, &n, &b, (unsigned char **)&property_value);
	if( (status!=Success) || (!n) || (!property_value) )
		exit(1);
	
	dest_window = property_value[0];
	XFree(property_value);

	/* pass command line */
	for (i=1; argv[i]; i++) {
		char *arg = argv[i];
		if (*arg=='-') {
			switch (arg[1]) {
				case 't' :
					if (argv[i+1] != NULL)
						ac_control_win(AC_CMD_SET_THEME, argv[i+1], dpy, dest_window);
					i++;
					break;

				case 'L' :
					if (argv[i+1] != NULL && argv[i+2] != NULL && argv[i+3] != NULL) {
						char app[512];
						sprintf(app,"%s,%s,%s",argv[i+1],argv[i+2],argv[i+3]);
						// For commandline only-
						// use root window
						ac_control_win(AC_CMD_LAUNCH_APP, app, dpy, DefaultRootWindow(dpy));
					}
					i+=2;
					break;

#ifdef __arm__
				case 'o' :
					if (argv[i+1] != NULL)
						ac_control_win(AC_CMD_SET_ROTATION, argv[i+1], dpy, dest_window);
					i++;
					break;
#endif

				case 'n':
					ac_control_win(AC_CMD_NEXT, NULL, dpy, dest_window);
					break;

				case 'p':
					ac_control_win(AC_CMD_PREV, NULL, dpy, dest_window);
					break;
					
				case 'q':
					ac_control_win(AC_CMD_EXIT, NULL, dpy, dest_window);
					break;
				
				case 'F' :
					ac_control_win(AC_CMD_SET_FLAGS, argv[i+1], dpy, dest_window);
					break;

				case 'T' :
					if (argv[i+1] != NULL)
						ac_control_win(AC_CMD_SET_TITLE_FONT, argv[i+1], dpy, dest_window);
					i++;
					break;

				case 's' :
					if (argv[i+1] != NULL)
						ac_control_win(AC_CMD_SET_STATUS_MSG, argv[i+1], dpy, dest_window);
					i++;
					break;
					
				case 'S' :
					if (argv[i+1] != NULL)
						ac_control_win(AC_CMD_SET_STATUS_FONT, argv[i+1], dpy, dest_window);
					i++;
					break;

				case 'P' :
					if (argv[i+1] != NULL)
						ac_control_win(AC_CMD_SET_PAGE_NUM, argv[i+1], dpy, dest_window);
					i++;
					break;

				case 'k':
					ac_control_win(AC_CMD_SET_KEYBOARD_CATEGORY, argv[i+1], dpy, dest_window);
					break;

				case 'l':
					ac_control_win(AC_CMD_SET_KEYBOARD_LAYOUT, argv[i+1], dpy, dest_window);
					break;

				case 'A':
					ac_control_win(AC_CMD_SET_ANNOTATE_KEY, argv[i+1], dpy, dest_window);
					break;

				case 'H':
					ac_control_win(AC_CMD_SET_HELP_KEY, argv[i+1], dpy, dest_window);
					break;

					
				case 'f' :
					if (argv[i+1] != NULL)
						ac_control_win(AC_CMD_SET_EMAIL_FILE, argv[i+1], dpy, dest_window);
					i++;
					break;

				case 'D' :
					if (argv[i+1] != NULL)
						ac_control_win(AC_CMD_DEL_ANNOTATION, argv[i+1], dpy, dest_window);
					i++;
					break;
				
				case 'h' :
					if (argv[i+1] != NULL) {
						if(!strcmp(argv[i+1],"on"))
							ac_control_win(AC_CMD_SHOW_HELP, NULL, dpy, dest_window);
						else if(!strcmp(argv[i+1],"off"))
							ac_control_win(AC_CMD_HIDE_HELP, NULL, dpy, dest_window);
					}
					i++;
					break;

				case 'i' :
					if (argv[i+1] != NULL) {
						if(!strcmp(argv[i+1],"on"))
							ac_control_win(AC_CMD_SHOW_INPUT, NULL, dpy, dest_window);
						else if(!strcmp(argv[i+1],"off"))
							ac_control_win(AC_CMD_HIDE_INPUT, NULL, dpy, dest_window);
					}
					i++;
					break;

				case 'a' :
				if (argv[i+1] != NULL) {
					if(!strcmp(argv[i+1],"on"))
						ac_control_win(AC_CMD_SHOW_ANNOTATE, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"off"))
						ac_control_win(AC_CMD_HIDE_ANNOTATE, NULL, dpy, dest_window);
					}
					i++;
					break;
				
				case 'e' :
					if (argv[i+1] != NULL) {
						if(!strcmp(argv[i+1],"on"))
							ac_control_win(AC_CMD_SHOW_EMAIL, NULL, dpy, dest_window);
						else if(!strcmp(argv[i+1],"off"))
							ac_control_win(AC_CMD_HIDE_EMAIL, NULL, dpy, dest_window);
					}
					i++;
					break;

				case 'r' :
					if (argv[i+1] != NULL) {
						if(!strcmp(argv[i+1],"on"))
							ac_control_win(AC_CMD_SHOW_PRINTER, NULL, dpy, dest_window);
						else if(!strcmp(argv[i+1],"off"))
							ac_control_win(AC_CMD_HIDE_PRINTER, NULL, dpy, dest_window);
					}
					i++;
					break;

				case 'B' :
					if (argv[i+1] != NULL) {
						if(!strcmp(argv[i+1],"on"))
							ac_control_win(AC_CMD_GRAB_ON, NULL, dpy, dest_window);
						else if(!strcmp(argv[i+1],"off"))
							ac_control_win(AC_CMD_GRAB_OFF, NULL, dpy, dest_window);
					}
					i++;
					break;

				case 'g' :
					if (argv[i+1] != NULL) {
					if(!strcmp(argv[i+1],"on"))
						ac_control_win(AC_CMD_SHOW_PAGE_NAV, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"off"))
						ac_control_win(AC_CMD_HIDE_PAGE_NAV, NULL, dpy, dest_window);
					}
					i++;
				break;

				case 'G' :
					if (argv[i+1] != NULL) {
					if(!strcmp(argv[i+1],"pen"))
						ac_control_win(AC_CMD_ENABLE_PAGE_PREV, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"pdis"))
						ac_control_win(AC_CMD_DISABLE_PAGE_PREV, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"nen"))
						ac_control_win(AC_CMD_ENABLE_PAGE_NEXT, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"ndis"))
						ac_control_win(AC_CMD_DISABLE_PAGE_NEXT, NULL, dpy, dest_window);
					}
					i++;
				break;

				case 'j' :
					if (argv[i+1] != NULL)
						ac_control_win(AC_CMD_SET_STATUS_JUSTIFY, argv[i+1], dpy, dest_window);
					i++;
					break;

				case 'b' :
					if (argv[i+1] != NULL) {
						if(!strcmp(argv[i+1],"on"))
							ac_control_win(AC_CMD_SHOW_RETRACTBAR, NULL, dpy, dest_window);
						else if(!strcmp(argv[i+1],"off"))
							ac_control_win(AC_CMD_HIDE_RETRACTBAR, NULL, dpy, dest_window);
					}
					i++;
					break;

				case 'K' :
					if(!argv[i+1])
						break;
					if(!strcmp(argv[i+1], "on"))
						ac_control_win(AC_CMD_SHOW_KEYBOARD, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1], "off"))
						ac_control_win(AC_CMD_HIDE_KEYBOARD, NULL, dpy, dest_window);
					break;

				case 'y' :
				if (argv[i+1] != NULL) {
					if(!strcmp(argv[i+1],"on"))
						ac_control_win(AC_CMD_SHOW_STATUS, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"off"))
						ac_control_win(AC_CMD_HIDE_STATUS, NULL, dpy, dest_window);
				}
				i++;
				break;

				case 'R' :
				if (argv[i+1] != NULL) {
					if(!strcmp(argv[i+1],"on"))
						ac_control_win(AC_CMD_SHOW_SPEAKER, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"off"))
						ac_control_win(AC_CMD_HIDE_SPEAKER, NULL, dpy, dest_window);
				}
				i++;
				break;

				case 'v' :
				if (argv[i+1] != NULL) {
					if(!strcmp(argv[i+1],"on"))
						ac_control_win(AC_CMD_SHOW_VOICE, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"off"))
						ac_control_win(AC_CMD_HIDE_VOICE, NULL, dpy, dest_window);
				}
				i++;
				break;

				case 'w' :
				if (argv[i+1] != NULL) {
					if(!strcmp(argv[i+1],"swap"))
						ac_control_win(AC_CMD_SWAP_CLOSE_UP, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"up"))
						ac_control_win(AC_CMD_SHOW_UP, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"close"))
						ac_control_win(AC_CMD_SHOW_CLOSE, NULL, dpy, dest_window);
					else if(!strcmp(argv[i+1],"tick"))
						ac_control_win(AC_CMD_SHOW_TICK, NULL, dpy, dest_window);
				}
				i++;
				break;

				case 'C':
					ac_control_win(AC_CMD_PRINTER_STATUS, NULL, dpy, dest_window);
					break;

				case 'E':
					{
						char data[MAX_PROPERTY_LENGTH];

						xwindow_get_text_property(dpy, DefaultRootWindow(dpy),
							"_AC_WM_THEME", data);
						printf("Theme=%s\n", data);
					}
					
					i++;
					break;

				case 'X' :
					{
						char data[MAX_PROPERTY_LENGTH];

						xwindow_get_text_property(dpy, DefaultRootWindow(dpy),
							"_AC_WM_PID_MAP", data);
						printf("Appid 2 Pid Map::%s\n", data);
						xwindow_get_text_property(dpy, DefaultRootWindow(dpy),
							"_AC_WM_LAYER_MAP", data);
						printf("Layer Map::%s\n", data);
					}
					
					i++;
					break;

				case 'I' :
					printf("%lu\n",getWindowAppid(strtoul(argv[i+1], NULL, 16), dpy));
					i++;
					break;

				case 'u' :
					ewmh_command("_NET_WM_STATE_FULLSCREEN", _NET_WM_STATE_TOGGLE, dpy);
					break;

				case 'z' :
					if (argv[i+1] != NULL) {
						if(!strcmp(argv[i+1],"on")) {
							ac_control_win(AC_CMD_SHOW_ANNOTATE, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_SHOW_EMAIL, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_SHOW_PAGE_NAV, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_SHOW_INPUT, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_SHOW_RETRACTBAR, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_SHOW_SPEAKER, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_SHOW_PRINTER, NULL, dpy, dest_window);
						} else if(!strcmp(argv[i+1],"off")) {
							ac_control_win(AC_CMD_HIDE_ANNOTATE, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_HIDE_EMAIL, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_HIDE_PAGE_NAV, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_HIDE_INPUT, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_HIDE_RETRACTBAR, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_HIDE_SPEAKER, NULL, dpy, dest_window);
							ac_control_win(AC_CMD_HIDE_PRINTER, NULL, dpy, dest_window);
						}
					}
					i++;
				break;

				default:
					usage(argv[0]);
				break;
			}
		}
	}
	XSync(dpy, True);
	XCloseDisplay(dpy);

	return 0;
}
