/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include <glob.h>
#include <string.h>
#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "utils.h"
#include "annotation_glue.h"

// Escape characters like space and forward slash
// which are not valid annotation keys
char *string_escape_annotation(const char *annotation_key)
{
	Bool changed=False;
	char *ret_annotation_key=NULL;

	if(annotation_key) {
		int i=strlen(annotation_key),j,k;

		ret_annotation_key=(char *)malloc(2*i);
		if(!ret_annotation_key) {
			err_exit("%s(): Cannot allocate memory", __func__);
			exit(-1);
		}

		for(j=0,k=0;j<=i;j++) {
			switch(annotation_key[j]) {
				case ' ':
				case '(':
				case ')':
				case '{':
				case '}':
				case '\\':
				case '*':
				case '?':
				case ';':
				case '\"':
				case '\'':
					ret_annotation_key[k++]='\\';
					ret_annotation_key[k++]=annotation_key[j];
					changed=True;
					break;

				case '/':
					ret_annotation_key[k++]='#';
					changed=True;
					break;

				default:
					ret_annotation_key[k++]=annotation_key[j];
					break;
			}
		}
	}

	if(!changed) {
		free(ret_annotation_key);
		return NULL;
	}

	return ret_annotation_key;
}

char *get_annotation_prefix(const char *app_name, const char *annotation_key, Bool escaped)
{
	char *annotation_path=NULL;
	const char *escaped_key;

	if (!escaped || !(escaped_key=string_escape_annotation(annotation_key))) {
		escaped_key = annotation_key;
	}

	annotation_path = (char *)malloc(strlen(ANNOTATION_TEMPLATE) + strlen(app_name) + strlen(escaped_key) + 2); // allocate one more for * at end
	if (!annotation_path)
		err_exit("%s(): Cannot allocate memory", __func__);
	sprintf(annotation_path, ANNOTATION_TEMPLATE, app_name, escaped_key);

	if (escaped_key != annotation_key)
		free((void *)escaped_key);

	return annotation_path;
}

char *get_annotation_wild_suffix(const char *app_name, const char *annotation_key, Bool escaped)
{
	char *annotation_path=NULL;
	const char *escaped_key;
	char *wild_escaped_key;

	if (!escaped || !(escaped_key=string_escape_annotation(annotation_key))) {
		escaped_key = annotation_key;
	}

	wild_escaped_key = (char *)malloc(strlen(escaped_key)+3);
	sprintf(wild_escaped_key,"*.%s",escaped_key);
	if (!wild_escaped_key)
		err_exit("%s(): Cannot allocate memory", __func__);

	annotation_path = (char *)malloc(strlen(ANNOTATION_TEMPLATE) + strlen(app_name) + strlen(wild_escaped_key) + 1); // allocate one more for * at end
	if (!annotation_path)
		err_exit("%s(): Cannot allocate memory", __func__);
	sprintf(annotation_path, ANNOTATION_TEMPLATE, app_name, wild_escaped_key);

	if (escaped_key != annotation_key)
		free((void *)escaped_key);
	free((void *)wild_escaped_key);

	return annotation_path;
}

void delete_annotation(const char *app_name, const char *annotation_key)
{
	int i;
	glob_t tglob;
	char *annotation_wildcard;

	annotation_wildcard=get_annotation_prefix(app_name, annotation_key, False);
	strcat(annotation_wildcard,"*");

	tglob.gl_offs=0;

	// Search for all files with same prefix
	if(glob(annotation_wildcard, 0, NULL, &tglob)) {
		debug_info("%s(): glob failed", __func__);
		globfree(&tglob);
		free(annotation_wildcard);
		return;
	}

	// ... and vaporize them
	for(i=0;i<tglob.gl_pathc;i++) {
		unlink(tglob.gl_pathv[i]);
	}

	globfree(&tglob);
	free(annotation_wildcard);

	annotation_wildcard=get_annotation_wild_suffix(app_name, annotation_key, False);

	tglob.gl_offs=0;

	// Search for all files with same prefix
	if(glob(annotation_wildcard, 0, NULL, &tglob)) {
		debug_info("%s(): glob failed", __func__);
		globfree(&tglob);
		free(annotation_wildcard);
		return;
	}

	// ... and vaporize them
	for(i=0;i<tglob.gl_pathc;i++) {
		unlink(tglob.gl_pathv[i]);
	}

	globfree(&tglob);
	free(annotation_wildcard);
}
