/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include <X11/Xlib.h>
#include <X11/Xproto.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <glob.h>
#include <math.h>
#include <time.h>

#include "alchemy.h"
#include "utils.h"
#include "grab.h"
#include "basic_window.h"
#include "ichoose_window.h"
#include "star_window.h"
#include "winlist_window.h"
#include "keyboard_glue.h"

static unsigned int nextpage_pressed = False;
static unsigned int prevpage_pressed = False;
static double nextpage_press_time = 0;
static double prevpage_press_time = 0;

// Life begins for a fresh new window
window *new_basic_window(wm *wm, Window map_window, window_type wtype)
{
	window *win,*lw = NULL;
	int tval;
	unsigned char tsval[MAX_PROPERTY_LENGTH];

	win = (window *)malloc(sizeof(window));
	if(!win)
		err_exit("Could not allocate window");

	win->wtype = wtype;
	win->app_window = map_window;
	win->next=win->prev=win->layer_above=win->layer_below=NULL;
	win->transient_of=NULL;
	win->transient_next=win->transient_prev=NULL;
	win->frame_widgets = NULL;
	win->wm = wm;
	win->x = win->y = win->height = win->width = 0;
	win->tfont = NULL;
	win->status_msg = NULL;
	win->sfont = NULL;
	win->page_num = NULL;
	win->pfont = NULL;
	win->annotation_key = NULL;
	win->help_key = strdup("default");
	win->is_mapped = False;
	win->fast_load = False;
	win->fast_visible = False;
	win->needs_grab = True;
	win->is_chikki = False;
	win->tjustify = JUSTIFY_CENTER;
	win->pjustify = JUSTIFY_RIGHT;
	win->sjustify = JUSTIFY_RIGHT;

	win->ignore_next_unmap = False;

	win->topbar_frame = win->bottombar_frame = NULL;
	win->keyboard_layout = strdup("qwerty");
	win->keyboard_category = strdup("english");
	win->onetap = False;
	win->app_speaks = False;

	XDefineCursor(wm->display, win->app_window, wm->normal_cursor);

	tval = window_get_int_property(win, _AC_W_APPID);
	if(tval < 0)
		win->appid = 0; // Set to 0 if not specified
	else
		win->appid = tval;

	tval = window_get_int_property(win, _AC_W_LAUNCHER);
	if(tval < 0)
		win->launcher_appid = AC_LAUNCHER_APPID; // Default is launcher
	else
		win->launcher_appid = tval;

	tval = window_get_int_property(win, _AC_W_FAST_LOAD);
	if(tval == 1) {
		win->fast_load = True;
		win->launcher_appid = AC_LAUNCHER_APPID;
	} else if(tval == 2) {
		win->fast_load = True;
		win->fast_visible = True;
		win->launcher_appid = AC_LAUNCHER_APPID;
	}

	tval = window_get_int_property(win, _AC_W_CHIKKI);
	if(tval == 1) {
		win->is_chikki = True;
	}

#ifdef __arm__
	tsval[0]=0;
	window_get_text_property(win, _AC_W_ROTATION_STATE, tsval);
	if(tsval[0] != 0) {
		int rot = tsval[0] - '0';
		if(rot >=0 && rot <= 3)
			win->rotation = rot;
		else
			win->rotation = wm->current_rot; // Set to wm default
							 // if invalid
	} else
		win->rotation = wm->current_rot; // Set to wm default 
						 // if not specified

#endif

	tval = window_get_int_property(win, _NET_WM_PID);
	if(tval < 0)
		win->pid = 0; // Set to 0 if not specified
	else
		win->pid = tval;

	win->oflags.all_flags = 0;
	win->flags.all_flags = 0;
	win->flags.bits.status = 1;
	win->flags.bits.status_msg = 1;
	win->flags.bits.star = 1;
	win->flags.bits.launcher = 1;
	win->flags.bits.needs_annotation = 1;
	win->flags.bits.close = 1;
	win->flags.bits.status_msg = 1;
	win->flags.bits.help = 1;
	win->flags.bits.next_page_enabled = 1;
	win->flags.bits.prev_page_enabled = 1;

	if(win->appid == AC_KEYBOARD_APPID) {
		wm->keyboard_window = win;
	} else if(win->appid == AC_STATUS_APPID) {
		wm->status_window = win;
		win->flags.bits.status = 0;
		win->flags.bits.launcher = 0;
	} else if(win->appid == AC_XSTROKE_APPID) {
		wm->xstroke_window = win;
	} else if(win->appid == AC_CLOCK_APPID) {
		wm->clock_window = win;
	}

	if(win->appid == AC_IDENTITY_APPID) {
		win->flags.all_flags = 0;
		window_set_int_property(win, _AC_W_FLAGS, 0 );
		wm->identity_save_window = wm->mapped_window;
	} else {
		tval = window_get_int_property(win, _AC_W_FLAGS);
		if(tval >= 0)
			win->flags.all_flags = tval;
		else
			window_set_int_property(win, _AC_W_FLAGS, win->flags.all_flags );
	}

	win->name =  NULL;
	tsval[0]=0;
	window_get_text_property(win, _NET_WM_NAME, tsval);
	if(tsval[0]!=0)
		win->name = strdup(tsval);

	if(!win->name) {
		XClassHint xch;
		// Try WM_CLASS  instead ...
		if(XGetClassHint(wm->display, win->app_window, &xch)) {
			win->name = strdup(xch.res_name);
			XFree(xch.res_name);
			XFree(xch.res_class);
		} else {
			// If nothing is possible, use default :)
			win->name = strdup("default");
		}
	}
	window_set_text_property(win, _NET_WM_NAME, win->name);

	win->title = NULL;
	tsval[0]=0;
	window_get_text_property(win, _NET_WM_VISIBLE_NAME, tsval);
	if(tsval[0]!=0)
		win->title = strdup(tsval);

	if(!win->title) {
		// Try WM_NAME  instead ...
		XFetchName(wm->display, win->app_window, &win->title);
	}

	// TODO : handle duplicate window names ??

	if(!win->title && win->name) {
		// Finally just settle on the name itself
		win->title = strdup(win->name);
		XStoreName(wm->display, win->app_window, win->title);
	}

	window_get_text_property(win, _AC_W_STATUS_MSG, tsval);
	if(tsval[0]!=0)
		win->status_msg = strdup(tsval);

	window_get_text_property(win, _AC_W_PAGE_NUM, tsval);
	if(tsval[0]!=0)
		win->page_num = strdup(tsval);

	window_get_text_property(win, _AC_W_ANNOTATE_KEY, tsval);
	if(tsval[0]!=0)
		win->annotation_key = strdup(tsval);

	tval = window_get_int_property(win, _AC_W_LAYER_WINDOWID);
	if(tval > 0) {
		// Aha : This window has a layer windowid
		// This means that I should attach as a layer
		// above the current window whose windowid is windowid

		lw = window_find_windowid(wm, (Window) tval);
	}

	if (win->appid == AC_CORONA_HELP_APPID && wm->mapped_window) {
		lw = wm->mapped_window;
		win->appid = 0;

		if (wm->mapped_window->name) {
			if(win->name)
				free(win->name);
			win->name=strdup(wm->mapped_window->name);
			window_set_text_property(win, _NET_WM_NAME, win->name);
		}
	}
		
	if(lw) {
			// Just in case other layers have come in-between
			while(lw->layer_above)
				lw=lw->layer_above;

			// Copy pid to this layer if it is unset
			// An example for this is Tk Dialog boxes
			if(win->pid == 0) {
				if(lw->pid) {
					win->pid = lw->pid;
					window_set_int_property(win, _NET_WM_PID, win->pid);
				}
			}

			// If there are layers above the launcher, they should
			// be treated different from the launcher
			if( win->appid == AC_LAUNCHER_APPID || (win->appid==0 && lw->appid == AC_LAUNCHER_APPID)) {
				win->flags.all_flags = 0;
				win->flags.bits.status_msg = 1;
				win->flags.bits.up = 1;
				win->flags.bits.star = 1;
				win->flags.bits.next_page_enabled = 1;
				win->flags.bits.prev_page_enabled = 1;
				window_set_int_property(win, _AC_W_FLAGS, win->flags.all_flags );
				win->appid = AC_LAUNCHER_APPID;
				window_set_int_property(win, _AC_W_APPID, win->appid);
			} else if(win->appid == 0) {
				if(lw->appid) {
					win->appid = lw->appid;
					window_set_int_property(win, _AC_W_APPID, win->appid);
				}
			}

#ifdef __arm__
			// Start in the same mode as the others
			win->rotation = lw->rotation;
#endif

			// Layered, push the layered window below this
			if((lw->next == lw) && (lw->prev == lw)) {
				// If lw was the sole window till we came along..
				win->next = win->prev = win;
			} else {
				win->next = lw->next;
				win->prev = lw->prev;
				win->prev->next = win;
				win->next->prev = win;
			}

			// Make layer links
			win->layer_below = lw;
			lw->layer_above = win;

			// null unnecessary links
			lw->next = lw->prev = NULL;

			// for circular list head; repoint
			if(wm->window_list == lw)
				wm->window_list = win;
	} else {
		// No layers: add to circular list

		if(win->appid == AC_LAUNCHER_APPID) {
			win->flags.all_flags = 0;
			win->flags.bits.star = 1;
			win->flags.bits.status = 1;
			win->flags.bits.left = 1;
			win->flags.bits.right = 1;
			win->flags.bits.next_page_enabled = 1;
			win->flags.bits.prev_page_enabled = 1;
			win->flags.bits.help = 1;
			window_set_int_property(win, _AC_W_FLAGS, win->flags.all_flags );

			wm->launcher_window = win;
		}
		if(wm->window_list) {
			// Adding to the circular list done here
			win->next = wm->window_list->next;
			win->prev = wm->window_list;
			win->next->prev = win;
			win->prev->next = win;
		} else {
			// I am the circular list
			wm->window_list = win;
			win->next = win->prev = win;
		}
	}

	wm->num_windows ++;
	if(wtype == NORMAL_WIN || wtype == TOOLBAR_WIN)
		ac_make_pid_appid_map(wm);

	return win;
}

// Set window to Normal, Iconic or Withdrawn states
unsigned long window_get_state(window *win)
{
	return window_get_int_property(win, WM_STATE);
}

// Set window to Normal, Iconic or Withdrawn states
void window_set_state(window *win, unsigned long state)
{
	CARD32 data[2];
	wm *wm=win->wm;

	data[0] = state;
	data[1] = None;

	XChangeProperty(wm->display, win->app_window,
			wm->atomlist[WM_STATE], wm->atomlist[WM_STATE],
			32, PropModeReplace, (unsigned char *)data, 2);
}

// Set the focus to this window if it requires focus
void window_handle_focus(window *win)
{
	XWMHints *xwmhints;
	Display *dpy;
	wm *wm=win->wm;
	
	dpy=wm->display;

	xwmhints=XGetWMHints(dpy, win->app_window);
	if(xwmhints) {
		if(xwmhints -> input) {
			XSetInputFocus(dpy, win->app_window, RevertToPointerRoot,
					CurrentTime);
			wm->focused_window = win;
		}
		XFree(xwmhints);
	} else {
			XSetInputFocus(dpy, win->app_window, 
					RevertToPointerRoot, CurrentTime);
			wm->focused_window = win;
	}
}

// Get the next normal window which can be mapped
window *window_get_next_mappable(window *w)
{
	window *tw;

	for(tw=w->next; tw!=w; tw=tw->next)
		if(tw->wtype == NORMAL_WIN)
			break;

	return tw;
}

// Get the previous normal window which can be mapped
window *window_get_prev_mappable(window *w)
{
	window *tw;

	for(tw=w->prev; tw!=w; tw=tw->prev)
		if(tw->wtype == NORMAL_WIN)
			break;

	return tw;
}

#ifdef __arm__
// Rotate the window to its required orientation
void window_handle_rotation(window *win)
{
	if(win->wtype == NORMAL_WIN) {
		wm *wm = win->wm;

		if(win->rotation != wm->current_rot)
			wm->rotate_to(wm, win->rotation);
	}
}
#endif


// Change to current window's keyboard layout and category
// Show/hide the keyboard as requested by the client
void window_handle_keyboard(window *win)
{
	wm *wm = win->wm;

	if(win->flags.bits.needs_keyboard && (!win->flags.bits.full_screen)) {
		if(wm->keyboard_window && !wm->is_keyboard_mapped)
			wm->keyboard_window->show(wm->keyboard_window);
		else if (!wm->keyboard_window) {
			keyboard_start_keyboard(win->keyboard_category,win->keyboard_layout);
		}
	} else {
		/* hide keyboard if currently being shown */
		if(wm->keyboard_window && wm->is_keyboard_mapped)
			wm->keyboard_window->hide(wm->keyboard_window);
	}

	if(wm->keyboard_window) {
		if(strcmp(win->keyboard_layout,"xstroke")) {
			// Ok do keyboard here, no xstroke
			if(keyboard_switch_category_layout(win->keyboard_category,
					win->keyboard_layout, wm->current_kbd_category, 
					wm->current_kbd_layout)) {
				free(wm->current_kbd_category);
				free(wm->current_kbd_layout);
				wm->current_kbd_category = strdup(win->keyboard_category);
				wm->current_kbd_layout = strdup(win->keyboard_layout);
			}
		} else {
			// Handle xstroke here
			if(strcmp(wm->current_kbd_layout,win->keyboard_layout)) {
				free(wm->current_kbd_layout);
				wm->current_kbd_layout = strdup("xstroke");
			}
		}
	}
}

// Window's life is over .. release its resources
void window_destroy(window *win)
{
	wm *wm = win->wm;

	// Free all allocated memory
	if(win->name)
		free(win->name);
	if(win->title)
		free(win->title);
	if(win->tfont)
		free(win->tfont);
	if(win->status_msg)
		free(win->status_msg);
	if(win->sfont)
		free(win->sfont);
	if(win->page_num)
		free(win->page_num);
	if(win->pfont)
		free(win->pfont);
	if(win->annotation_key)
		free(win->annotation_key);
	if(win->help_key)
		free(win->help_key);
	if(win->keyboard_layout)
		free(win->keyboard_layout);
	if(win->keyboard_category)
		free(win->keyboard_category);

	wm->current_theme->cleanup_theme(wm->current_theme, win);

	if(win->appid == AC_LAUNCHER_APPID && !win->layer_below) {
		wm->launcher_window = NULL;
	} else if(win->appid == AC_KEYBOARD_APPID) {
		wm->keyboard_window = NULL;
	} else if(win->appid == AC_STATUS_APPID) {
		wm->status_window = NULL;
	}

	if(win->layer_below || win->layer_above) {

		// Fix the doubly linked layer list
		if(win->layer_below) {
			win->layer_below->layer_above = win->layer_above;
		}

		if(win->layer_above) {
			win->layer_above->layer_below = win->layer_below;
		} else if((win->next == win) && (win->prev == win)) {
			// This is the only window left (Though it may have other layers
			// below)
			win->layer_below->next = win->layer_below; 
			win->layer_below->prev = win->layer_below; 
			wm->window_list = win->layer_below;
		} else {
			// Raise the layer below and link that to circular list
			win->prev->next = win->layer_below;
			win->next->prev = win->layer_below;
			win->layer_below->next = win->next;
			win->layer_below->prev = win->prev;
			if(wm->window_list == win)
				wm->window_list = win->layer_below;
		}
	} else {
		// Remove from circular list
		if((win->next == win) && (win->prev == win)) {
			wm->window_list = wm->mapped_window = NULL;
			wm->focused_window = NULL;
		} else {
			win->prev->next = win->next;
			win->next->prev = win->prev;
			if(wm->window_list == win)
				wm->window_list = win->next;
		}
		if(win == wm->mapped_window)
			wm->mapped_window = NULL;
	}

	XDestroyWindow(wm->display, win->frame_window);

	wm->num_windows --;

	if(win->wtype == NORMAL_WIN || win->wtype == TOOLBAR_WIN)
		ac_make_pid_appid_map(wm);

	free(win);
}

// Look at annotate/annotate_there icon visibility
void window_handle_annotate_icon(window *win)
{
	char *annotation_path;
	glob_t tglob;

	if(!win->flags.bits.needs_annotation)
		return;

	annotation_path = window_get_annotation_prefix(win, False);
	strcat(annotation_path,"*");

	if(glob(annotation_path, 0, NULL, &tglob)) {
		DBG(debug_info("%s(): glob failed", __func__);)
		win->set_flag(win, ANNOTATE, True);
		win->set_flag(win, ANNOTATION_PRESENT, False);
		window_set_int_property(win, _AC_W_ANNOTATE_STATE, ANNOTATION_NOT_THERE);
	} else if(tglob.gl_pathc) {
		win->set_flag(win, ANNOTATE, False);
		win->set_flag(win, ANNOTATION_PRESENT, True);
		window_set_int_property(win, _AC_W_ANNOTATE_STATE, ANNOTATION_THERE);
	} else {
		win->set_flag(win, ANNOTATE, True);
		win->set_flag(win, ANNOTATION_PRESENT, False);
		window_set_int_property(win, _AC_W_ANNOTATE_STATE, ANNOTATION_NOT_THERE);
	}

	win->redraw(win);
	globfree(&tglob);
	free(annotation_path);
}

// Send the target window a XConfigureEvent (Its dimensions have changed)
void window_send_configure_event(window *w)
{
	XConfigureEvent xce;

	xce.type = ConfigureNotify;
	xce.x = w->x;
	xce.y = w->y;
	xce.width = w->width;
	xce.height = w->height;

	xce.border_width = 0;
	xce.above = None;
	xce.override_redirect = 0;
	
	XSendEvent(w->wm->display, w->app_window, False, SubstructureNotifyMask,
			(XEvent *)&xce);
	XSync(w->wm->display, False);
}

// Send the target window any WM_PROTOCOLS message
void window_send_protocol_event(window *w, Atom protocol)
{
	XEvent xcm;

	xcm.type = ClientMessage;

	xcm.xclient.window = w->app_window;
	xcm.xclient.message_type = w->wm->atomlist[WM_PROTOCOLS];
	xcm.xclient.format = 32;
	xcm.xclient.data.l[0] = protocol;
	xcm.xclient.data.l[1] = CurrentTime;

	XSendEvent(w->wm->display, w->app_window, False, NoEventMask, &xcm);
	XSync(w->wm->display, False);
}

// Read a Xwindow property atom as atom
Atom xwindow_get_atom_property(Window win, wm *wm, atomlist_index index)
{
	Atom at, actual_type;
	int status, actual_format;
	unsigned long *property_value;
	Atom property_return;
	unsigned long n,b;

	at=wm->atomlist[index];

	status = XGetWindowProperty(wm->display, win, at, 0,
			MAX_PROPERTY_LENGTH, False, XA_ATOM, &actual_type,
			&actual_format, &n, &b, (unsigned char **)&property_value);
	if( (status!=Success) || (!n) || (!property_value) || (actual_type!=XA_ATOM))
		return 0;
	property_return = property_value[0];
	XFree(property_value);
	return property_return;
}

// Read a window property atom as atom
Atom window_get_atom_property(window *w, atomlist_index index)
{
	Atom at, actual_type;
	int status, actual_format;
	unsigned long *property_value;
	Atom property_return;
	unsigned long n,b;

	at=w->wm->atomlist[index];

	status = XGetWindowProperty(w->wm->display, w->app_window, at, 0,
			MAX_PROPERTY_LENGTH, False, XA_ATOM, &actual_type,
			&actual_format, &n, &b, (unsigned char **)&property_value);
	if( (status!=Success) || (!n) || (!property_value) || (actual_type!=XA_ATOM))
		return 0;
	property_return = property_value[0];
	XFree(property_value);
	return property_return;
}

// Read a window property atom as int
int window_get_int_property(window *w, atomlist_index index)
{
	Atom at, actual_type;
	int status, actual_format, int_ret;
	int *property_value=0;
	unsigned long n,b;

	at=w->wm->atomlist[index];

	status = XGetWindowProperty(w->wm->display, w->app_window, at, 0,
			MAX_PROPERTY_LENGTH, False, AnyPropertyType, &actual_type,
			&actual_format, &n, &b, (unsigned char **)&property_value);
	if( (status!=Success) || (!n) || (!property_value) )
		return -1;
	
	int_ret = property_value[0];
	XFree(property_value);
	return int_ret;
}

// Read a window property atom as text
void window_get_text_property(window *w, atomlist_index index, unsigned char *string_ret)
{
	unsigned char *prop_ret;
	Atom at, actual_type;
	int status, actual_format;
	unsigned long n,b;

	at=w->wm->atomlist[index];

	status = XGetWindowProperty(w->wm->display, w->app_window, at, 0,
			MAX_PROPERTY_LENGTH, False, AnyPropertyType, &actual_type,
			&actual_format, &n, &b, &prop_ret);
	if( (status!=Success) || (!n) || (!string_ret) )
		string_ret[0] = 0;
	else {
		strcpy(string_ret, prop_ret);
		XFree(prop_ret);
	}
	
}

// Set a window property atom as int
int window_set_int_property(window *win, atomlist_index index, const unsigned int property_value)
{
	Atom at;
	int status;
	wm *wm=win->wm;

	at=wm->atomlist[index];

	status = XChangeProperty(wm->display, win->app_window, at,
			XA_CARDINAL, 32, PropModeReplace,
			(unsigned char *)&property_value, 1);
	return status;
}

// Get a X window property atom as text
void xwindow_get_text_property(wm *wm, Window win, unsigned int atomlist_index, unsigned char *string_ret)
{
	unsigned char *prop_ret;
	Atom at, actual_type;
	int status, actual_format;
	unsigned long n,b;

	at=wm->atomlist[atomlist_index];

	status = XGetWindowProperty(wm->display, win, at, 0,
			MAX_PROPERTY_LENGTH, False, AnyPropertyType, &actual_type,
			&actual_format, &n, &b, &prop_ret);

	if( (status!=Success) || (!n) || (!prop_ret) )
		string_ret[0] = 0;
	else {
		strcpy(string_ret, prop_ret);
		XFree(prop_ret);
	}
}

// Set a X window property atom as text
int xwindow_set_text_property(wm *wm, Window win, unsigned int atomlist_index, const char *property_value)
{
	Atom at;
	int status;
	Display *display = wm->display;

	at=wm->atomlist[atomlist_index];

	status = XChangeProperty(display, win, at, XA_STRING, 8,
			PropModeReplace, property_value, strlen(property_value));
	XFlush(display);
	return status;
}

// Set a window property atom as text
int window_set_text_property(window *win, unsigned int atomlist_index, const char *property_value)
{
	Atom at;
	int status;
	wm *wm=win->wm;

	at=wm->atomlist[atomlist_index];

	status = XChangeProperty(wm->display, win->app_window, at, XA_STRING,
			8, PropModeReplace, property_value, strlen(property_value));
	return status;
}

// If the window has set up a WM_PROTOCOL for killing itself, inform it
// to commit suicide, else lets murder it :)
void window_handle_deletion(window *w)
{
	int i,num;
	Atom *protocols;

	if(XGetWMProtocols(w->wm->display, w->app_window, &protocols, &num)) {
		for(i=0;i<num;i++)
			if(protocols[i] == w->wm->atomlist[WM_DELETE_WINDOW])
				window_send_protocol_event(w, w->wm->atomlist[WM_DELETE_WINDOW]);
		XFree(protocols);
	} else {
		if(w->pid > 0)
			kill(w->pid, SIGINT);
		
		XKillClient(w->wm->display, w->app_window);
	}
}

// find the window with specified window id
window *window_find_windowid(wm *wm, Window wid)
{
	window *w;

	//For_All_Mappable_Do(wm, w);
	For_All_Windows_Do(wm, w);
	if(w->app_window == wid)
		return w;
	//For_All_Mappable_Done(wm, w);
	For_All_Windows_Done(wm, w);
	return NULL;
}

// find the window with specified appid
window *window_find_appid(wm *wm, unsigned long appid)
{
	window *w;

	For_All_Mappable_Do(wm, w);
	if(w->appid == appid)
		return w;
	For_All_Mappable_Done(wm, w);
	return NULL;
}

// find if the specified window exists
Bool window_exists(wm *wm, window *win)
{
	window *w;

	For_All_Mappable_Do(wm, w);
	if(w == win)
		return True;
	For_All_Mappable_Done(wm, w);

	return False;
}

// Callback when the close button is pressed
Bool window_close_callback(px_widget *widget, px_event *ev)
{
	wm *wm;
	window *win = (window *)ev->private_data;

	wm=win->wm;

	if(win->fast_load) {
		// Do not delete, just hide
		if(wm->launcher_window) {
			window *launch_top=wm->launcher_window;
				
			/* Navigate through the layers till the top */
			while(launch_top->layer_above)
				launch_top=launch_top->layer_above;
			launch_top->show(launch_top);

			win->hide(win);
			win->fast_visible=False;
		}
	} else
		window_handle_deletion(win);

	return True;
}

// Callback when the tick button is pressed
Bool window_tick_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;

	if(ev->xev->type == ButtonPress)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_TICK]);
	else if(ev->xev->type == ButtonRelease)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_TICK_REL]);

	return True;
}

// Callback when the up button is pressed
Bool window_up_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;

	if(ev->xev->type == ButtonPress)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_UP]);
	else if(ev->xev->type == ButtonRelease)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_UP_REL]);

	return True;
}

// Callback when the voice button is pressed
Bool window_voice_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;

	if(ev->xev->type == ButtonPress)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_VOICE]);
	else if(ev->xev->type == ButtonRelease)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_VOICE_REL]);

	return True;
}

void xstroke_start_reconition(wm *wm)
{
	window *xstroke_window=wm->xstroke_window;

	if(xstroke_window) {
		XEvent xcm;

		xcm.type = ClientMessage;

		xcm.xclient.window = xstroke_window->app_window;
		xcm.xclient.message_type = wm->atomlist[XSTROKE_RECOGNITION_CMD];
		xcm.xclient.format = 32;
		xcm.xclient.data.l[0] = XSTROKE_START_RECOGNITION;
		xcm.xclient.data.l[1] = CurrentTime;

		XSendEvent(wm->display, xstroke_window->app_window, False, NoEventMask, &xcm);
		XSync(wm->display, False);
	} else {
		char cmd[TMP_BUF_SIZE];
		char env1[TMP_BUF_SIZE],env2[TMP_BUF_SIZE];

		sprintf(cmd,XSTROKE_CMDLINE,58,300);
		sprintf(env1,"_AC_W_APPID=%d",AC_XSTROKE_APPID);
		sprintf(env2,"NET_WM_NAME=xstroke");
		spawn_child(cmd,env1,env2,NULL);
	}
}

void xstroke_stop_recognition(wm *wm)
{
	window *xstroke_window=wm->xstroke_window;

	if(xstroke_window) {
		XEvent xcm;

		xcm.type = ClientMessage;

		xcm.xclient.window = xstroke_window->app_window;
		xcm.xclient.message_type = wm->atomlist[XSTROKE_RECOGNITION_CMD];
		xcm.xclient.format = 32;
		xcm.xclient.data.l[0] = XSTROKE_STOP_RECOGNITION;
		xcm.xclient.data.l[1] = CurrentTime;

		XSendEvent(wm->display, xstroke_window->app_window, False, NoEventMask, &xcm);
		XSync(wm->display, False);
	}
}

// Callback when the xstroke button is pressed
Bool window_xstroke_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;
	xstroke_start_reconition(win->wm);

	return True;
}

// Callback when the printer button is pressed
Bool window_printer_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;

	if(ev->xev->type == ButtonPress)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_PRINTER]);
	else if(ev->xev->type == ButtonRelease)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_PRINTER_REL]);

	return True;
}

// Callback when the email button is pressed
Bool window_email_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;

	win->onetap = False;
	if(ev->xev->type == ButtonPress)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_EMAIL]);
	else if(ev->xev->type == ButtonRelease)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_EMAIL_REL]);

	return True;
}

// Callback when the onetap button is pressed
Bool window_onetap_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;

	win->onetap = True;
	if(ev->xev->type == ButtonPress)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_EMAIL]);
	else if(ev->xev->type == ButtonRelease)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_EMAIL_REL]);

	return True;
}

// Callback when the nextpage button is pressed
Bool window_nextpage_press_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;
	
	if(nextpage_pressed) {
		double diff = fabs(difftime(nextpage_press_time, time(NULL)));
		if(diff <= 1.0)
			return True;
	}
	window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_NEXTPAGE]);
	nextpage_pressed = True;
	nextpage_press_time = time(NULL);

	return True;
}

// Callback when the nextpage button is Released
Bool window_nextpage_rel_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;
	
	window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_NEXTPAGE_REL]);
	nextpage_pressed = False;

	return True;
}

// Callback when the prevpage button is pressed
Bool window_prevpage_press_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;
	
	if(prevpage_pressed) {
		double diff = fabs(difftime(prevpage_press_time, time(NULL)));
		if(diff <= 1.0)
			return True;
	}
	window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_PREVPAGE]);
	prevpage_pressed = True;
	prevpage_press_time = time(NULL);

	return True;
}

Bool window_prevpage_rel_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;
	
	window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_PREVPAGE_REL]);
	prevpage_pressed = False;

	return True;
}

// Callback when the help button is pressed
Bool window_help_callback(px_widget *widget, px_event *ev)
{
	window *win;
	FILE *fp;
	char *help_path, *background_path;
	char env1[TMP_BUF_SIZE],env2[TMP_BUF_SIZE],env3[TMP_BUF_SIZE];
	char *cmd, *real_help_template;
	wm *wm;

	win = (window *)ev->private_data;
	wm = win->wm;

	if(win->is_chikki) {
		real_help_template = CHIKKI_HELP_TEMPLATE;
	} else {
		real_help_template = HELP_TEMPLATE;
	}

	if(!win->help_key) {
		help_path = (char *)malloc(strlen(real_help_template) + strlen(win->name) + strlen("default") + 1);
		if (!help_path)
			err_exit("%s(): Cannot allocate memory", __func__);
		sprintf(help_path, real_help_template, win->name, "default");
	} else {
		if(!strcmp(win->help_key, ANNOTATE_HELP_KEY)) {
			help_path = (char *)malloc(strlen(real_help_template) + strlen(ANNOTATE_APP_NAME) + strlen(ANNOTATE_HELP_KEY) + 1);
			if (!help_path)
				err_exit("%s(): Cannot allocate memory", __func__);
			sprintf(help_path, real_help_template, ANNOTATE_APP_NAME, ANNOTATE_HELP_KEY);
		} else {
			help_path = (char *)malloc(strlen(real_help_template) + strlen(win->name) + strlen(win->help_key) + 1);
			if (!help_path)
				err_exit("%s(): Cannot allocate memory", __func__);
			sprintf(help_path, real_help_template, win->name, win->help_key);
		}
	}

	// Path to store background image
	background_path = (char *)malloc(strlen("/tmp/") + strlen(win->name) + strlen(BACKGROUND_GRAB_IMG) + 2);
	if (!background_path)
		err_exit("%s(): Cannot allocate memory", __func__);
	sprintf(background_path, "/tmp/%s_%s",win->name, BACKGROUND_GRAB_IMG);

	fp=fopen(background_path, "wb");
	if(fp) {
		if(wm->keyboard_window && wm->is_keyboard_mapped) {
			wm->keyboard_window->hide(wm->keyboard_window);
		}
		save_image(win->wm->display, win->app_window, fp, 0, 0, 0, 0);
		fclose(fp);
	} else {
		DBG(debug_info("Unable to capture background info");)
	}

	cmd = (char *)malloc(strlen(HELP_CMDLINE) + strlen(win->name) + 
			strlen(help_path) + 1);
	if(!cmd)
		err_exit("%s(): Cannot allocate memory", __func__);

	sprintf(env1,"_AC_W_APPID=%lu",win->appid);
	sprintf(env2,"_NET_WM_NAME=%s",win->name);
	sprintf(env3,"_AC_W_LAYER_WINDOWID=%lu",win->app_window);
	sprintf(cmd, HELP_CMDLINE, help_path, win->name);

	spawn_child(cmd,env1,env2,env3,NULL);

	free(cmd);
	free(background_path);
	free(help_path);

	return True;
}

// Callback when the pref button is pressed
Bool window_pref_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;

	if(ev->xev->type == ButtonPress)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_PREF]);
	else if(ev->xev->type == ButtonRelease)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_PREF_REL]);

	return True;
}

// Callback when the retractable toolbar button is pressed
Bool window_retractbar_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;

	if(ev->xev->type == ButtonPress)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_RETRACTBAR]);
	else if(ev->xev->type == ButtonRelease)
		window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_RETRACTBAR_REL]);

	return True;
}

// Callback when the next application button is pressed
Bool window_nextapp_callback(px_widget *widget, px_event *ev)
{
	window *newwin, *win = (window *)ev->private_data;

	newwin = window_get_next_mappable(win);
	if(newwin != win) {
		win->hide(win);
		newwin->show(newwin);
	}

	return True;
}

// Callback when the prev application button is pressed
Bool window_prevapp_callback(px_widget *widget, px_event *ev)
{
	window  *newwin, *win = (window *)ev->private_data;

	newwin = window_get_prev_mappable(win);
	if(newwin != win) {
		win->hide(win);
		newwin->show(newwin);
	}

	return True;
}

// Callback when home(launcher button) button is pressed
Bool window_home_callback(px_widget *widget, px_event *ev)
{
	wm *wm;
	window *win = (window *)ev->private_data;
	wm=win->wm;

	if(wm->launcher_window) {
		window *launch_top=wm->launcher_window;
		        
		/* Navigate through the layers till the top */
		while(launch_top->layer_above)
			launch_top=launch_top->layer_above;
		launch_top->show(launch_top);

		win->hide(win);
	}

	return True;
}

char *window_get_annotation_prefix(window *win, Bool escaped)
{
	if(!win->annotation_key) {
		return get_annotation_prefix(win->name,"default", escaped);
	} else {
		return get_annotation_prefix(win->name,win->annotation_key, escaped);
	}
}

// Callback when annotate button is pressed
void window_launch_annotate(window *win)
{
	FILE *fp;
	char *annotation_path;
	char *background_path;
	char env1[TMP_BUF_SIZE],env2[TMP_BUF_SIZE],env3[TMP_BUF_SIZE];
	char *cmd;
	static time_t prev_annotate_time = 0;
	wm *wm;

	// To prevent multiple grabs from screwing up the final image
	// This typically happens when the button is clicked multiple times

	if(fabs(difftime(prev_annotate_time, time(NULL)))<=2.5) {
		debug_info("Annotate launch locked");
		return;
	}

	wm = win->wm;

	// Here prefix must be escaped for bash's sake
	annotation_path = window_get_annotation_prefix(win, True);

	// Path to store background image
	background_path = (char *)malloc(strlen("/tmp/") + strlen(win->name) + strlen(BACKGROUND_GRAB_IMG) + 2);
	if (!background_path)
		err_exit("%s(): Cannot allocate memory", __func__);
	sprintf(background_path, "/tmp/%s_%s",win->name, BACKGROUND_GRAB_IMG);

	fp=fopen(background_path, "wb");
	if(fp) {
		if(wm->keyboard_window && wm->is_keyboard_mapped) {
			wm->keyboard_window->hide(wm->keyboard_window);
		}
		save_image(wm->display, win->app_window, fp, 0, 0, 0, 0);
		fclose(fp);
	} else {
		DBG(debug_info("Unable to capture background info");)
	}
	free(background_path);

	cmd = (char *)malloc(strlen(ANNOTATE_CMDLINE) + strlen(win->name) + 
			strlen(annotation_path) + 1);
	if(!cmd)
		err_exit("%s(): Cannot allocate memory", __func__);

	sprintf(env1,"_AC_W_APPID=%lu",win->appid);
	sprintf(env2,"_NET_WM_NAME=%s",win->name);
	sprintf(env3,"_AC_W_LAYER_WINDOWID=%lu",win->app_window);

	if(win->needs_grab)
		sprintf(cmd, ANNOTATE_CMDLINE, annotation_path, win->name );
	else
		sprintf(cmd, ANNOTATE_NOGRAB_CMDLINE, annotation_path, win->name );

	prev_annotate_time = time(NULL);
	
	spawn_child(cmd,env1,env2,env3,NULL);

	free(annotation_path);
	free(cmd);
}

// Callback when annotate button is pressed
Bool window_annotate_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;
	window_launch_annotate(win);

	return True;
}

// Callback when keyboard button is pressed
Bool window_keyboard_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;

	if(win->flags.bits.needs_keyboard) {
		win->set_flag(win, NEEDS_KEYBOARD, False);
	} else {
		win->set_flag(win, NEEDS_KEYBOARD, True);
		win->set_flag(win, XSTROKE, False);
	}

	window_handle_keyboard(win);

	return True;
}

// Callback when star button is pressed
Bool window_star_callback(px_widget *widget, px_event *ev)
{
	wm *wm;
	window *win = (window *)ev->private_data;

	wm=win->wm;

	XGrabServer(wm->display);
	new_star_window(win->wm);
	XUngrabServer(wm->display);

	return True;
}

// Callback when title label is pressed
Bool window_title_callback(px_widget *widget, px_event *ev)
{
	wm *wm;
	window *win = (window *)ev->private_data;

	wm=win->wm;
	if(win->appid != AC_STATUS_APPID) {
		XGrabServer(wm->display);
		new_winlist_window(wm);
		XUngrabServer(wm->display);
	}

	return True;
}

// Callback when keyboard chooser button is pressed
Bool window_keyboard_chooser_callback(px_widget *widget, px_event *ev)
{
	wm *wm;
	window *win = (window *)ev->private_data;
	
	wm=win->wm;

	XGrabServer(wm->display);
	new_ichoose_window(wm);
	XUngrabServer(wm->display);

	return True;
}

// Callback when status button is pressed
Bool window_status_callback(px_widget *widget, px_event *ev)
{
	wm *wm;
	window *win = (window *)ev->private_data;
	wm = win->wm;

	if(wm->status_window) {
		wm->status_window->show(wm->status_window);
	} else {
		char env1[TMP_BUF_SIZE], env2[TMP_BUF_SIZE], env3[TMP_BUF_SIZE];

		sprintf(env1,"_AC_W_APPID=%d", AC_STATUS_APPID);
		sprintf(env2,"_NET_WM_NAME=Status");
		sprintf(env3,"_AC_W_LAYER_WINDOWID=%lu",win->app_window);
		spawn_child(STATUS_CMDLINE, env1, env2, env3, NULL);
	}

	return True;
}

// Callback when the speaker icon is pressed
Bool window_speaker_callback(px_widget *widget, px_event *ev)
{
	window *win = (window *)ev->private_data;
	
	if(win->app_speaks) {
		// Send an event to the application
		if(ev->xev->type == ButtonPress)
			window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_SPEAKER]);
		else if(ev->xev->type == ButtonRelease)
			window_send_protocol_event(win, win->wm->atomlist[_AC_BUTTON_SPEAKER_REL]);
	} else if(ev->xev->type == ButtonRelease) {
		// Write a selection to flite from xclip
		char buf[TMP_BUF_SIZE];
		int fd;

		fd=open("/dev/dsp",O_WRONLY|O_NONBLOCK);
		if(fd<0) {
			sprintf(buf,"/usr/local/bin/bubble \"Speaker is in use\" &");
			spawn_child(buf,NULL);
		}
		close(fd);

#ifdef __arm__
		sprintf(buf,"a=`%s -d :0.0 -o`; echo $a | %s", XCLIP_CMDLINE, FLITE_CMDLINE);
#else
		sprintf(buf,"a=`%s -o`; echo $a | %s", XCLIP_CMDLINE, FLITE_CMDLINE);
#endif
		spawn_child(buf,NULL);
	}

	return True;
}

