/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include <sys/sysmacros.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xlibint.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xproto.h>
#include <X11/xpm.h>
#include <X11/Xatom.h>
#include <X11/extensions/shape.h>
#include <X11/Xmu/WinUtil.h>

#include <X11/cursorfont.h>

#include <png.h>

#define TOP 0
#define BOTTOM 0

#define SCALE_FACTOR (65535.0/31.0)

extern int count, height, width;

void Fatal_Error(char *err)
{

	fprintf(stderr,"%s\n",err);
	exit(-1);

}

int png_write_rgba(FILE *fp, unsigned char *data, int w, int h, int bit_depth)
{
	png_structp png_ptr;
	png_infop info_ptr;
	register int i,j;
	int pass;
	png_color_8 sig_bit;
	unsigned char *tmp=NULL, *pp;
	unsigned short pixel;

	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, 
		(png_voidp)NULL, (png_error_ptr)NULL, (png_error_ptr)NULL);

	if (!png_ptr) {
		return -1;
	}
	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) {
		png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
	   return -1;
	}

	png_init_io(png_ptr, fp);

	png_set_filter(png_ptr, 0, PNG_FILTER_NONE );

	png_set_compression_level(png_ptr, 1);
	//png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);

	sig_bit.red = 8;
	sig_bit.blue = 8;
	sig_bit.green = 8;

	png_set_sBIT(png_ptr, info_ptr, &sig_bit);

	png_set_packing (png_ptr);

	png_set_IHDR(png_ptr, info_ptr, w, h, 8, PNG_COLOR_TYPE_RGB,
		0, PNG_COMPRESSION_TYPE_BASE, 
		PNG_FILTER_TYPE_DEFAULT);

	png_write_info(png_ptr, info_ptr);

	pp = (unsigned char *)malloc(h * w );

	if(!pp) {
		png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
		return -1;
	}

	for(pass=0; pass<png_set_interlace_handling(png_ptr); pass++) {
		unsigned short red=0, blue=0, green=0;
		for(j=0;j<h;j++) {
			tmp = pp;
			for(i=0; i<w; i++) {
				pixel = ((unsigned short) data[2*j*w+2*i+1] << 8 ) | 
							(unsigned short) data[2*j*w+2*i];


				if(bit_depth >= 16) {
#ifdef __arm__
					//red   = (65535.0/63.0) * (float)((pixel & 0xfc00) >> 10);
					//green = (SCALE_FACTOR) * (float)((pixel & 0x3e0) >> 5);
					//blue  = (SCALE_FACTOR) * (float)((pixel & 0x1f));

					red   = ((pixel & 0xfc00) >> 10) << 10;
					green = ((pixel & 0x3e0) >> 5) << 11;
					blue  = ((pixel & 0x1f)) << 11;
#else 
					red   = (SCALE_FACTOR) * (float)((pixel & 0xf800) >> 11);
					green = (SCALE_FACTOR) * (float)((pixel & 0x7e0) >> 6);
					blue  = (SCALE_FACTOR) * (float)((pixel & 0x1f));
#endif
				} 

				if(bit_depth == 12) {
                                        red   =  ((pixel & 0x0f00) >> 8) << 12;
                                        green =  ((pixel & 0xf0) >> 4) << 12;
                                        blue  =  (pixel & 0xf) << 12;
				}

				*tmp++ = red >> 8;
				//*tmp++ = red & 0xff;

				*tmp++ = green >> 8;
				//*tmp++ = green & 0xff;

				*tmp++ = blue >> 8;
				//*tmp++ = blue & 0xff;
			}
			png_write_row(png_ptr, pp);
		}
	}

	png_write_end(png_ptr, NULL);

	free(pp);
	png_destroy_write_struct(&png_ptr, &info_ptr);

	return 0;
}


int simputer_png_write(FILE *fp, unsigned char *data, int w, int h, int bit_depth)
{
	png_structp png_ptr;
	png_infop info_ptr;
	int i;
	png_bytep rowp[h];


	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, 
		(png_voidp)NULL, (png_error_ptr)NULL, (png_error_ptr)NULL);

	if (!png_ptr) {
         return -1;
	}
        info_ptr = png_create_info_struct(png_ptr);
        if (!info_ptr) {
           png_destroy_write_struct(&png_ptr,
             (png_infopp)NULL);
           return -1;
        }

	png_init_io(png_ptr, fp);

	png_set_filter(png_ptr, 0, PNG_FILTER_NONE | PNG_FILTER_SUB |
	           PNG_FILTER_PAETH);

	png_set_compression_level(png_ptr, 3);

        
	png_set_IHDR(png_ptr, info_ptr, w, h, bit_depth, PNG_COLOR_TYPE_GRAY, 
		PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, 
		PNG_FILTER_TYPE_DEFAULT);	

	for(i=0;i<h;i++) {
		rowp[i] = data + i*w ;
	}

	png_write_info(png_ptr, info_ptr);

	png_write_image(png_ptr, rowp);

	png_write_end(png_ptr, info_ptr);

	png_destroy_write_struct(&png_ptr, &info_ptr);

    return 0;
}

int save_image(Display *dis, Window window, FILE *fp, int left, int top, int right, int bottom)
{
	XImage *img;
	unsigned char *data = NULL;
	int i,j;
	int screen;
	XWindowAttributes win_info;
	int new_width, new_height, new_depth;
	int min, max, bytes_per_pixel;
	unsigned long val;
	int absx, absy, x, y, dwidth, dheight, ret;
	Window dummywin, win;

	if(!fp) {
		return -1;
	}

	if( (left < 0) ||  (right < 0) || (top < 0) || (bottom < 0)) {
		return -1;
	}

	screen = DefaultScreen(dis);

    // Use XGetWindowAttributes instead of DisplayHeight & DisplayWidth,
    // which do not handle rotated displays. This was causing grab to clip
    // a portion of the grabbed screen
    // - Vivek & Phani
    XGetWindowAttributes(dis, DefaultRootWindow(dis), &win_info);
    dwidth = win_info.width;
    dheight = win_info.height;

	win = XmuClientWindow (dis, window);

	//printf("window = %x win = %x \n",window, win);

	XGetWindowAttributes(dis, win, &win_info);

	new_width = win_info.width;
	new_height = win_info.height;
	new_depth = win_info.depth;

	if(!XTranslateCoordinates(dis, win, RootWindow(dis, screen), 0, 0, &absx, &absy, &dummywin)) {
		return -1;
	}

	win_info.x = absx;
	win_info.y = absy;

	if(absx < 0) {
		new_width += absx;
		absx = 0;
	}

	if(absy < 0) {
		new_height += absy;
		absy = 0;
	}

	if(absx + new_width > dwidth) {
		new_width = dwidth - absx;
	}

	if(absy + new_height > dheight) {
		new_height = dheight - absy;
	}

	x = absx - win_info.x;
	y = absy - win_info.y;

	//new_height = new_height - TOP - BOTTOM ;

	if(new_height <= 0) {
		return -1;
	}

	//printf("width = %d height = %d \n",new_width, new_height);
	//printf("absx = %d absy = %d \n",absx, absy);
	//printf("x = %d y = %d \n",x, y);

	//img = XGetImage(dis, win, x, y, new_width, new_height, AllPlanes, ZPixmap);



	new_width -= (left + right);
	new_height -= (top + bottom);

	if( (new_height <= 0) || (new_width <= 0) ) {
		return -1;
	}

	img = XGetImage(dis, RootWindow(dis, screen), absx+left, absy+top, new_width, new_height, AllPlanes, ZPixmap);

	if(!img) {
		return(-1);
	}

	if(new_depth <= 16) {

		if(new_width & 0x1) {
			new_width += 1;
		}

		data = (unsigned char *)malloc(new_width*new_height*2);

		if(new_depth <4) {
			/*
			 *  depths 1,2 are  not supported :(
			 */
			XDestroyImage(img);
			free(data);
			return -1;
		}

		if(new_depth == 4) {

			char left, right;

			/*
			 *	If its a 4 bit display, image width and bytes_per_line
			 *	may not match. so copy byte by byte. and convert 4 bpp
			 *	to 8 bpp.
			 */

			// new_height += top;

			for(i=0;i<new_height;i++) {
				for(j=0;j<new_width;j++) {
					// data[(i-left)*new_width+j]
				 	left = img->data[i*img->bytes_per_line+j];

					right = left & 0xf;
					left >>= 4;
					data[i*new_width+j] = (right << 4) | left;

				}
			}

			// new_height -= top;

			// new_width <<= 1;
                        
			new_depth = 4 ;

			//simputer_png_write();
			ret = simputer_png_write(fp, data, new_width, new_height, new_depth);

			XDestroyImage(img);
			free(data);

			return 0;

		} if (new_depth == 12) {
			memcpy(data, (img->data), new_height*new_width*2);	
		} else {
			/*
			 * if depth is 8, 16 proceed ...
			 */
			max = new_width*new_height*(img->bits_per_pixel>>3);
			min = left*new_width*(img->bits_per_pixel>>3);

			// memcpy(data, (img->data)+min, max);
			memcpy(data, (img->data), new_height*new_width*(new_depth>>3));
		}
	} else {
		int red, green, blue;

		new_depth = 16;
		data = (unsigned char *)malloc(new_width*new_height*2);

		min = 0; //new_width*(left)*(img->bits_per_pixel>>3);
		max = new_width*(new_height)*(img->bits_per_pixel>>3);

		bytes_per_pixel = img->bits_per_pixel >>3;

		j=0;
        	val = 0;
		for(i=min;i<max;i+=bytes_per_pixel) {
			memcpy(&(val), &(img->data[i]), 4);
			/*
			 * for 24 bit depth, r,g,b masks are 0xff0000, 
                         * 0xff00 and 0xff
			 */

			red   = ( val & 0xf80000 ) >> 8;
			green = ( val & 0xfc00 ) >> 5;
			blue  = ( val & 0xf8 ) >> 3;

			val = red | green | blue;

			data[j]   = val & 0xff;
			data[j+1] = (val >> 8) & 0xff;
			j+=2; 
		}
	}

	// new_height = new_height - top - bottom ;

	png_write_rgba(fp, data, new_width, new_height, new_depth);

	XDestroyImage(img);
	free(data);

	return 0;
}
