/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "alchemy.h"
#include "basic_window.h"
#include "ichoose_window.h"
#include <picowidget.h>
#include <frame.h>
#include <listbox.h>
#include <font.h>
#include "utils.h"
struct _layout_list *layout_list = NULL;

px_widget *ichoose_create_listbox(wm *wm, px_widget *frame, int *width, int *height, int *n);

// Create a new window showing the list of keyboard layouts
window *new_ichoose_window(wm *wm)
{
	Window frame_window;
	Display *dpy=wm->display;
	XSetWindowAttributes win_attr;
	window *w;
	int n, width, height;
	px_widget *listbox;

	win_attr.override_redirect = True;
	win_attr.background_pixel = wm->black_pixel;
	win_attr.border_pixel = wm->black_pixel;
	win_attr.event_mask = ButtonPressMask|ButtonReleaseMask|ExposureMask|
		EnterWindowMask|LeaveWindowMask|PointerMotionMask;

	// Create with small size, Resize and Reposition after getting
	// the actual size and position
	frame_window = XCreateWindow(dpy, wm->root, 0,0, 1, 1, 0,
			CopyFromParent, CopyFromParent, CopyFromParent, 
			CWOverrideRedirect|CWBackPixel|CWBorderPixel|CWEventMask,
			&win_attr);

	// Allocate structures, there is no separate app_window here
	w = new_basic_window(wm, frame_window, MENU_WIN);
	w->frame_window = frame_window;
	w->app_window = frame_window;
	w->appid = AC_KEYBOARD_MENU_APPID;
	
	// The topbar frame is used to hold the listbox widget
	w->topbar_frame=new_toplevel_frame(frame_window, 0, 0, 1, 1, NULL);

	// Create the listbox and figure out dimensions
	listbox = ichoose_create_listbox(wm, w->topbar_frame, &width, &height, &n);
	listbox->redraw(listbox);
	w->topbar_frame->set_size(w->topbar_frame, width, height);
	
	w->x = (wm->display_width - width)/2;
	w->y = wm->mapped_window->height + wm->mapped_window->y -  height;

	// Move/Resize window according to calculated dimensions
	XMoveResizeWindow(dpy, frame_window, w->x, w->y, width, height);
	XSetWindowBorderWidth(dpy, frame_window, 1 );
	
	w->destroy = ichoose_window_destroy;
	
	w->topbar_frame->redraw(w->topbar_frame);
	w->topbar_frame->show(w->topbar_frame);
	
	window_set_state(w, NormalState);

	// Show the window
	XMapWindow(dpy, w->frame_window);

	wm->is_menu_mapped = True;

	// Grab pointer
	XGrabPointer(dpy, wm->root, True, (ButtonPressMask|ButtonReleaseMask),
			GrabModeAsync, GrabModeAsync, None, wm->normal_cursor, CurrentTime);
	
	XSync(dpy, False);

	return w;
}

static void *menu_background;
static char menu_bg_type;
static Bool menu_initialized=False;
static Bool xstroke_available=False;

// Create the listbox of keyboard layouts
px_widget *ichoose_create_listbox(wm *wm, px_widget *frame, int *width, int *height, int *n)
{
	int i;
	int twidth, theight;
	px_widget *list;
	px_widget* label;
	char* keyboard_category = wm->mapped_window->keyboard_category;
	struct _layout_list *node = NULL;	
	const int MAXSIZE = 255;

	list = new_widget(LISTBOX, frame, 0,0,1,1, NULL);
	*width = 0;
	*height = list->specific.listbox.inter_widget_space;
	*n = 0;

	if(!menu_initialized) {
		menu_background = wm->current_theme->get_menu_background(
				wm->current_theme, &menu_bg_type);
		menu_initialized = True;
		if(!access(XSTROKE_LOCATION, R_OK | X_OK))
			xstroke_available=True;
	}

	if(menu_bg_type) {
		list->set_background_color(list, (px_color *) menu_background);
	} else {
		list->set_background_pixmap_from_pixmap(list, (px_pixmap *) menu_background);
	}

   	if (layout_list == NULL || strcmp(keyboard_category, layout_list->name) != 0)
	{
		char *directory = "/usr/local/settings/keyboard/kmlfiles/";
		char *filename = (char*)malloc(sizeof(char) * (strlen(directory) + strlen(keyboard_category) + 10));
		char ctemp[MAXSIZE];
		FILE* fileptr;
	   
		strcpy(filename, directory);
		strcat(filename, keyboard_category);
		strcat(filename, ".category");

		fileptr = fopen(filename,"rt");
		free(filename);
		filename = NULL;

		if (layout_list != NULL)
		{
			struct _layout_list *node, *next_node;

			node = layout_list;
			while (node != NULL)
			{
				next_node = node->next;
				free(node->name);
				free(node);
				node = next_node;
			}
			layout_list = NULL;
		}

		if (fileptr != NULL)
		{
			struct _layout_list* last_node = NULL;
			
		
			layout_list = (struct _layout_list*)malloc(sizeof(struct _layout_list));
			layout_list->name = (char*) malloc(sizeof(char)*(strlen(keyboard_category) + 1));
			strcpy(layout_list->name, keyboard_category);
			layout_list->next = NULL;
			last_node = layout_list;

			while(fgets(ctemp,MAXSIZE,fileptr) != NULL)
			{
				struct _layout_list* node = (struct _layout_list*)malloc(sizeof(struct _layout_list));
				printf("Creating node with tag %s\n", ctemp);
				ctemp[strlen(ctemp) - 1]=0;
				node->name = strdup(ctemp);
				node->next = NULL;
	
				last_node->next = node;
				last_node = node;
			}
			fclose(fileptr);
		}
		else // This is not a valid category or category files do not exist.
		{
			int twidth, theight;
			char msg[] = "No layouts available";
				

			label = new_widget(LABEL, list, 0, 0, 1, 1, NULL);
			label->specific.label.set_wrap_type(label, NO_WRAP);
			if(menu_bg_type) {
				label->set_background_color(label, (px_color *) menu_background);
			}

			list->specific.listbox.set_child(list, label, *n);
			label->specific.label.set_font(label, DEFAULT_FONT);
			label->specific.label.set_normal_color(label, "gray");

			// Set name and callbacks

			label->specific.label.set_text(label, msg);
		
			// Get font metrics
			twidth = font_get_width(label->specific.label.layout.font, msg, strlen(msg));
			theight = font_get_height(label->specific.label.layout.font, msg);

			if(twidth > *width)
				*width = twidth;

			label->height = theight + 4;
			*height += label->height + list->specific.listbox.inter_widget_space;
		}
			
	}

	if (layout_list != NULL)
		node = layout_list->next;
	else
		node = NULL;

	while (node != NULL)
	{

		label = new_widget(LABEL, list, 0, 0, 1, 1, node->name);
		label->specific.label.set_wrap_type(label, NO_WRAP);
		if(menu_bg_type) {
			label->set_background_color(label, (px_color *) menu_background);
		}
		list->specific.listbox.set_child(list, label, *n);
		label->specific.label.set_font(label, DEFAULT_FONT);
		
		// Set name and callbacks
		label->specific.label.set_text(label, node->name);
		label->callbacks[BUTTON_PRESSED] = ichoose_button_pressed;
		label->callbacks[BUTTON_RELEASED] = ichoose_button_released;
		label->callbacks[MOUSE_ENTER] = ichoose_mouse_enter;
		label->callbacks[MOUSE_LEAVE] = ichoose_mouse_leave;
		
		DBG(debug_info("Created label with %s\n", node->name);)
		// Get font metrics
		twidth = font_get_width(label->specific.label.layout.font, node->name, strlen(node->name));
		theight = font_get_height(label->specific.label.layout.font, node->name);

		if(twidth > *width)
			*width = twidth;

		label->height = theight + 4;
		*height += label->height + list->specific.listbox.inter_widget_space;

		(*n)++;
		node = node->next;
	}

	if(xstroke_available && !strcmp(wm->current_kbd_category,"english")) {
		label = new_widget(LABEL, list, 0, 0, 1, 1, "xstroke");
		label->specific.label.set_wrap_type(label, NO_WRAP);
		if(menu_bg_type) {
			label->set_background_color(label, (px_color *) menu_background);
		}
		list->specific.listbox.set_child(list, label, *n);
		label->specific.label.set_font(label, DEFAULT_FONT);
		
		// Set name and callbacks
		label->specific.label.set_text(label, "xstroke");
		label->callbacks[BUTTON_PRESSED] = ichoose_button_pressed;
		label->callbacks[BUTTON_RELEASED] = ichoose_button_released;
		label->callbacks[MOUSE_ENTER] = ichoose_mouse_enter;
		label->callbacks[MOUSE_LEAVE] = ichoose_mouse_leave;
		
		// Get font metrics
		twidth = font_get_width(label->specific.label.layout.font, "xstroke", strlen("xstroke"));
		theight = font_get_height(label->specific.label.layout.font, "xstroke");

		if(twidth > *width)
			*width = twidth;

		label->height = theight + 4;
		*height += label->height + list->specific.listbox.inter_widget_space;

		(*n)++;
	}

	*width += 4;
	
	for(i=0; i<list->specific.listbox.no_of_children; i++) {
		px_widget *l = list->specific.listbox.widget_list[i];

		l->set_size(l, *width, l->height);
		l->redraw(l);
	}
	
	list->set_size(list, *width, *height);

	return list;
}

// Destroy this menu
void ichoose_window_destroy(window *w)
{
	Display *dpy = w->wm->display;

	XUnmapWindow(dpy, w->frame_window);
	XUngrabPointer(dpy, CurrentTime);
	window_set_state(w, WithdrawnState);

	w->wm->is_menu_mapped = False;

	XGrabServer(dpy);
	window_destroy(w);
	XUngrabServer(dpy);
}

// Click happens here: do not know what to do :)
Bool ichoose_button_pressed(px_widget *pw, px_event *pxe)
{
	px_color pxc;

	pxc.type = SOLID;
	widget_xcolor_from_str(&pxc.first_color, "black");
	pw->set_background_color(pw, &pxc);
	pw->specific.label.set_normal_color(pw, "white");
	pw->redraw(pw);
	pw->show(pw);

	return True;
}

// Pointer released; must switch to the specified window
Bool ichoose_button_released(px_widget *pw, px_event *pxe)
{
	window *w;
	wm* wm;
	window *win;
	
	w = (window *)pxe->private_data;
	wm = w->wm;
	win = wm->mapped_window;
	
	if(win) {
		if(!strcmp(pw->tag,"xstroke")) {
				if(!win->flags.bits.xstroke ||
						win->flags.bits.input) {
					win->set_flag(win, NEEDS_KEYBOARD, False);
					win->set_flag(win, INPUT, False);
					win->set_flag(win, XSTROKE, True);
					win->redraw(win);
					xstroke_start_reconition(wm);
				}
		} else {
				if(win->flags.bits.xstroke ||
						!win->flags.bits.input) {
					win->set_flag(win, XSTROKE, False);
					win->set_flag(win, INPUT, True);
					win->redraw(win);
				}
				xstroke_stop_recognition(wm);
		}
		ichoose_send_message(wm, pw->tag);
	}
	w->flags.bits.needs_keyboard = True; // Set flag for window to be destroyed

	return True;
}

// Pointer enters the label
Bool ichoose_mouse_enter(px_widget *pw, px_event *pxe)
{
	window *w;
	px_color pxc;

	w = (window *)pxe->private_data;
	pxc.type = SOLID;
	widget_xcolor_from_str(&pxc.first_color, "black");
	pw->set_background_color(pw, &pxc);
	pw->specific.label.set_normal_color(pw, "white");
	pw->redraw(pw);
	pw->show(pw);

	return True;
}

// Pointer leaves the label
Bool ichoose_mouse_leave(px_widget *pw, px_event *pxe)
{
	window *w;
	px_color pxc;

	w = (window *)pxe->private_data;

	//widget_xcolor_from_str(&pxc.first_color, "white");
	if(menu_bg_type) {
		pw->set_background_color(pw, (px_color *) menu_background);
	} else {
		pxc.type = TRANSPARENT;
		pw->set_background_color(pw, &pxc);
	}
	pw->specific.label.set_normal_color(pw, "black");
	pw->redraw(pw);
	pw->show(pw);

	return True;
}

// Switch to the specified window
void ichoose_send_message(wm* wm, const char *layout)
{
	window* w = wm->mapped_window;

	if (w->keyboard_layout != NULL)
	{
		free(w->keyboard_layout);
		w->keyboard_layout = NULL;
	}

	w->keyboard_layout = strdup(layout);
	window_handle_keyboard(w);
}

