/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include "alchemy.h"
#include "utils.h"
#include "splash_window.h"
#include "basic_window.h"

// Splash!! A new splash window is conjured
window *new_splash_window(wm *wm, Window map_window)
{
	window *win;
	winlist *winlist_node;
	XWindowAttributes winattr;

	// Allocate and assign common data structures
	win=new_basic_window(wm, map_window, SPLASH_WIN);

	// Set Function pointers
	win->redraw = &splash_window_redraw;
	win->refresh_widgets = &splash_window_refresh_widgets;
	win->set_size = &splash_window_set_size;
	win->set_position = &splash_window_set_position;
	win->show = &splash_window_show;
	win->hide = &splash_window_hide;
	win->iconify = &splash_window_hide;
	win->destroy = &splash_window_destroy;
	win->set_flags = &splash_window_set_flags;
	win->set_flag = &splash_window_set_flag;
	win->app_window = map_window;

	// Figure out window size
	XGetWindowAttributes(wm->display, win->app_window, &winattr);
	win->x = winattr.x;
	win->y = winattr.y;
	win->width = winattr.width;
	win->height = winattr.height;

	if(win->width > (wm->display_width - 5))
		win->width = wm->display_width - 5;
	if((win->height) > (wm->display_height - 30))
		win->height = wm->display_height - 30;

	if((win->width != winattr.width) || (win->height != winattr.height)) {
		XResizeWindow(wm->display, win->app_window, win->width, win->height);
		//window_send_configure_event(win); // Seems un-necessary
	}
	XSetWindowBorderWidth(wm->display, win->app_window, 0);

	// No reparenting!! Heh Heh
	win->frame_window = win->app_window; 

	// Add to the list called on_top_list; this will always ensure
	// that the splash windows stay on top
	// Not for the clock though :)
	if(win->appid != AC_CLOCK_APPID) {
		winlist_node = (winlist *)malloc(sizeof(winlist_node));
		if (!winlist_node)
			err_exit("%s(): Cannot allocate memory", __func__);
		winlist_node->win = win;
		winlist_node->next = wm->on_top_list;
		wm->on_top_list = winlist_node;
	}

	XAddToSaveSet(wm->display, win->app_window);
	XSelectInput(wm->display, win->app_window, ColormapChangeMask|PropertyChangeMask);

	return win;
}

// Nothing to redraw :)
void splash_window_redraw(window *win)
{
}

// Nothing to refresh either :)
void splash_window_refresh_widgets(window *win)
{
}

// Lower this window
void splash_window_hide(window *win)
{
	window_set_state(win, IconicState);
	win->is_mapped = False;
	XLowerWindow(win->wm->display, win->frame_window);
}

// Show the window
void splash_window_show(window *win)
{
	wm *wm=win->wm;
	
	if(win->appid == AC_CLOCK_APPID) {
		if(!wm->mapped_window || (wm->mapped_window && 
			(wm->mapped_window != wm->launcher_window)))
			return;
	}

	window_set_state(win, NormalState);
	
	win->is_mapped = True;
	XMapRaised(win->wm->display, win->frame_window);

	// No Input Focus for splash windows
	//window_handle_focus(win);
}

// Kill this splash
void splash_window_destroy(window *win)
{
	wm *wm=win->wm;
	winlist *winlist_node=wm->on_top_list;

	// Remove from list of on_top windows
	if(winlist_node) {
		if(winlist_node->win == win) {
			// First node
			wm->on_top_list = winlist_node->next;
			free(winlist_node);
		} else {
			// Other nodes in list
			while(winlist_node->next) {
				// Is the next node matching
				if(winlist_node->next->win == win) {
					winlist *temp_node;

					temp_node = winlist_node->next;
					winlist_node->next = temp_node->next;
					free(temp_node);
					break;
				}
				winlist_node = winlist_node->next;
			}
		}
	}

	window_destroy(win);
}

// Set flags ... no action to be taken though
void splash_window_set_flags(window *win, unsigned long flags)
{
	win->flags.all_flags = flags;

	window_set_int_property(win, _AC_W_FLAGS, flags );
}

// Set flag ... no action to be taken though
void splash_window_set_flag(window *win, flag_num flag_no, Bool value)
{
	if(value)
		win->flags.all_flags |= (1<<flag_no);
	else
		win->flags.all_flags &= (~(1<<flag_no));

	window_set_int_property(win, _AC_W_FLAGS, win->flags.all_flags );
}

// Pass through to XResizeWindow
void splash_window_set_size(window *win, unsigned long width, unsigned long height)
{
	win->width=width;
	win->height=height;
	XResizeWindow(win->wm->display, win->frame_window, width, height);
	window_send_configure_event(win);
}

// Pass through to XMoveWindow
void splash_window_set_position(window *win, int x, int y)
{
	win->x=x;
	win->y=y;
	XMoveWindow(win->wm->display, win->frame_window, x, y);
}
