/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#ifndef _THEME_H_
#define _THEME_H_

#include <picowidget.h>

/*Various button states supported in the theme */
#define NORMAL_BUTTON   	0
#define PRESSED_BUTTON  	1
#define RELEASED_BUTTON 	2
#define DISABLED_BUTTON 	3
#define NO_OF_BUTTON_STATES 	4

/*Prev and next page buttons, part of page navigation.
 *Other buttons are defined in the flag_num enumeration i
 * alchemy.h
 * */
#define SEPARATOR 		100
#define PREV_PAGE 		101
#define NEXT_PAGE 		102


#define THEME_PATH_SIZE         100


/* Typedefs */
typedef struct _theme_top_level_frame   theme_top_level_frame;
typedef struct _theme_box               theme_box;
typedef struct _theme_button            theme_button;
typedef struct _theme_button_state      theme_button_state;
typedef struct _theme_label             theme_label;
typedef struct _theme_pixmap		theme_pixmap;

typedef enum _top_level_frame_type{
	TLF_TOPBAR,
	TLF_BOTTOMBAR,
	TLF_DIALOG,
} top_level_frame_type;	

#define NON_TLF 10

typedef enum _theme_label_type{
	LABEL_TITLE=50,
	LABEL_STATUS_MSG,
	LABEL_PAGE_NUM,
}theme_label_type;	

/*Top-level frames.
 * These top level frames will contain some list of boxes*/
struct _theme_top_level_frame{
	top_level_frame_type frame_type; 
	unsigned long height;
	theme_box *boxes; 
	theme_top_level_frame *next;
};	

struct _theme_button_state{
	char *image; /*Path to the pixmap. NULL value for "Released" state*/
	char *mask; /* path to the maks (used to check if an even has to be processed).
			NULL value for "Released" state*/
	char *audio; /* Wav file to play. NULL value for "Normal" or "Disabled" */
};	

/*The button structure*/
struct _theme_button{
	unsigned char btype; /* Type of the button, as defined by the flag_num enumeration in alchemy.h*/	
	long x; /*Horizontal position*/
	long y; /*Vertical position*/
	unsigned long width;
	unsigned long height;
	theme_button_state state[NO_OF_BUTTON_STATES]; /* Normal, Pressed, Released, Disabled*/
	theme_button *next;
};


struct _theme_label{
	theme_label_type ltype;
	long x; /*Horizontal position*/
	long y; /*Vertical position*/
	unsigned long width; /* Width of the box*/
	unsigned long height; /* Height of the box*/
	char *font; 
	theme_label *next;
};

/* Width of the box can be specified either as an absolute value
 * of in terms of the percentage of the parent window*/
typedef enum _width_type{
	PERCENTAGE_WIDTH,
	ABSOLUTE_WIDTH
} width_type;	

struct _width{
	width_type wtype;
	unsigned long value;
};	

/*The backgound of a widget can either be a pixmap or colour*/
typedef enum _theme_bg_type{
	PIXMAP,
	COLOR
}theme_bg_type;	


/*Background information for different theme widgets*/
typedef union theme_background{
	struct _pmap{	
		char *pixmap_h; /* path for the pixmap in vertical position*/
		char *pixmap_v; /* path for the pixmap in horizontal position*/
	}pmap;	
	struct _color{
		char *fill_type; /* Way to fill the background, e.g., solid, gradientvert, gradienthoriz */
		char *color1;  /* Color to fill */
		char *color2;  /* Color to fill..required if fill_type is gradient */
	}color;	
}theme_background;	


/* A box may contain some buttons, labels or some embedded boxes */
struct _theme_box{
	long x; /* Horizontal position*/
	long y; /* Vertical position */
	struct _width width;
	unsigned long height;
	theme_bg_type bg_type; /* Specifies the type of the background. It's either "pixmap" or "color" */
	theme_background background; /*The background for the box*/
	
	//theme_box     *box_list; /* A list of embedded boxes */
	theme_button  *button_list; /* List of buttons inside the box*/
	theme_label   *label_list;  /*List of labels inside the box*/
	theme_box *next;
};	

/*Node for the binary tree structure containing the theme*/
struct _theme_pixmap{
	char *path; /*Path for the pixmap*/
	px_pixmap *background; /* px_pixmap structure defined in picoxt.h */
	theme_pixmap *left; /* Left pointer of a node*/
	theme_pixmap *right; /* Right pointer of a node*/
};

typedef struct _theme_menu{
	theme_bg_type bg_type; /* Type of the background for the menu*/
	theme_background background; /*The background for the menu*/
}theme_menu;	

#endif
