/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

/*
 *Parser for the XML theme file
 */

#include <expat.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "themeparse.h"
#include "utils.h"
#include "basic_window.h"
#include "dialog_window.h"
#include <font.h>
#include <image.h>
#define TRUE  1
#define FALSE 0

//#define THEME_DEBUG

extern Display *display;

static int theme_path_offset = 0;
	
/* Allocate the ac_theme structure */
ac_theme *new_theme(wm *wm)
{
	ac_theme *theme;

 	if (!(theme = (ac_theme *)malloc(sizeof(ac_theme)))){
		err_exit("Not enough memory\n");
	}	

	theme->top_level_frame = NULL;
	theme->theme_path = NULL;
	theme->menu = NULL;
	theme->pixmap_tree = NULL;
	
	theme->load_theme = load_ac_theme;
	theme->free_theme = free_ac_theme;
	theme->render_theme = render_ac_theme;
	theme->cleanup_theme = cleanup_ac_theme;
	theme->get_topbar_height = get_ac_topbar_height ;
	theme->get_bottombar_height = get_ac_bottombar_height ;
	theme->get_dialogbar_height = get_ac_dialogbar_height ;
	theme->get_menu_background = get_ac_menu_background;
	theme->get_menu_x = get_ac_menu_x;
	theme->wm = wm;

	return theme;
}

/* Parse the XML theme file and populate the ac_theme structure.
 *
 * The theme_path is the name of a directory. We have
 * assumed that the actual theme file is "ac.theme" under
 * the directory specified in the parameter "theme_path"
 * */
void load_ac_theme(ac_theme *self, const char *theme_path, Window root_win)
{
	FILE *conf_fp;
	char buff[THEME_BUFFSIZE];
	theme_parser_data *parser_data; /* Local parser data used to send it to the handler*/
	XML_Parser parser = NULL ;
	unsigned char backslash_in_theme_path = TRUE; /* A flag to know whether the "theme_path" contains
				a terminating back-slash */
	char *theme_file; /* The absolute theme file name created by appending "ac.theme"
				to the theme directory in "theme_path" parameter */
	int theme_dir_len;

	
	if (!(parser_data = (theme_parser_data *)malloc(sizeof(theme_parser_data)))){
		err_exit("Not enough memory\n");
	}	
	
	parser_data->theme = self;
	parser_data->depth = 0;  /* Depth is initially 0*/
	self->win = root_win;
	
	if (!parser)	
		parser = XML_ParserCreate(NULL);
	if (!parser){
		err_exit("Couldn't allocate memory for parser\n");
	}	
  	XML_SetUserData(parser, parser_data); /* Send "parser_data" to the handlers */
  	XML_SetElementHandler(parser, start_tag_handler, end_tag_handler);

	theme_dir_len = strlen(theme_path);
	if (theme_path[theme_dir_len - 1] != '/'){
		backslash_in_theme_path = FALSE;
		theme_dir_len += 1; /* One more byte to accomodate the backslash character*/
	}	

	theme_path_offset = parser_data->theme_dir_len =  theme_dir_len;
	/* Store the theme path in the structure*/
	self->theme_path = (char *)malloc(theme_dir_len + 1);
	if (!self->theme_path){
		err_exit("Not enough memory\n");
	}	

	strcpy(self->theme_path, theme_path); /* Theme path is locally saved in the ac_theme structure*/
	if (!backslash_in_theme_path){
		strcat(self->theme_path, "/");
	}	

	/*Make the absolute theme file name*/
	theme_file = (char *)malloc(theme_dir_len + strlen("ac.theme") + 1);
	if (!theme_file){
		err_exit("Not enough memory\n");
	}	
	strcpy(theme_file, self->theme_path);
	strcat(theme_file, "ac.theme");


	if (!(conf_fp = fopen(theme_file, "r"))){
		err_exit("%s: Couldn't open XML theme file\n", theme_file);
	}		
		
	for (;;){
		int done;
		int len;
		
    		len = fread(buff, 1, THEME_BUFFSIZE, conf_fp);
		done = len < sizeof(buff); /* Nothing more to read */
		
		/* Parse the content in the buffer*/
    		if (!XML_Parse(parser, buff, len, done)) {
      			fprintf(stderr, "%s at line %d in %s\n",
	      			XML_ErrorString(XML_GetErrorCode(parser)),
	      			XML_GetCurrentLineNumber(parser), theme_file);
      			exit(-1);
		}	

    		if (done)
      			break;
	}	
	/*Free various resources*/
	XML_ParserFree(parser);
	free(parser_data);
	free(theme_file);
	fclose(conf_fp);
}

/*Handler for the start tag*/
void start_tag_handler(void *user_data, const char *tag_name, const char **attr){
	theme_parser_data *parser_data = user_data;
#if 0 
	for (i = 0 ; i < depth; i++)
		printf("\t");
	printf("%s", tag_name);
	
  	for (i = 0; attr[i]; i += 2) {
    		printf(" %s=%s", attr[i], attr[i + 1]);
  	}
  	printf("\n");
#endif	
	parse_start_tag(parser_data, tag_name, attr );
  	(parser_data->depth)++;
}

/*Handler for the end tag*/
void end_tag_handler(void *user_data, const char *name)
{
	theme_parser_data *parser_data = user_data;
  	(parser_data->depth)--;
}

/*The width of a box can be either an absolute value or specified in % value*/
void set_box_width(theme_box *p_box, const char *width_string)
{
	if (!width_string){
		/* This is actually a wrong condition. The width shouldn't be zero.*/
		p_box->width.wtype = ABSOLUTE_WIDTH;
		p_box->width.value = 0;
	}else {
		int len = strlen(width_string);
		if (width_string[len - 1] == '%'){
			char *str_without_pc = NULL;
			str_without_pc = (char *)malloc(len);
			if (!str_without_pc)
				err_exit("Not enough memory\n");
			strncpy(str_without_pc, width_string, len - 1);
			str_without_pc[len - 1] = '\0'; 
			p_box->width.wtype = PERCENTAGE_WIDTH;
			p_box->width.value = strtoul(str_without_pc, NULL, 10);
			free(str_without_pc);
			
		}
		else{
			/*The width is specified in absolute values*/
			p_box->width.wtype = ABSOLUTE_WIDTH;
			p_box->width.value = strtoul(width_string, NULL, 10);
		}	
	}	
}

/* Parse the start tag and its attributes and populate the structures defined in theme.h */
void parse_start_tag(theme_parser_data *pdata,  const char *tag_name, const char **attr)
{
	switch (*tag_name) {
		case 'a':
			/* <annotate> */
			if (!strcmp(tag_name, "ac_theme")){
				if (pdata->depth != 0){
					err_exit("%s: <ac_theme> can be used only as the begin tag of the theme file\n");
				}	
			}	
			else if (!strcmp(tag_name, "annotate")){
				if (pdata->depth < 3){
					err_exit("%s:<annotate> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "annotate_present")){
				if (pdata->depth < 3){
					err_exit("%s:<annotate_present> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'b':
			/* <bottombar>, <box>*/
			if (!strcmp(tag_name, "bottombar")){
				/*The <bottombar> tag is expected only at the depth 1 of the XML theme file*/
				if (pdata->depth < 1){
						err_exit("%s: <bottombar> tag must be inside <ac_theme> tag\n",
							pdata->theme->theme_path);
					
				} else	if (pdata->depth == 1)	{
					theme_top_level_frame *ptr;
					int i;
					
					ptr = (theme_top_level_frame *)malloc(sizeof(theme_top_level_frame));
					if (!ptr){
						err_exit("Not enough memory\n");
					}	
					ptr->boxes = NULL;
					ptr->frame_type = TLF_BOTTOMBAR; 
					for (i = 0; attr[i]; i += 2){
						if (!strcmp(attr[i], "height"))
							ptr->height = strtoul(attr[i+1], NULL, 10);
					}	

					/*  Adding an element to the head of the linked list*/
					ptr->next = pdata->theme->top_level_frame ;
					pdata->theme->top_level_frame = ptr;
				}
				else {
					err_exit("%s:Embedded <bottombar> tag is not supported\n",
							pdata->theme->theme_path);
				}	
			}	
			else if (!strcmp(tag_name, "box")){
				if (pdata->depth < 2){
					/* Error*/	
					err_exit("%s:<box> tag must be inside a <topbar> or <bottombar>\n",
							pdata->theme->theme_path);
				}
				else {
					theme_box *ptr;
					int i;

					if (!(ptr = (theme_box *)malloc(sizeof(theme_box)))){
						err_exit("Not enough memory\n");
					}
					
					ptr->button_list = NULL;
					ptr->label_list = NULL;	
					ptr->background.color.color1 = NULL;
					ptr->background.color.color2 = NULL;
					ptr->background.color.fill_type  = NULL;
					ptr->background.pmap.pixmap_h = NULL;
					ptr->background.pmap.pixmap_v = NULL;
						
					for (i = 0; attr[i]; i += 2){
						if (!strcmp(attr[i], "x"))
							ptr->x = strtoul(attr[i + 1], NULL, 10);
						else if (!strcmp(attr[i], "y"))
							ptr->y = strtoul(attr[i + 1], NULL, 10);
						else if (!strcmp(attr[i], "width"))
							set_box_width(ptr, attr[i+1]);
						else if (!strcmp(attr[i], "height"))
							ptr->height = strtoul(attr[i + 1], NULL, 10);
						else if (!strcmp(attr[i], "bg_type")){
							if (!strcmp(attr[i+1], "pixmap"))
								ptr->bg_type = PIXMAP;
							else if ((!strcmp(attr[i+1], "colour")) || (!strcmp(attr[i+1], "color")))
								ptr->bg_type = COLOR;
							else 
								err_exit("%s: Unsupported background type\n", attr[i+1]);
						}	
						else if (!strcmp(attr[i], "pixmap_v")){
							/* Read the pixmap type only if bg_type has value "pixmap" */
							if (ptr->bg_type == PIXMAP){
								ptr->background.pmap.pixmap_v
									= (char *)malloc(pdata->theme_dir_len +
											strlen(attr[i+1]) + 1);
								if (!ptr->background.pmap.pixmap_v){
									err_exit("Not enough memory\n");
								}	
								strcpy(ptr->background.pmap.pixmap_v, pdata->theme->theme_path);
								strcat(ptr->background.pmap.pixmap_v, attr[i+1]);
								add_to_pixmap_tree(pdata->theme->win, 
								  &(pdata->theme->pixmap_tree), ptr->background.pmap.pixmap_v);
							}
						}
						else if (!strcmp(attr[i], "pixmap_h")){
							/* Read the pixmap type only if bg_type has value "pixmap" */
							if (ptr->bg_type == PIXMAP){
								ptr->background.pmap.pixmap_h
									= (char *)malloc(pdata->theme_dir_len +
											strlen(attr[i+1]) + 1);
								if (!ptr->background.pmap.pixmap_h){
									err_exit("Not enough memory\n");
								}	
								strcpy(ptr->background.pmap.pixmap_h, pdata->theme->theme_path);
								strcat(ptr->background.pmap.pixmap_h, attr[i+1]);
								add_to_pixmap_tree(pdata->theme->win, 
								  &(pdata->theme->pixmap_tree), ptr->background.pmap.pixmap_h);
							}
						}
						else if(!strcmp(attr[i], "fill_type")){	
							/* Read the color type only if bg_type attribute has value "color" */
							if (ptr->bg_type == COLOR){
								ptr->background.color.fill_type = strdup(attr[i+1]);
								if (!ptr->background.color.fill_type)
									err_exit("Not enough memory\n");
							}	
						}
						else if(!strcmp(attr[i], "color1")){	
							/* Read the color1 type only if bg_type attribute has value "color" */
							if (ptr->bg_type == COLOR){
								ptr->background.color.color1 = strdup(attr[i+1]);
								if (!ptr->background.color.color1)
									err_exit("Not enough memory\n");
							}	
						}
						else if(!strcmp(attr[i], "color2")){	
							/* Read the color2 type only if bg_type attribute has value "color" 
							 * and fill_type has value "gradient"*/
							if ((ptr->bg_type == COLOR) && 
							           (is_gradient_string(ptr->background.color.fill_type))){
								ptr->background.color.color2 = strdup(attr[i+1]);
								if (!ptr->background.color.color2)
									err_exit("Not enough memory\n");
							}	
						}
						else {
							err_exit("%s: Box attribute not supported\n", attr[i]);
						}		
					}	
					//ptr->box_list = NULL; /*Embedded boxes are not supported at present*/
					
					if (pdata->depth == 2){
						/* A top level frame(topbar or bottombar) can have a list of boxes. 
					 	* So we maintain a linked list of these top-level boxes here.*/

						/* Adding the box to the head of the linked list*/
						ptr->next = pdata->theme->top_level_frame->boxes;
						pdata->theme->top_level_frame->boxes = ptr;
					}
					else {
				  		/* Embedded box. This box is inside another box. So we
				   		* add these to the box list of the parent box*/	

						/* IMPORANT NOTE: Only one level of embedded box is supported
					 	* by this code. */
						/*ptr->next = pdata->theme->top_level_frame->boxes->box_list;
						pdata->theme->top_level_frame->boxes->box_list = ptr;*/
					}	
				}	
			}
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'c':
			/* <close> */
			if (!strcmp(tag_name, "close")){
				if (pdata->depth < 3){
					err_exit("%s:<close> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'd':
			/* <disabled>*/
			if (!strcmp(tag_name, "disabled")){
				int i;
				theme_button *button_list
					= pdata->theme->top_level_frame->boxes->button_list;

				for (i = 0; attr[i]; i += 2){
					if (!strcmp(attr[i], "image")){
						char *tmp = NULL;
						tmp = (char *)malloc(pdata->theme_dir_len +strlen(attr[i+1]) + 1);
						if (!tmp){
							err_exit("Not enough memory\n");
						}	
						strcpy(tmp, pdata->theme->theme_path);
						strcat(tmp, attr[i+1]);
						add_to_pixmap_tree(pdata->theme->win, &(pdata->theme->pixmap_tree),
							tmp);
						button_list->state[DISABLED_BUTTON].image = tmp;
					}	
					else if (!strcmp(attr[i], "mask")){
						char *tmp = NULL;
						if (attr[i+1]){
							tmp = (char *)malloc(pdata->theme_dir_len +strlen(attr[i+1]) + 1);
							if (!tmp){
								err_exit("Not enough memory\n");
							}	
							strcpy(tmp, pdata->theme->theme_path);
							strcat(tmp, attr[i+1]);
						}
						button_list->state[DISABLED_BUTTON].mask = tmp;
					}	
				}
			}	
			else if (!strcmp(tag_name, "dialog")){
				/*The <dialog> tag is expected only at the depth 1 of the XML theme file*/
				if (pdata->depth < 1)	{
						err_exit("%s: <dialog> tag must be inside <ac_theme> tag\n",
							pdata->theme->theme_path);
					
				}else if (pdata->depth == 1)	{
					theme_top_level_frame *ptr;
					int i;
					ptr = (theme_top_level_frame *)malloc(sizeof(theme_top_level_frame));
					if (!ptr){
						err_exit("Not enough memory\n");
					}	
					ptr->boxes = NULL;
					ptr->frame_type = TLF_DIALOG; 
					for (i = 0; attr[i]; i += 2){
						if (!strcmp(attr[i], "height"))
							ptr->height = strtoul(attr[i+1], NULL, 10);
					}	

					ptr->next = pdata->theme->top_level_frame ;
					pdata->theme->top_level_frame = ptr;

				}
				else {
					err_exit("%s:Embedded <dialog> tag is not supported\n",
							pdata->theme->theme_path);
				}	

			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'e':
			/* <email> */
			if (!strcmp(tag_name, "email")){
				if (pdata->depth < 3){
					err_exit("%s:<email> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'h':
			if (!strcmp(tag_name, "help")){
				if (pdata->depth < 3){
					err_exit("%s:<help> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'i':
			/* <input> */
			if (!strcmp(tag_name, "input")){
				if (pdata->depth < 3){
					err_exit("%s:<input> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "input_choose")){
				if (pdata->depth < 3){
					err_exit("%s:<input_choose> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'l':
			if (!strcmp(tag_name, "launcher")){
				if (pdata->depth < 3){
					err_exit("%s:<launcher> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "left")){
				if (pdata->depth < 3){
					err_exit("%s:<left> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'm':
			if (!strcmp(tag_name, "menu")){
				theme_menu *ptr = NULL;
				int i;

				if (!(ptr = (theme_menu *)malloc(sizeof(theme_menu)))){
					err_exit("Not enough memory\n");
				}
				ptr->background.color.color1 = NULL;
				ptr->background.color.color2 = NULL;
				ptr->background.color.fill_type  = NULL;
				ptr->background.pmap.pixmap_h = NULL;
				ptr->background.pmap.pixmap_v = NULL;
				for (i = 0; attr[i]; i += 2){
					if (!strcmp(attr[i], "bg_type")){
						if (!strcmp(attr[i+1], "pixmap"))
							ptr->bg_type = PIXMAP;
						else if ((!strcmp(attr[i+1], "colour")) || (!strcmp(attr[i+1], "color")))
							ptr->bg_type = COLOR;
						else 
							err_exit("%s: Unsupported background type\n", attr[i+1]);
					}	
					else if (!strcmp(attr[i], "pixmap")){
						/* Read the pixmap type only if bg_type has value "pixmap" */
						if (ptr->bg_type == PIXMAP){
							ptr->background.pmap.pixmap_v
								= (char *)malloc(pdata->theme_dir_len +
										strlen(attr[i+1]) + 1);
							if (!ptr->background.pmap.pixmap_v){
								err_exit("Not enough memory\n");
							}	
							strcpy(ptr->background.pmap.pixmap_v, pdata->theme->theme_path);
							strcat(ptr->background.pmap.pixmap_v, attr[i+1]);
						}
					}
					else if(!strcmp(attr[i], "fill_type")){	
						/* Read the color type only if bg_type attribute has value "color" */
						if (ptr->bg_type == COLOR){
							ptr->background.color.fill_type = strdup(attr[i+1]);
							if (!ptr->background.color.fill_type)
								err_exit("Not enough memory\n");
						}	
					}
					else if(!strcmp(attr[i], "color1")){	
						/* Read the color1 type only if bg_type attribute has value "color" */
						if (ptr->bg_type == COLOR){
							ptr->background.color.color1 = strdup(attr[i+1]);
							if (!ptr->background.color.color1)
								err_exit("Not enough memory\n");
						}	
					}
					else if(!strcmp(attr[i], "color2")){	
						/* Read the color2 type only if bg_type attribute has value "color" 
						 * and fill_type has value "gradient"*/
						if ((ptr->bg_type == COLOR) && 
						             (!strcmp(ptr->background.color.fill_type, "gradient"))){
							ptr->background.color.color2 = strdup(attr[i+1]);
							if (!ptr->background.color.color2)
								err_exit("Not enough memory\n");
						}	
					}
				}
				pdata->theme->menu = ptr;
				
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'n':
			/* <normal> */
			if (!strcmp(tag_name, "normal")){
				int i;
				theme_button *button_list
					= pdata->theme->top_level_frame->boxes->button_list;
				
				for (i = 0; attr[i]; i += 2){
					if (!strcmp(attr[i], "image")){
						char *tmp = NULL;
						if (attr[i+1]){
							tmp= (char *)malloc(pdata->theme_dir_len +strlen(attr[i+1]) + 1);
							if (!tmp){
								err_exit("Not enough memory\n");
							}	
							strcpy(tmp, pdata->theme->theme_path);
							strcat(tmp, attr[i+1]);
							add_to_pixmap_tree(pdata->theme->win, &(pdata->theme->pixmap_tree),
								tmp);
							button_list->state[NORMAL_BUTTON].image = tmp;
						}	
					}	
					else if (!strcmp(attr[i], "mask")){
						char *tmp = NULL;
						if (attr[i+1]){
							tmp = (char *)malloc(pdata->theme_dir_len +strlen(attr[i+1]) + 1);
							if (!tmp){
								err_exit("Not enough memory\n");
							}	
							strcpy(tmp, pdata->theme->theme_path);
							strcat(tmp, attr[i+1]);
							button_list->state[NORMAL_BUTTON].mask = tmp;
						}	
					}	
				}	
			}	
			else if (!strcmp(tag_name, "next")){
				if (pdata->depth < 3){
					err_exit("%s:<next> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "next_page")){
				if (pdata->depth < 3){
					err_exit("%s:<next_page> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'o':	
			if (!strcmp(tag_name, "onetap")){
				if (pdata->depth < 3){
					err_exit("%s:<onetap> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'p':
			/* <pressed>, <page_nav> */
			if (!strcmp(tag_name, "pressed")) {
				int i;
				theme_button *button_list
					= pdata->theme->top_level_frame->boxes->button_list;

				for (i = 0; attr[i]; i += 2){
					if (!strcmp(attr[i], "image")){
						char *tmp = NULL;
						if (attr[i+1]){
							tmp = (char *)malloc(pdata->theme_dir_len +strlen(attr[i+1]) + 1);
							if (!tmp){
								err_exit("Not enough memory\n");
							}	
							strcpy(tmp, pdata->theme->theme_path);
							strcat(tmp, attr[i+1]);
							add_to_pixmap_tree(pdata->theme->win, &(pdata->theme->pixmap_tree),
								tmp);
							button_list->state[PRESSED_BUTTON].image = tmp;
						}
					}	
					else if (!strcmp(attr[i], "mask")){
						char *tmp = NULL;
						if (attr[i+1]){
							tmp = (char *)malloc(pdata->theme_dir_len +strlen(attr[i+1]) + 1);
							if (!tmp){
								err_exit("Not enough memory\n");
							}	
							strcpy(tmp, pdata->theme->theme_path);
							strcat(tmp, attr[i+1]);
							button_list->state[PRESSED_BUTTON].mask = tmp;
						}
					}	
					else if (!strcmp(attr[i], "audio")){
						if (attr[i+1]){
							char *tmp = NULL;
							tmp = (char *)malloc(pdata->theme_dir_len +strlen(attr[i+1]) + 1);
							if (!tmp){
								err_exit("Not enough memory\n");
							}	
							strcpy(tmp, pdata->theme->theme_path);
							strcat(tmp, attr[i+1]);
							button_list->state[PRESSED_BUTTON].audio = tmp;
						}
					}	
				}	
			}
			else if (!strcmp(tag_name, "page_nav")) {
				if (pdata->depth < 3){
					err_exit("%s:<page_nav> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "page_num")){
				if (pdata->depth < 3){
					err_exit("%s:<page_num> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_label(tag_name, pdata, attr);
				}
			}	
			else if (!strcmp(tag_name, "printer")){
				if (pdata->depth < 3){
					err_exit("%s:<printer> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "prev")){
				if (pdata->depth < 3){
					err_exit("%s:<prev> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "prev_page")){
				if (pdata->depth < 3){
					err_exit("%s:<prev_page> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "pref")){
				if (pdata->depth < 3){
					err_exit("%s:<pref> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'r':
			if (!strcmp(tag_name, "retractbar")){
				if (pdata->depth < 3){
					err_exit("%s:<retractbar> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}
			}	
			else if (!strcmp(tag_name, "right")){
				if (pdata->depth < 3){
					err_exit("%s:<right> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}
			}	
			else if (!strcmp(tag_name, "released")) {
				int i;
				theme_button *button_list
					= pdata->theme->top_level_frame->boxes->button_list;

				for (i = 0; attr[i]; i += 2){
					if (!strcmp(attr[i], "audio")){
						char *tmp = (char *)malloc(pdata->theme_dir_len +strlen(attr[i+1]) + 1);
						if (!tmp){
							err_exit("Not enough memory\n");
						}	
						strcpy(tmp, pdata->theme->theme_path);
						strcat(tmp, attr[i+1]);
						button_list->state[RELEASED_BUTTON].audio = tmp;
					}	
				}
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 's':
			/* <status> */
			if (!strcmp(tag_name, "status")){
				if (pdata->depth < 3){
					/* Error*/
					err_exit("%s:<status> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "status_msg")){
				if (pdata->depth < 3){
					err_exit("%s:<status_msg> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_label(tag_name, pdata, attr);
				}
			}	
			else if (!strcmp(tag_name, "star")){
				if (pdata->depth < 3){
					err_exit("%s:<star> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "seperator")){
				if (pdata->depth < 3){
					err_exit("%s:<seperator> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "speaker")){
				if (pdata->depth < 3){
					err_exit("%s:<speaker> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 't':	
			/* <title>, <topbar> */
			if (!strcmp(tag_name, "topbar")) {
				/*The <topbar> tag is expected only at the depth 1 of the XML theme file*/
				if (pdata->depth < 1)	{
						err_exit("%s: <topbar> tag must be inside <ac_theme> tag\n",
							pdata->theme->theme_path);
					
				}else if (pdata->depth == 1)	{
					theme_top_level_frame *ptr;
					int i;
					ptr = (theme_top_level_frame *)malloc(sizeof(theme_top_level_frame));
					if (!ptr){
						err_exit("Not enough memory\n");
					}	
					ptr->boxes = NULL;
					ptr->frame_type = TLF_TOPBAR; 
					for (i = 0; attr[i]; i += 2){
						if (!strcmp(attr[i], "height"))
							ptr->height = strtoul(attr[i+1], NULL, 10);
					}	

					ptr->next = pdata->theme->top_level_frame ;
					pdata->theme->top_level_frame = ptr;

				}
				else {
					err_exit("%s:Embedded <topbar> tag is not supported\n",
							pdata->theme->theme_path);
				}	
			}
			else if (!strcmp(tag_name, "title")) {
				if (pdata->depth < 3){
					/* Error*/
					err_exit("%s:<title> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_label(tag_name, pdata, attr);
				}	
			}	
			else if (!strcmp(tag_name, "tick")){
				if (pdata->depth < 3){
					/* Error*/
					err_exit("%s:<tick> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'u':
			/* <up> */
			if (!strcmp(tag_name, "up")){
				if (pdata->depth < 3){
					/* Error*/
					err_exit("%s:<up> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'v':
			/* <voice> */
			if (!strcmp(tag_name, "voice")){
				if (pdata->depth < 3){
					/* Error*/
					err_exit("%s:<voice> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				}	
				else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		case 'x':
			/* <xstroke> */
			if (!strcmp(tag_name, "xstroke")){
				if (pdata->depth < 3){
					/* Error*/
					err_exit("%s:<xstroke> tag must be inside a <box>\n",
							pdata->theme->theme_path);
				} else {
					parse_button(tag_name, pdata, attr);
				}	
			}	
			else {
				err_exit("%s: Unsupported tag\n", tag_name);
				break;
			}	
			break;
		default:
			/* Unsupported tag */
			err_exit("%s: Unsupported tag\n", tag_name);
			break;
	}	
}

/* Hello label, what's your type? */
theme_label_type get_label_type(const char *label_num)
{
	if (!strcmp(label_num, "title"))
		return LABEL_TITLE;
	else if (!strcmp(label_num, "status_msg"))
		return LABEL_STATUS_MSG;
	else if (!strcmp(label_num, "page_num"))
		return LABEL_PAGE_NUM;
	return -2;
}

/* Allocate the label structure and set its attributes*/
void parse_label(const char *label_name, theme_parser_data *pdata, const char **attr)
{
	theme_label *ptr = NULL;
	int i;

#ifdef THEME_DEBUG	
	printf("label: %s\n", label_name);
#endif	
	if  (pdata->theme->top_level_frame->frame_type == TLF_DIALOG){
		if (strcmp(label_name, "title"))
			err_exit("<%s> tag can't be inside a <dialog>\n", label_name);
	}		
	
	ptr = (theme_label *)malloc(sizeof(theme_label));
	if (!ptr){
		err_exit("Not enough memory\n");
	}	
	ptr->ltype = get_label_type(label_name);
	if (ptr->ltype < 0)
		err_exit("%s: Tag not supported\n", label_name);

	for (i = 0; attr[i]; i += 2){
		if (!strcmp(attr[i], "x"))
			ptr->x = strtoul(attr[i+1], NULL, 10);
		else if (!strcmp(attr[i], "y"))
			ptr->y = strtoul(attr[i+1], NULL, 10);
		else if (!strcmp(attr[i], "width"))
			ptr->width = strtoul(attr[i+1], NULL, 10);
		else if (!strcmp(attr[i], "height"))
			ptr->height = strtoul(attr[i+1], NULL, 10);
		else if (!strcmp(attr[i], "font")){
			int len = 0;
			
			if (attr[i+1])
				len = strlen(attr[i+1]);
			ptr->font = (char *)malloc(len + 1);
			if (!ptr->font)
				err_exit("Not enough memory\n");
			strcpy(ptr->font, attr[i+1]);	
		}	
	}	
	if (pdata->depth == 3){
		ptr->next = pdata->theme->top_level_frame->boxes->label_list;
		pdata->theme->top_level_frame->boxes->label_list = ptr;
	}
	else {
		//To be implemented later
	}	
}

/* Allocate the button structure and set its attributes*/
void parse_button(const char *button_name, theme_parser_data *pdata, const char **attr)
{
	/* Allocate space for the button information*/
	theme_button *ptr;
	int i;
	
	if (!(ptr = (theme_button *)malloc(sizeof(theme_button)))){
		err_exit("Not enough memory\n");
	}	
	ptr->btype = get_button_type(button_name);
	for (i = 0; i < NO_OF_BUTTON_STATES; i++){
		ptr->state[i].image = NULL;
		ptr->state[i].mask = NULL;
		ptr->state[i].audio = NULL;
	}	
		
	for (i = 0; attr[i]; i += 2){
		if (!strcmp(attr[i], "x")){
			ptr->x = strtoul(attr[i+1], NULL, 10);
			if (!strcmp(button_name, "star"))
					pdata->theme->menu_x = ptr->x;//the star menu will have the same x-pixel
													//value as that of the star button 
		}	
		else if (!strcmp(attr[i], "y"))
			ptr->y = strtoul(attr[i+1], NULL, 10);
		else if (!strcmp(attr[i], "width")){
			ptr->width = strtoul(attr[i+1], NULL, 10);
		}	
		else if (!strcmp(attr[i], "height"))
			ptr->height = strtoul(attr[i+1], NULL, 10);
	}	

	if (pdata->depth == 3){
		ptr->next = pdata->theme->top_level_frame->boxes->button_list;
		pdata->theme->top_level_frame->boxes->button_list = ptr;
	}
	else {
		/*To be implemented later*/
	}	
}

/* Return the button type corresponding to the "button"
 * parameter. The button_type is defined in the flag_num
 * enumeration in alchemy.h and also NEXT_PAGE and PREV_PAGE
 * #defines in themeparse.h*/
unsigned char get_button_type(const char *button)
{
	switch (*button){
		case 'a':
			if (!strcmp(button, "annotate"))
				return ANNOTATE;
			else if (!strcmp(button, "annotate_present"))
				return ANNOTATION_PRESENT;
			break;
		case 'c':
			if (!strcmp(button, "close"))
				return CLOSE;
			break;
		case 'e':
			if (!strcmp(button, "email"))
				return EMAIL;
			break;
		case 'h':
			if (!strcmp(button, "help"))
				return HELP;
			break;
		case 'i':
			if (!strcmp(button, "input"))
				return INPUT;
			else if (!strcmp(button, "input_choose"))
				return INPUT_CHOOSE;
			break;
		case 'l':
			if (!strcmp(button, "launcher"))
				return LAUNCHER;
			else if (!strcmp(button, "left"))
				return LEFT;
			break;
		case 'n':
			if (!strcmp(button, "next"))
				return NEXT;
			else if (!strcmp(button, "next_page"))
				return NEXT_PAGE;
			break;
		case 'o':
			if (!strcmp(button, "onetap"))
				return ONETAP;
			break;
		case 'r':
			if (!strcmp(button, "retractbar"))
				return RETRACTBAR;
			else if (!strcmp(button, "right"))
				return RIGHT;
			break;
		case 's':
			if (!strcmp(button, "status"))
				return STATUS;
			else if (!strcmp(button, "star"))
				return STAR;
			else if (!strcmp(button, "seperator"))
				return SEPARATOR;
			else if (!strcmp(button, "speaker"))
				return SPEAKER;
			break;
		case 'p':	
			if (!strcmp(button, "printer"))
				return PRINTER;
			else if (!strcmp(button, "prev"))
				return PREV;
			else if (!strcmp(button, "pref"))
				return PREF;
			else if (!strcmp(button, "page_nav"))
				return PAGE_NAV;
			else if (!strcmp(button, "prev_page"))
				return PREV_PAGE;
			break;
		case 't':	
			if (!strcmp(button, "tick"))
				return TICK;
			break;
		case 'u':	
			if (!strcmp(button, "up"))
				return UP;
			break;
		case 'v':	
			if (!strcmp(button, "voice"))
				return VOICE;
			break;
		case 'x':	
			if (!strcmp(button, "xstroke"))
				return XSTROKE;
			break;
		default:
			err_exit("%s:Tag not supported in the theme file", button);
			break;
	}	
	return 0;
}	

/* Render the theme. 
 *
 * First read from the structures defined in theme.h and then render them.
 */
void render_ac_theme(ac_theme *self, window *win)
{
	theme_top_level_frame *tlf; /* Pointer for browsing through the top_level_frame list*/
#ifdef THEME_DEBUG		
	int tlf_no = 0;
#endif	
	//if ((!win->topbar_frame) && (!win->bottombar_frame)){
	if (!win->frame_widgets){
		/*It's a new window to be rendered. Allocate 
		 * and initialize the frame_widgets array. Three element are taken extra
		 * to store the three different labels in this array
		 */
		int j;
		win->frame_widgets = (px_widget **)malloc((NR_BUTTONS + 3) * sizeof(px_widget *)); 
		if (!win->frame_widgets){
			err_exit("Not enough memory\n");
		}
		for (j = 0; j < NR_BUTTONS + 3; j++){
			win->frame_widgets[j] = NULL;
		}	
	}	
		
	tlf = self->top_level_frame;
	/* Traverse all the top level frames */
	while (tlf){
		theme_box *boxes = NULL;
		int box_number = 0;
		px_widget *parent = NULL;
		char goto_next_tlf = FALSE;
		char in_topbar = FALSE;
		char in_botbar = FALSE;
		
#ifdef THEME_DEBUG		
		printf("Top-level-frame: %d  ", tlf_no++);
		if (tlf->frame_type == TLF_TOPBAR)
			printf(" Topbar\n");
		else if (tlf->frame_type == TLF_BOTTOMBAR)
			printf(" Bottomar\n");
		else 
			printf(" Dialog\n");
		printf("Height: %ld\n", tlf->height);	
#endif		
		switch (win->wtype){
			case NORMAL_WIN:
				switch (tlf->frame_type){
					case TLF_TOPBAR:
						in_topbar = TRUE;
						if (!win->topbar_frame){
							/* The window is rendered for the first time */
							parent = new_toplevel_frame(win->frame_window, 0, 0, 
								win->width, tlf->height, NULL);	
							if (!parent){
								err_exit("Top level frame couldn't be created\n");
							}
							win->topbar_frame = parent;
						}	
						else{
							/*It's a re-render of the theme. So don't need to create the
							 * top level frame.
							 * */
							parent = win->topbar_frame;
#ifdef __arm__							
							check_for_rotation(win, parent, win->width, self, &(self->pixmap_tree), NULL);
#endif							
						}	
						break;
					case TLF_BOTTOMBAR:
						in_botbar = TRUE;
						if (!win->bottombar_frame){
							/* The window is rendered for the first time */
							parent = new_toplevel_frame(win->frame_window, 0, win->y + win->height, 
								win->width, tlf->height, NULL);	
							if (!parent){
								err_exit("Top level frame couldn't be created\n");
							}	
							win->bottombar_frame = parent;
						} else {
							/*It's a re-render of the theme. So don't need to create the
							 * top level frame. 
							 * */
							parent = win->bottombar_frame;
#ifdef __arm__							
							check_for_rotation(win, parent, win->width, self, &(self->pixmap_tree), NULL);
#endif							
						}	
						break;
					default:	
						/* The theme for dialog is not displayed when the 
					 	* window type is normal.  */
						goto_next_tlf = TRUE;
						break;
				}	
				break;
			case DIALOG_WIN:
				switch (tlf->frame_type){
					case TLF_TOPBAR:
						/* The theme for normal window(topbar & bottombar) is not displayed 
					 	* when the window type is dialog  */
						goto_next_tlf = TRUE;
						break;
					case TLF_BOTTOMBAR:
						/* The theme for normal window(topbar & bottombar) is not displayed 
					 	* when the window type is dialog  */
						goto_next_tlf = TRUE;
						break;
					case TLF_DIALOG:	
						parent = new_toplevel_frame(win->frame_window, 0, 0, 
							win->width, tlf->height, NULL);	
						if (!parent){
							err_exit("Top level frame couldn't be created\n");
						}	
						win->topbar_frame = parent;
						parent->callbacks[MOTION] = dialog_window_motion_callback;
						parent->callbacks[BUTTON_PRESSED] = dialog_window_click_callback;
						parent->callbacks[BUTTON_RELEASED] = dialog_window_release_callback;
						break;
				}	
				break;
			default:
				err_exit("Invalid window type for render_theme\n");
				break;
		}

		if (goto_next_tlf){
			/*This top-level frame is not going to be displayed.
			 * So visit the next top-level-frame in the theme structure*/
			tlf = tlf->next;
			continue;
		}	
		
		/* Traverse all the boxes and render all the objects in
		 * a box*/
		boxes = tlf->boxes;
		while (boxes){
			theme_button *buttons = NULL;
			theme_label  *labels = NULL;
			px_widget *this_box = NULL;
			unsigned char page_num_to_be_rendered = FALSE;
			int box_width;

#ifdef THEME_DEBUG		
			int label_no = 0;

			printf("    Box number: %d\n", box_number);	
			printf("    x = %ld, y = %ld, width = %ld, height = %ld, bg_type = %d\n", 
					boxes->x, boxes->y, boxes->width.value, boxes->height, boxes->bg_type);
			if (boxes->bg_type == PIXMAP){
				printf("    pixmap = %s\n", boxes->background.pixmap);
			}	
			else{ 
				printf("    fill_type = %s, color1 = %s, color2 = %s\n", 
						boxes->background.color.fill_type, boxes->background.color.color1,
											boxes->background.color.color2);
			}
#endif		
			/*calculate the box width*/
			if (boxes->width.wtype == PERCENTAGE_WIDTH)
				box_width = (win->width * boxes->width.value)/100;
			else 
				box_width = boxes->width.value;
			
			if (in_topbar){
				if (!win->topbar_frame->specific.frame.widget_list[box_number]){
					/*Create the box*/
					this_box = render_box(self->win, &(self->pixmap_tree), win, parent, boxes, box_width, self->wm);

				} else {
					/*This is a re-render of the window*/
					this_box = win->topbar_frame->specific.frame.widget_list[box_number];
#ifdef __arm__							
					check_for_rotation(win, this_box, box_width, self, &(self->pixmap_tree), boxes);
#endif					
				}	
			}
			else if (in_botbar){
				if (!win->bottombar_frame->specific.frame.widget_list[box_number]){
					/*Create the box*/
					this_box = render_box(self->win, &(self->pixmap_tree), win, parent, boxes, box_width, self->wm);

				}else {
					/*This is a re-render of the window*/
					this_box = win->bottombar_frame->specific.frame.widget_list[box_number];
#ifdef __arm__							
					check_for_rotation(win, this_box, box_width, self, &(self->pixmap_tree), boxes);
#endif					
				}	
			}else {
				/*Create the box*/
				this_box = render_box(self->win, &(self->pixmap_tree), win, parent, boxes, box_width, NULL);

			}	
			box_number++;	
			
			/* Render all the buttons inside a box*/
			buttons = boxes->button_list;	
			while (buttons){
				switch (buttons->btype){
					/*Prev_page and Next_page buttons are not directly
					 * supported in the num_flags in alchemy.h .These are influenced 
					 * by the PAGE_NAV and NEXT_PAGE_ENABLED and PREV_PAGE_ENABLED tags*/
					case PREV_PAGE:
					case NEXT_PAGE:
					case SEPARATOR:	
						render_page_nav(self->win, &(self->pixmap_tree), win, buttons, box_width, 
								this_box, &page_num_to_be_rendered);
						break;
					default:	
						/*Handle other buttons
						 * First check if button was already rendered last time. It's useful
						 * while re-rendering the theme.*/
						switch (win->oflags.all_flags & (1 << buttons->btype)){
							case 0:
								/*The button was not rendered last time*/
								/*Now check if it's to be rendered now*/
								switch (win->flags.all_flags & (1 << buttons->btype)){
									case 0:
										/*The button is not to be rendered this time also.
										 * So nothing to do!*/  
										break;
									default:
										/*Render the buttons*/
									        render_button(self->win, &(self->pixmap_tree), 
												win,buttons, box_width,this_box);	
										break;
								}	
								break;
							default:
								/*The button was rendered last time*/
								/*Now check if it's to be rendered now*/
								switch (win->flags.all_flags & (1 << buttons->btype)){
									case 0:
										/*Hide the button. It's not needed this time*/
										hide_widget(win, buttons->btype);
										break;
									default:
										/*The button is to be rendered this time. But it's
										 * alredy rendered. So nothing to do! */
										break;
								}	
								break;
						}	
						break;
				}	
				buttons = buttons->next; 
			}	
			
			labels = boxes->label_list;
			/* Render all the labels inside a box*/
			while (labels){
				int left_most_x;
				int top_most_y;
				px_widget *this_label = NULL;
				int label_len;
#ifdef THEME_DEBUG		
				printf("        Label number = %d\n", label_no++);	
				printf("        x = %ld, y = %ld, width = %ld, height = %ld, font= %s\n", 
						labels->x, labels->y, labels->width, labels->height,labels->font);
#endif		

				/* A negative x value means render the buttons from the right
				 * side of the parent box */
				if (labels->x < 0){
					left_most_x = box_width - abs(labels->x);
					/*Width of the label can't be larger than
					 * the x-position from left */
					if (labels->width > abs(labels->x))
						labels->width = abs(labels->x);
				}	
				else 
					left_most_x = labels->x;
				top_most_y = labels->y;
				
				/*Create the label now */
				switch (labels->ltype){
					case LABEL_TITLE:
						if (win->title){
							if (!is_already_created(win, LABEL_TITLE)){
								/*Create the widget when the window is rendered for the first 
							 	* time*/
								this_label = new_widget(LABEL, this_box, left_most_x, top_most_y,
									labels->width, labels->height, NULL);	
								this_label->specific.label.set_wrap_type(this_label, NO_WRAP);
								save_widget(win, LABEL_TITLE, this_label);
							}else
								this_label = get_widget(win, LABEL_TITLE) ;

						
							if (win->tfont){
								this_label->specific.label.set_font(this_label, win->tfont);
								if (this_label->specific.label.layout.font != NULL)
									label_len = font_get_width(
											this_label->specific.label.layout.font, 
											win->title, strlen(win->title));
								else
									label_len = labels->width;
							}	
							else{ 
								this_label->specific.label.set_font(this_label, labels->font);	
								if (this_label->specific.label.layout.font != NULL)
									label_len = font_get_width(
											this_label->specific.label.layout.font, 
											win->title, strlen(win->title));
								else
									label_len = labels->width;
							}	
							
							
							if (label_len <= labels->width)
								(*this_label->specific.label.set_justification)
									(this_label, win->tjustify);	
							else
								(*this_label->specific.label.set_justification)
									(this_label, JUSTIFY_LEFT);	

							(*this_label->specific.label.set_text)(this_label, win->title);	
						
							(*this_label->redraw)(this_label);
							(*this_label->show)(this_label);
							if (tlf->frame_type != TLF_DIALOG)
								this_label->callbacks[BUTTON_PRESSED] = window_title_callback; 
						}
						break;
					case LABEL_STATUS_MSG:
						if (win->status_msg){
							switch (win->flags.all_flags & (1 << STATUS_MSG)){
								case 0:
									hide_widget(win, LABEL_STATUS_MSG);
									break;
								default:
									render_status_msg(win, labels, this_box, 
											left_most_x, top_most_y, box_width);
							}
						}	
						break;
					case LABEL_PAGE_NUM:
						/*Page number is part of the page-navigation*/
						if ((win->page_num) && (page_num_to_be_rendered)){
							if (!is_already_created(win, LABEL_PAGE_NUM)){
								//Create the widget when the window is rendered for the first time
								this_label = new_widget(LABEL, this_box, left_most_x, top_most_y,
									labels->width, labels->height, NULL);	
								this_label->specific.label.set_wrap_type(this_label, NO_WRAP);
								save_widget(win, LABEL_PAGE_NUM, this_label);
							} else{
								this_label = get_widget(win, LABEL_PAGE_NUM) ;
								this_label->set_position(this_label, left_most_x, top_most_y);
							}	

							if (win->pfont)
								this_label->specific.label.set_font(this_label, win->pfont);	
							else 
								this_label->specific.label.set_font(this_label, labels->font);	

							(*this_label->specific.label.set_justification)(this_label, win->pjustify);	
							(*this_label->specific.label.set_text)(this_label, win->page_num);	
						
							(*this_label->redraw)(this_label);
							(*this_label->show)(this_label);
						}	
						break;
				}	

				labels = labels ->next;
			}	
			boxes = boxes->next;
		}	
		tlf = tlf->next;
	}	
	win->oflags.all_flags= win->flags.all_flags; //Save the flags
}



/*Render a box*/
px_widget* render_box(Window root, theme_pixmap **header, window *win, px_widget *parent, theme_box *boxes, int box_width, wm *wm)
{
	px_widget *this_box = NULL;
	
	this_box = new_widget(FRAME, parent, boxes->x, boxes->y, box_width, 
			boxes->height, NULL);	
	switch (boxes->bg_type){
		case PIXMAP:{
				px_pixmap *b_pmap = NULL;
				char *pmap_path = NULL;
#ifdef __arm__				
				if (wm){ /*Topbar or bottombar*/
					switch (wm->current_rot){
						case ROT_0:
						case ROT_180:
							pmap_path = boxes->background.pmap.pixmap_v;
							break;
						case ROT_90:
						case ROT_270:
							pmap_path = boxes->background.pmap.pixmap_h;
							break;
					}		
				}
				else {
					pmap_path = boxes->background.pmap.pixmap_v; /*For dialog window*/
				}
#else
					pmap_path = boxes->background.pmap.pixmap_v;
#endif				
				b_pmap= lookup_pixmap_tree(header, pmap_path, NULL);    
				if (b_pmap)
				{
					/*Found in the cache*/
					this_box->set_background_pixmap_from_pixmap(this_box, b_pmap);
				}
				else{
					/*Not in cache. Read from file and add to cache*/
					this_box->set_background_pixmap_from_file(this_box, pmap_path);
					add_to_pixmap_tree(root, header, pmap_path);
				}	
			}	
			break;
		case COLOR:{
				px_color *p = (px_color *)malloc(sizeof(px_color));
				if (!p)
					err_exit("Not enough memory\n");
				if (!strcmp(boxes->background.color.fill_type, "solid")){
					p->type = SOLID;
					set_color(boxes->background, p, 's');
				}	
				else if (!strcmp(boxes->background.color.fill_type, "transparent"))
					p->type = TRANSPARENT;
				else if (!strcmp(boxes->background.color.fill_type, "gradienthor")){
					p->type = GRADIENTHOR;
					set_color(boxes->background, p, 'g');
				}	
				else if (!strcmp(boxes->background.color.fill_type, "gradientvert")){
					p->type = GRADIENTVERT;
					set_color(boxes->background, p, 'g');
					break;
				}
				this_box->set_background_color(this_box, p);
				free(p);
			   }
			   break;
	}		   
	(*this_box->redraw)(this_box);
	(*this_box->show)(this_box);

	return this_box;
}

#ifdef  __arm__
/*Check if the width of the box has changed. This may happen due
 * to rotation of the simputer. Load the different pixmap for the box */
void check_for_rotation(window *win, px_widget *this_box, unsigned int width, ac_theme *theme, theme_pixmap **header, theme_box *boxes) {
	wm *wm = theme->wm;
	Window root = theme->win;
	
	if (width != this_box->width){
		/*There is a rotation of the simputer. So render everything. */
		this_box->set_size(this_box, width, this_box->height);
		if (boxes && (boxes->bg_type == PIXMAP)){
			px_pixmap *b_pmap = NULL;
			char *pmap_path = NULL;
			switch (wm->current_rot){
				case ROT_0:
				case ROT_180:
					pmap_path = boxes->background.pmap.pixmap_v;
					b_pmap= lookup_pixmap_tree(header, boxes->background.pmap.pixmap_v, NULL);    
					break;
				case ROT_90:		
				case ROT_270:
					pmap_path = boxes->background.pmap.pixmap_h;
					break;
			}
			b_pmap= lookup_pixmap_tree(header, pmap_path, NULL);    
			if (b_pmap)
			{
				/*Found in the cache*/
				this_box->set_background_pixmap_from_pixmap(this_box, b_pmap);
			}
			else{
				/*Not in cache. Read from file and add to cache*/
				this_box->set_background_pixmap_from_file(this_box, pmap_path);
				add_to_pixmap_tree(root, header, pmap_path);
			}	
		}		
		win->oflags.all_flags = 0; /*Enforcing render-all*/
		(*this_box->redraw)(this_box);
		(*this_box->show)(this_box);
	}	
	return;
}
#endif

/*Get the index of the widget for the frame_widget array
 *
 * The SEPARATOR, PREV_PAGE and NEXT_PAGE buttons are not
 * defined in the flag_num (alchemy.h). So these are mapped to the
 * logical buttons PAGE_NAV, PREV_PAGE_ENABLED and NEXT_PAGE_ENABLED
 * repectively.
 *
 * Besides, the labels are mapped to the last three indices of the
 * frame-widget array. (See, it's size is (BR_BUTTONS + 3) to accomodate
 * the labels)
 */
int get_widget_index(char wtype)
{
	int windex;
	switch(wtype){
		case LABEL_TITLE:
			windex = NR_BUTTONS;
			break;
		case LABEL_STATUS_MSG:
			windex = NR_BUTTONS + 1;
			break;
		case LABEL_PAGE_NUM:
			windex = NR_BUTTONS + 2;
			break;
		case SEPARATOR:
			windex = PAGE_NAV;
			break;
		case PREV_PAGE:
			windex = PREV_PAGE_ENABLED;
			break;
		case NEXT_PAGE:
			windex = NEXT_PAGE_ENABLED;
			break;
		default:
			windex = wtype;
			break;
	}	
	return windex;
}

/*Check if the widget type was already created for this window*/
char is_already_created(window *win, char wtype)
{
	int windex = get_widget_index(wtype);

	if (win->frame_widgets[windex])
		return TRUE;
	else 
		return FALSE;
}

/* Render one button */
void render_button(Window root, theme_pixmap **header, window *win, theme_button *buttons, int box_width, px_widget *parent_box)
{
	int left_most_x ;
	int top_most_y;
	px_widget *this_button;	
	
#ifdef THEME_DEBUG		
	int j;

	printf("        x = %ld, y = %ld, width = %ld, height = %ld\n", 
		buttons->x, buttons->y, buttons->width, buttons->height);
	for (j = 0; j < NO_OF_BUTTON_STATES; j++){
		printf("           image = %s\n", buttons->state[j].image);
		printf("           mask = %s\n", buttons->state[j].mask);
		//printf("           action = %s\n", buttons->state[j].action);
		printf("           audio = %s\n", buttons->state[j].audio);
	}	
#endif		
	/* A negative x value means render the buttons from the right
 	* side of the parent box */
	if (buttons->x < 0){
		left_most_x = box_width - abs(buttons->x);
		/*Width of the button can't be larger than
		 * the x-position from left */
		if (buttons->width > abs(buttons->x))
			buttons->width = abs(buttons->x);
	}	
	else 
		left_most_x = buttons->x;
	top_most_y = buttons->y;

	if (!is_already_created(win, buttons->btype)){
		this_button = new_widget(BUTTON, parent_box, left_most_x, top_most_y,
			buttons->width, buttons->height, NULL);	
		this_button->specific.button.set_wrap_type(this_button, NO_WRAP);
		save_widget(win, buttons->btype, this_button);
		
		/* Set the pixmap for various states */
		set_button_pixmap(root, header, buttons, NORMAL_BUTTON, this_button);
		set_button_pixmap(root, header, buttons, PRESSED_BUTTON, this_button);
		set_button_pixmap(root, header, buttons, DISABLED_BUTTON, this_button);
		
		set_button_callbacks(this_button, buttons->btype);	
		
	}else{
		this_button = get_widget(win, buttons->btype);
		this_button->set_position(this_button, left_most_x, top_most_y);
	}	
			
	set_button_state(win, buttons->btype, this_button);	
	
	(*this_button->redraw)(this_button);
	(*this_button->show)(this_button);
}

/* All buttons except PREV_PAGE and NEXT_PAGE will be in normal state.
 * The state of  PREV_PAGE and NEXT_PAGE is governed by the flags
 * PREV_PAGE_ENABLED and NEXT_PAGE_ENABLED  respectively
 */
void set_button_state(window *win, unsigned char btype, px_widget *button)
{
	switch (btype){
		case PREV_PAGE:
			set_prev_next_page_state(win, PREV_PAGE_ENABLED, button);
			break;
		case NEXT_PAGE:
			set_prev_next_page_state(win, NEXT_PAGE_ENABLED, button);
			break;
		default:
			/*by default, all buttons are in normal state*/
			button->set_state(button, NORMAL);
			break;
		
	}	
}

/* The state of the previous and next page buttons are determined by the
 * PREV_PAGE_ENABLED and  NEXT_PAGE_ENABLED flags. Depending on the flag,
 * the button may be either on normal state or  go to disabled state
 */
void set_prev_next_page_state(window *win, unsigned char btype, px_widget *button)
{
	switch (win->flags.all_flags & (1 << btype)){
			case 0:
				button->set_state(button, DISABLED);
				break;
			default:
				/*State is normal*/
				button->set_state(button, NORMAL);
				break;
	}
}

/*Set the pixmaps for various button states. First a pixmap is looked at the cache. If it's there, it's set. 
 * Otherwise the pixmap is read from file and added to the cache. */
void set_button_pixmap(Window win, theme_pixmap **header, theme_button *buttons, char state, px_widget *this_button)
{
	px_pixmap *pmap = NULL;

	switch (state){
		case NORMAL_BUTTON:
			pmap= lookup_pixmap_tree(header, buttons->state[NORMAL_BUTTON].image, NULL);    
			if (pmap){
				/*Found in the cache*/
				this_button->specific.button.set_normal_pixmap_from_pixmap(this_button, pmap); 
			}	
			else{
				/*Not in cache. Read from file and add to cache*/
				this_button->specific.button.set_normal_pixmap_from_file(this_button,
						buttons->state[NORMAL_BUTTON].image); 
				add_to_pixmap_tree(win, header, buttons->state[NORMAL_BUTTON].image);
			}	
			break;
		case PRESSED_BUTTON:
			pmap= lookup_pixmap_tree(header, buttons->state[PRESSED_BUTTON].image, NULL);    
			if (pmap){
				/*Found in the cache*/
				this_button->specific.button.set_active_pixmap_from_pixmap(this_button, pmap); 
			}	
			else{
				/*Not in cache. Read from file and add to cache*/
				this_button->specific.button.set_active_pixmap_from_file(this_button,
						buttons->state[PRESSED_BUTTON].image); 
				add_to_pixmap_tree(win, header, buttons->state[PRESSED_BUTTON].image);
			}	
			break;
		case DISABLED_BUTTON:
			pmap= lookup_pixmap_tree(header, buttons->state[DISABLED_BUTTON].image, NULL);    
			if (pmap){
				/*Found in the cache*/
				this_button->specific.button.set_disabled_pixmap_from_pixmap(this_button, pmap); 
			}	
			else{
				/*Not in cache. Read from file and add to cache*/
				this_button->specific.button.set_disabled_pixmap_from_file(this_button,
						buttons->state[DISABLED_BUTTON].image); 
				add_to_pixmap_tree(win, header, buttons->state[DISABLED_BUTTON].image);
			}	
			break;
	}	
}

/*Hide a widget, like a button or a label*/
void hide_widget(window *win, unsigned char wtype)
{
	px_widget *this_widget = NULL;
	
	this_widget = get_widget(win, wtype);
	if (this_widget)
		this_widget->hide(this_widget);
}

/* Add this widget to the frame widget list*/
void save_widget(window *win, char wtype, px_widget *label)
{
	int windex = get_widget_index(wtype);

	win->frame_widgets[windex] = label;
}

/*Get the label widget from the frame widget list*/
px_widget *get_widget(window *win, char wtype)
{
	int windex = get_widget_index(wtype);
	
	return win->frame_widgets[windex];
}	

/*Render page navigation. The page navigation is to be rendered only if the
 * PAGE_NAV bit is set in the flags of window */
void render_page_nav(Window root, theme_pixmap **header, window *win,
		theme_button *buttons, int box_width, 
		px_widget *parent_box, unsigned char *render_page_num)
{
	
	switch (win->oflags.all_flags & (1 << PAGE_NAV)){
		case 0:
			switch (win->flags.all_flags &(1 << PAGE_NAV)){
				case 0:
					/*page nav was earlier off. Now also it's ff. So no action required*/
					break;
				default:
					/*Page nav was earlier off. Now it's to be turned on*/
					*render_page_num = TRUE;
				       	render_button(root, header, win, buttons, box_width, parent_box);	
					break;
			}	
			break;
		default:
			switch (win->flags.all_flags &(1 << PAGE_NAV)){
				case 0:
					/*Page nav was earlier on. Now it's to be hidden*/
					hide_widget(win, buttons->btype);
					if (buttons->btype == SEPARATOR){
						/*Hide the label only once */
						hide_widget(win, LABEL_PAGE_NUM);
					   /*px_widget *page_num = get_widget(win, LABEL_PAGE_NUM);
					   if (page_num)
					   	page_num->hide(page_num);*/
					}	
					break;
				default:
					/*page nav was earlier on. Now also it's on. But the prev_page and next_page
					 * state may change. So render it*/
					*render_page_num = TRUE;
				       	render_button(root, header, win, buttons, box_width, parent_box);	
					break;
			}	
			break;
	}	
}

/* Set the various callbacks depending on the button type*/
void set_button_callbacks(px_widget *button, unsigned char btype)
{
	switch (btype){
		/* See the callbacks in basic_window.c */
		case CLOSE:
			button->callbacks[BUTTON_PRESSED] = window_close_callback; 
			break;
		case UP:
			button->callbacks[BUTTON_RELEASED] = window_up_callback; 
			break;
		case TICK:
			button->callbacks[BUTTON_RELEASED] = window_tick_callback; 
			break;
		case STATUS:
			button->callbacks[BUTTON_PRESSED] = window_status_callback; 
			break;
		case PRINTER:
			button->callbacks[BUTTON_PRESSED] = window_printer_callback; 
			button->callbacks[BUTTON_RELEASED] = window_printer_callback; 
			break;
		case PREV:
			button->callbacks[BUTTON_RELEASED] = window_prevapp_callback; 
			break;
		case NEXT:
			button->callbacks[BUTTON_RELEASED] = window_nextapp_callback; 
			break;
		case HELP:
			button->callbacks[BUTTON_RELEASED] = window_help_callback; 
			break;
		case PREF:
			button->callbacks[BUTTON_RELEASED] = window_pref_callback; 
			break;
		case LAUNCHER:
			button->callbacks[BUTTON_RELEASED] = window_home_callback; 
			break;
		case ANNOTATE:
			button->callbacks[BUTTON_RELEASED] = window_annotate_callback; 
			break;
		case ANNOTATION_PRESENT:
			button->callbacks[BUTTON_RELEASED] = window_annotate_callback; 
			break;
		case EMAIL:
			button->callbacks[BUTTON_PRESSED] = window_email_callback; 
			button->callbacks[BUTTON_RELEASED] = window_email_callback; 
			break;
		case INPUT:
			button->callbacks[BUTTON_RELEASED] = window_keyboard_callback; 
			break;
		case INPUT_CHOOSE:
			button->callbacks[BUTTON_PRESSED] = window_keyboard_chooser_callback; 
			break;
		case RETRACTBAR:
			button->callbacks[BUTTON_PRESSED] = window_retractbar_callback; 
			button->callbacks[BUTTON_RELEASED] = window_retractbar_callback; 
			break;
		case ONETAP:
			button->callbacks[BUTTON_PRESSED] = window_onetap_callback; 
			button->callbacks[BUTTON_RELEASED] = window_onetap_callback; 
			break;
		case NEXT_PAGE:
			button->callbacks[BUTTON_PRESSED] = window_nextpage_press_callback; 
			button->callbacks[BUTTON_RELEASED] = window_nextpage_rel_callback; 
			button->callbacks[MOUSE_ENTER] = window_nextpage_press_callback; 
			button->callbacks[MOUSE_LEAVE] = window_nextpage_rel_callback; 
			break;
		case PREV_PAGE:
			button->callbacks[BUTTON_PRESSED] = window_prevpage_press_callback; 
			button->callbacks[BUTTON_RELEASED] = window_prevpage_rel_callback; 
			button->callbacks[MOUSE_ENTER] = window_prevpage_press_callback; 
			button->callbacks[MOUSE_LEAVE] = window_prevpage_rel_callback; 
			break;
		case RIGHT:
			button->callbacks[BUTTON_PRESSED] = window_nextpage_press_callback; 
			button->callbacks[BUTTON_RELEASED] = window_nextpage_rel_callback; 
			button->callbacks[MOUSE_ENTER] = window_nextpage_press_callback; 
			button->callbacks[MOUSE_LEAVE] = window_nextpage_rel_callback; 
			break;
		case LEFT:
			button->callbacks[BUTTON_PRESSED] = window_prevpage_press_callback; 
			button->callbacks[BUTTON_RELEASED] = window_prevpage_rel_callback; 
			button->callbacks[MOUSE_ENTER] = window_prevpage_press_callback; 
			button->callbacks[MOUSE_LEAVE] = window_prevpage_rel_callback; 
			break;
		case STAR:
			button->callbacks[BUTTON_PRESSED] = window_star_callback; 
			break;
		case SPEAKER:
			button->callbacks[BUTTON_PRESSED] = window_speaker_callback; 
			button->callbacks[BUTTON_RELEASED] = window_speaker_callback; 
			break;
		case VOICE:
			button->callbacks[BUTTON_RELEASED] = window_voice_callback; 
			break;
		case XSTROKE:
			button->callbacks[BUTTON_RELEASED] = window_xstroke_callback; 
			break;
		default:
			break;
	}	
}


/* Render the status message */
void render_status_msg(window *win, theme_label *labels, px_widget *parent, int left_most_x, int top_most_y, int box_width)
{
	if (win->status_msg){
		//unsigned int lbl_width;
		
		px_widget *this_label = NULL;
		if (!is_already_created(win, LABEL_STATUS_MSG)){
			///Create the widget now
			this_label = new_widget(LABEL, parent, left_most_x, top_most_y,
				labels->width, labels->height, NULL);	
			this_label->specific.label.set_wrap_type(this_label, NO_WRAP);
			save_widget(win, LABEL_STATUS_MSG, this_label);
		}else {
			this_label = get_widget(win, LABEL_STATUS_MSG) ;
			this_label->set_position(this_label, left_most_x, top_most_y);
		}	
		
		if (win->sfont)
			this_label->specific.label.set_font(this_label, win->sfont);	
		else 
			this_label->specific.label.set_font(this_label, labels->font);	

		(*this_label->specific.label.set_justification)(this_label, win->sjustify);	
		(*this_label->specific.label.set_text)(this_label, win->status_msg);
		// Rajat. Change the size of widget to just fit in the text.

/*		if (this_label->specific.label.text != NULL)
		{
			lbl_width = font_get_width (this_label->specific.label.font, 
					this_label->specific.label.text,
					strlen(this_label->specific.label.text)) + 3;

			if (this_label->width < lbl_width || (this_label->width - lbl_width) > 10)
			{
				this_label->set_size(this_label, lbl_width, this_label->height);
				printf("---------------- Label set size to %d\n", lbl_width);
			}

			switch (win->sjustify)
			{
				case JUSTIFY_LEFT :
					this_label->set_position(this_label, left_most_x, top_most_y);
					break;

				case JUSTIFY_RIGHT: {
					int x = box_width - lbl_width - 1;

					if (x < left_most_x)
						x = left_most_x;

					this_label->set_position(this_label, x, top_most_y);
				}
					break;

				case JUSTIFY_CENTER : {
					int area_middle = (box_width - left_most_x) / 2 + left_most_x;
					int x = area_middle - lbl_width / 2;
					this_label->set_position(this_label, x, top_most_y);
				}
					break;

			}
		}
		else
		{
			this_label->set_size(this_label, 1, this_label->height);
		}*/
					
		(*this_label->redraw)(this_label);
		(*this_label->show)(this_label);
	}	
}

/* Get the height of the topbar */
unsigned long get_ac_topbar_height (ac_theme *theme)
{
	theme_top_level_frame *ptr = theme->top_level_frame;

	while (ptr){
		if (ptr->frame_type == TLF_TOPBAR)
			return ptr->height;
		ptr = ptr->next;
	}	
	return 0;
}

/* Get the height of the bottombar */
unsigned long get_ac_bottombar_height (ac_theme *theme)
{
	theme_top_level_frame *ptr = theme->top_level_frame;

	while (ptr){
		if (ptr->frame_type == TLF_BOTTOMBAR)
			return ptr->height;
		ptr = ptr->next;
	}	
	return 0;
}

/* Get the height of the dialog bar */
unsigned long get_ac_dialogbar_height (ac_theme *theme)
{
	theme_top_level_frame *ptr = theme->top_level_frame;

	while (ptr){
		if (ptr->frame_type == TLF_DIALOG)
			return ptr->height;
		ptr = ptr->next;
	}	
	return 0;
}

/* Get the background for the menu. If the value pointed to bg_type_flag is 0,
 * it returns a pointer to px_pixmap; else it returns a pointer to px_color*/
void *get_ac_menu_background(ac_theme *theme, char *bg_type_flag)
{
	px_pixmap *m_pmap = NULL;
	px_color  *m_color = NULL;
	theme_menu *menu = theme->menu;
	Window win=theme->win;
		
	switch (theme->menu->bg_type){
		case PIXMAP:
			*bg_type_flag = 0;
			m_pmap = (px_pixmap *)malloc(sizeof(px_pixmap));
			if (!m_pmap)
				err_exit("Not enough memory\n");
			image_initialize_pixmap(m_pmap);
			image_load_file(win, m_pmap, menu->background.pmap.pixmap_v);
			return m_pmap; 
		case COLOR:
			*bg_type_flag = 1;
			m_color = (px_color *)malloc(sizeof(px_color));
			if (!m_color)
				err_exit("Not enough memory\n");
			if (!strcmp(menu->background.color.fill_type, "solid")){
				m_color->type = SOLID;
				set_color(menu->background, m_color, 's');
			}	
			else if (!strcmp(menu->background.color.fill_type, "transparent"))
				m_color->type = TRANSPARENT;
			else if (!strcmp(menu->background.color.fill_type, "gradienthor")){
				m_color->type = GRADIENTHOR;
				set_color(menu->background, m_color, 'g');
			}	
			else if (!strcmp(menu->background.color.fill_type, "gradientvert")){
				m_color->type = GRADIENTVERT;
				set_color(menu->background, m_color, 'g');
			}	
			return m_color;
	}	
	return NULL;
}

/*Get the x-pixel for the menu*/
unsigned long get_ac_menu_x(ac_theme *self){
	return self->menu_x;
}		

/*Set the color types. If the color type is gradient(flag is 'g'), then only fill
 * the second color */
void set_color(theme_background bg, px_color *p, char flag)
{
	switch (flag){
		case 'g':
			widget_xcolor_from_str(&(p->second_color), bg.color.color2);
		default:
			widget_xcolor_from_str(&(p->first_color), bg.color.color1);
			break;
	}	
}

/*Does "string" starts with "substr"? */
char is_gradient_string(const char *fill_type)
{

	if ((!strcmp(fill_type, "gradienthor")) || (!strcmp(fill_type, "gradientvert")))
		return TRUE;
	err_exit("%s: Unsupported fill_type value\n", fill_type);
	return FALSE;
}

/* Add the pixmap to the BS-tree with the key, value pair*/
void add_to_pixmap_tree(Window win, theme_pixmap **header, char *path)
{
	px_pixmap *pmap = NULL;

	pmap = (px_pixmap *)malloc(sizeof(px_pixmap));
	if (!pmap)
		err_exit("Not enough memory\n");
	image_initialize_pixmap(pmap);
	image_load_file(win, pmap, path);
	lookup_pixmap_tree(header, path, pmap);
}	

/*Lookup for the key(which is the pixmap path) in the BS-tree of pixmaps. If it's 
 * available, return a pointer to that node, otherwise add the pixmap with the key and value
 *
 * If the value is NULL, it works as just a lookup; otherwise, it adds a node to the tree*/
px_pixmap *lookup_pixmap_tree(theme_pixmap **header, const char *key, px_pixmap *value)
{
	theme_pixmap *cur = NULL;
	theme_pixmap *prev = NULL;
	theme_pixmap *node = NULL;

	cur = *header;
	prev = cur;

	key += theme_path_offset;

	while (cur){
		if (!strcmp(key, cur->path))
			return cur->background;
		prev = cur;
		if (strcmp(key, cur->path) < 0)
			cur = cur->left;
		else
			cur = cur->right;
	}
	
	/* The key was not found. So add the value in the tree if the value is non-NULL*/
	if (value){
		node = make_node(key, value);
		if (!prev)
			*header = node;
		else 
			if (strcmp(key, prev->path) < 0)
				prev->left = node;
			else
				prev->right = node;
		return node->background;
	}	
	return NULL;
}	

/*Create one node of the BS-tree*/
theme_pixmap *make_node(const char *key, px_pixmap *value)
{
	theme_pixmap *node = NULL;

	node = (theme_pixmap *)malloc(sizeof(theme_pixmap));
	if (!node)
		err_exit("Not enough memory\n");
	node->path = strdup(key);
	if (!node->path)
		err_exit("Not enough memory\n");
	node->background = value;
	node->left = NULL;
	node->right = NULL;
	return node;	
}



/* De allocate the structures allocated in render_ac_theme*/
void cleanup_ac_theme(ac_theme *self, window *win){
	if(win->frame_widgets){
		free(win->frame_widgets);
		win->frame_widgets = NULL;
	}	

	if(win->topbar_frame) {
		win->topbar_frame->destroy(win->topbar_frame);
		free(win->topbar_frame);
		win->topbar_frame = NULL;
	}

	if(win->bottombar_frame) {
		win->bottombar_frame->destroy(win->bottombar_frame);
		free(win->bottombar_frame);
		win->bottombar_frame = NULL;
		
	}
	win->oflags.all_flags = 0;
}

/*Free the theme when a new theme is to be loaded*/
void free_ac_theme(ac_theme *self){
	efree(self->theme_path);
	free_top_level_frame(self->top_level_frame);
	free_theme_menu(self->menu);
	free_pixmap_tree(self->pixmap_tree);

	self->theme_path = NULL;
	self->top_level_frame = NULL;
	self->menu = NULL;
	self->pixmap_tree = NULL;


	return;

}


void free_top_level_frame(theme_top_level_frame *tlf) {
	theme_button *tmp_button;
	theme_box    *tmp_box;
	theme_label  *tmp_label;
	theme_top_level_frame  *tmp_tlf;
	
	while (tlf){
		//Deallocate the top level frame
		theme_box *boxes = tlf->boxes;
		while (boxes){
			//Free each of the boxes
			theme_button *buttons = boxes->button_list;
			theme_label *labels = boxes->label_list;
			while (buttons){
				//Free each of the buttons
				int i;
				for (i = 0; i < NO_OF_BUTTON_STATES; i++){
					efree(buttons->state[i].image);
					efree(buttons->state[i].mask);
					efree(buttons->state[i].audio);
				}	
				tmp_button = buttons;
				buttons = buttons->next;
				free(tmp_button);
			}	
			while (labels) {
				//Free each of the labels
				efree(labels->font);
				tmp_label = labels;	
				labels = labels->next;
				free(tmp_label);
			}	
			free_background(boxes);
			tmp_box = boxes;
			boxes = boxes->next;
			free(tmp_box);
		}
		tmp_tlf = tlf;	
		tlf = tlf->next;
		free(tmp_tlf);
	}
}	

void free_pixmap_tree(theme_pixmap *node) {
	if (!node)
		return;
	//Free the left and right sub-tree first
	free_pixmap_tree(node->left);
	free_pixmap_tree(node->right);

	//Free the node now
	efree(node->path);
	image_destroy(node->background);
	efree(node->background);
	efree(node);
}

//Deallocate menu information in theme
void free_theme_menu(theme_menu *menu) {
	if (menu){
		switch (menu->bg_type){
			case PIXMAP:
				efree(menu->background.pmap.pixmap_v);
				break;
			case COLOR:
				efree(menu->background.color.fill_type);
				efree(menu->background.color.color1);
				efree(menu->background.color.color2);
				break;
		}	
		free(menu);
	}	
}	
//Free with error check(not null)
void efree(void *ptr){
	if (ptr)
		free(ptr);
	return;
}		

void free_background(theme_box *boxes){
	switch (boxes->bg_type){
		case PIXMAP:
			efree(boxes->background.pmap.pixmap_v);
			efree(boxes->background.pmap.pixmap_h);
			break;
		case COLOR:
			efree(boxes->background.color.fill_type);
			efree(boxes->background.color.color1);
			efree(boxes->background.color.color2);
			break;
	}	
}
