/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#ifndef _THEMEPARSER_H
#define _THEMEPARSER_H

#include "alchemy.h"
#include "theme.h"

#define THEME_BUFFSIZE 		1024



/* Structure to be sent to the start-tag and end-tag handlers*/
typedef struct{
	ac_theme *theme;
	int depth; /* Depth of a tag in XML tree structure*/
	int theme_dir_len; /* String length of the theme directory.*/
} theme_parser_data;	


void start_tag_handler(void *user_data, const char *name, const char **attributes);
void end_tag_handler(void *user_data, const char *name);
void load_ac_theme(ac_theme *self, const char *theme_path, Window root_win);
void render_ac_theme(ac_theme *self, window *win);
void cleanup_ac_theme(ac_theme *self, window *win);
void parse_start_tag(theme_parser_data *pdata,  const char *tag_name, const char **attributes);
void parse_button(const char *button_name, theme_parser_data *pdata, const char **attr);
void parse_label(const char *label_name, theme_parser_data *pdata, const char **attr);
unsigned char get_button_type(const char *button);
void set_button_callbacks(px_widget *button, unsigned char btype);

void render_button(Window root, theme_pixmap **header, window *win, theme_button *buttons, int box_width, px_widget *parent_box) ;
void hide_widget(window *win, unsigned char wtype);
void render_page_nav(Window root, theme_pixmap **header, window *win, theme_button *buttons, int box_width, px_widget *parent_box, 
		unsigned char *render_page_num);
px_widget* render_box(Window root, theme_pixmap **header, window *win, px_widget *parent, theme_box *boxes, int box_width, wm *wm);
void save_widget(window *win, char wtype, px_widget *label);
px_widget *get_widget(window *win, char wtype);
char is_already_created(window *win, char wtype);
void set_color(theme_background bg, px_color *p, char flag);
char is_gradient_string(const char *fill_type);

void add_to_pixmap_tree(Window win, theme_pixmap **header, char *path);
px_pixmap *lookup_pixmap_tree(theme_pixmap **header, const char *key, px_pixmap *value);
void set_button_pixmap(Window win, theme_pixmap **header, theme_button *buttons, char state, px_widget *this_button);
void set_button_state(window *win, unsigned char btype, px_widget *button);
void set_prev_next_page_state(window *win, unsigned char btype, px_widget *button);
	
void handle_status_msg(window *win, theme_label *labels, px_widget *parent, int left_most_x, int top_most_y);
void render_status_msg(window *win, theme_label *labels, px_widget *parent, int left_most_x, int top_most_y, int box_width);

unsigned long get_ac_topbar_height (ac_theme *theme);
unsigned long get_ac_bottombar_height (ac_theme *theme);
unsigned long get_ac_dialogbar_height (ac_theme *theme);
theme_pixmap *make_node(const char *key, px_pixmap *value);
void check_for_rotation(window *win, px_widget *this_box, unsigned int width, ac_theme *theme, theme_pixmap **header, theme_box *boxes);

void *get_ac_menu_background(ac_theme *theme, char *bg_type_flag);
unsigned long get_ac_menu_x(ac_theme *self);

	
void free_top_level_frame(theme_top_level_frame *tlf);
void free_pixmap_tree(theme_pixmap *header) ;
void free_theme_menu(theme_menu *m) ;
void efree(void *ptr);
void free_background(theme_box *boxes);
void free_pixmap_tree(theme_pixmap *node);
void free_ac_theme(ac_theme *self);

#endif
