/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include "alchemy.h"
#include "toolbar_window.h"
#include "basic_window.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

window *new_toolbar_window(wm *wm, Window map_window)
{
	window *win;
	XSetWindowAttributes attr;
	XWindowAttributes winattr;

	// Allocate and assign common data structures
	win=new_basic_window(wm, map_window, TOOLBAR_WIN);

	// Set Function pointers
	win->redraw = &toolbar_window_redraw;
	win->refresh_widgets = &toolbar_window_refresh_widgets;
	win->set_size = &toolbar_window_set_size;
	win->set_position = &toolbar_window_set_position;
	win->show = &toolbar_window_show;
	win->hide = &toolbar_window_hide;
	win->iconify = &toolbar_window_hide;
	win->destroy = &toolbar_window_destroy;
	win->set_flags = &toolbar_window_set_flags;
	win->set_flag = &toolbar_window_set_flag;
	win->app_window =  map_window;

	// Figure out window size
	XGetWindowAttributes(wm->display, win->app_window, &winattr);
	win->x = winattr.x;
	win->y = winattr.y;
	win->width = winattr.width;
	win->height = winattr.height;

	// Reparenting happens here
	// Create toplevel
	attr.override_redirect = True;
	attr.border_pixel = attr.background_pixel = wm->black_pixel;
	attr.event_mask = SubstructureRedirectMask|SubstructureNotifyMask|
		ButtonPressMask|ButtonReleaseMask|ExposureMask;

	win->frame_window = XCreateWindow(wm->display, wm->root, win->x,
			win->y, win->width, win->height, 0,
			CopyFromParent, CopyFromParent, CopyFromParent,
			CWOverrideRedirect|CWBorderPixel|CWEventMask|CWBackPixel,
			&attr);

	// To ensure that windows stay even when we are not there
	XAddToSaveSet(wm->display, win->app_window);

	XSelectInput(wm->display, win->app_window,
		ColormapChangeMask|PropertyChangeMask);

	// Ok I am the father, you are the son
	XReparentWindow(wm->display, win->app_window, win->frame_window, 0, 0);

	// Adjust toolbar sizes
	if((win->height) > (wm->display_height - 30))
		win->height = wm->display_height - 30;

	if(win->height != winattr.height) {
		XResizeWindow(wm->display, win->app_window, win->width, win->height);
		//window_send_configure_event(win); // Seems unnecessary
	}

	window_set_state(win, WithdrawnState);

	return win;
}

// Redraw ? Whats that ?
void toolbar_window_redraw(window *win)
{
}

// Nothing to refresh either .. :)
void toolbar_window_refresh_widgets(window *win)
{
}

// Unmap and retile other mapped window
void toolbar_window_hide(window *win)
{
	wm *wm = win->wm;

	window_set_state(win, IconicState);
	win->ignore_next_unmap = True;
	if(win->appid == AC_KEYBOARD_APPID)
		wm->is_keyboard_mapped = False;
	win->is_mapped = False;
	wm->retile(wm, win, False);
	XUnmapWindow(wm->display, win->frame_window);
}

// Map and retile other mapped windows
void toolbar_window_show(window *win)
{
	int disp;
	wm *wm=win->wm;
	
	if(win->fast_load && !win->fast_visible) {
		win->fast_visible = True;
		return;
	}
	
	if(window_get_state(win) == NormalState)
		return;

	if(win->appid == AC_KEYBOARD_APPID) {
		if(wm->mapped_window && !wm->mapped_window->flags.bits.needs_keyboard)
			return;

		wm->is_keyboard_mapped = True;
	}

	disp = wm->get_toolbar_displacement(wm);

	win->set_position(win, 0, wm->display_height - (disp + win->height));
	
	win->is_mapped = True;

	XMapRaised(wm->display, win->frame_window);
	XMapSubwindows(wm->display, win->frame_window);

	window_handle_focus(win);

	wm->retile(wm, win, True);
	window_set_state(win, NormalState);
}

// Ok toolbar, we do not need u anymore
void toolbar_window_destroy(window *win)
{
	wm *wm=win->wm;

	if(win->appid == AC_KEYBOARD_APPID) {
		wm->is_keyboard_mapped = False;
		wm->keyboard_window = NULL;
	}

	if(win->is_mapped)
		wm->retile(wm, win, False);

	if(wm->mapped_window)
		wm->mapped_window->show(wm->mapped_window);


	window_destroy(win);
}

// Set flags .. no specific action
void toolbar_window_set_flags(window *win, unsigned long flags)
{
	win->flags.all_flags = flags;

	window_set_int_property(win, _AC_W_FLAGS, flags );
}

// Set flag .. no specific action
void toolbar_window_set_flag(window *win, flag_num flag_no, Bool value)
{
	if(value)
		win->flags.all_flags |= (1<<flag_no);
	else
		win->flags.all_flags &= (~(1<<flag_no));

	window_set_int_property(win, _AC_W_FLAGS, win->flags.all_flags );
}

// Set size .. just XResize, both frame and app_windows
void toolbar_window_set_size(window *win, unsigned long width, unsigned long height)
{
	win->width = width;
	win->height = height;
	XResizeWindow(win->wm->display, win->frame_window, width, height);
	XResizeWindow(win->wm->display, win->app_window, width, height);
	window_send_configure_event(win);
}

// Set size .. just Xmove the frame
void toolbar_window_set_position(window *win, int x, int y)
{
	win->x = x;
	win->y = y;
	XMoveWindow(win->wm->display, win->frame_window, x, y);
}
