/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include "alchemy.h"
#include "basic_window.h"
#include "winlist_window.h"
#include <picowidget.h>
#include <frame.h>
#include <label.h>
#include <listbox.h>
#include <font.h>
#include <widget_factory.h>
#include "utils.h"
#include "stdlib.h"

// Create a new window showing the list of all normal windows
window *new_winlist_window(wm *wm)
{
	Window frame_window;
	Display *dpy = wm->display;
	XSetWindowAttributes win_attr;
	window *w;
	int n, width, height;
	px_widget *listbox;

	win_attr.override_redirect = True;
	win_attr.background_pixel = wm->black_pixel;
	win_attr.border_pixel = wm->black_pixel;
	win_attr.event_mask = ButtonPressMask|ButtonReleaseMask|ExposureMask|
		EnterWindowMask|LeaveWindowMask|PointerMotionMask;

	// Create with small size, Resize and Reposition after getting
	// the actual size and position
	frame_window = XCreateWindow(dpy, wm->root, 0,0, 1, 1, 0,
			CopyFromParent, CopyFromParent, CopyFromParent, 
			CWOverrideRedirect|CWBackPixel|CWBorderPixel|CWEventMask,
			&win_attr);

	// Allocate structures, there is no separate app_window here
	w = new_basic_window(wm, frame_window, MENU_WIN);
	w->frame_window = frame_window;
	w->app_window = frame_window;
	
	// The topbar frame is used to hold the listbox widget
	w->topbar_frame=new_toplevel_frame(frame_window, 0, 0, 1, 1, NULL);

	// Create the listbox and figure out dimensions
	listbox = create_listbox(wm, w->topbar_frame, &width, &height, &n);
	listbox->redraw(listbox);
	w->topbar_frame->set_size(w->topbar_frame, width, height);
	
	// Move/Resize window according to calculated dimensions
	XMoveResizeWindow(wm->display, frame_window,
			(w->wm->display_width - width)/2,
			w->wm->current_theme->get_topbar_height(w->wm->current_theme),
			width, height);
	XSetWindowBorderWidth(dpy, frame_window, 1 );
	
	w->destroy = winlist_window_destroy;
	
	w->topbar_frame->redraw(w->topbar_frame);
	w->topbar_frame->show(w->topbar_frame);
	
	window_set_state(w, NormalState);

	// Show the window
	XMapWindow(dpy, frame_window);
	wm->is_menu_mapped = True;

	// Grab pointer
	XGrabPointer(dpy, wm->root, True, (ButtonPressMask|ButtonReleaseMask),
			GrabModeAsync, GrabModeAsync, None, wm->normal_cursor, CurrentTime);
	
	XSync(dpy, False);

	return w;
}

static void *menu_background;
static char menu_bg_type;
static Bool menu_initialized=False;

// Create the listbox of window names of normal windows
px_widget *create_listbox(wm *wm, px_widget *frame, int *width, int *height, int *n)
{
	int i;
	window *w;
	px_widget *list;

	list = new_widget(LISTBOX, frame, 0,0,1,1, NULL);
	*width = 0;
	*height = list->specific.listbox.inter_widget_space;
	*n = 0;

	if(!menu_initialized) {
		menu_background = wm->current_theme->get_menu_background(
				wm->current_theme, &menu_bg_type);
		menu_initialized=True;
	}

	if(menu_bg_type) {
		list->set_background_color(list, (px_color *) menu_background);
	} else {
		list->set_background_pixmap_from_pixmap(list, (px_pixmap *) menu_background);
	}

	For_All_Mappable_Do(wm, w);
		if(w->wtype==NORMAL_WIN && (!w->fast_load || w->fast_visible)) {
			int twidth, theight;
			px_widget *label;
			char tag_num[5];
			
			// Store the pointer itself .. neat hack
			sprintf(tag_num, "%x", (unsigned int) w);
			label = new_widget(LABEL, list, 0, 0, 1, 1, tag_num);
			label->specific.label.set_wrap_type(label, NO_WRAP);
			if(menu_bg_type) {
				label->set_background_color(label,(px_color *) menu_background);
			}
			list->specific.listbox.set_child(list, label, *n);
			label->specific.label.set_font(label, NINE_POINT_FONT);

			// Set name and callbacks
			label->specific.label.set_text(label, w->name);
			label->callbacks[BUTTON_PRESSED] = winlist_button_pressed;
			label->callbacks[BUTTON_RELEASED] = winlist_button_released;
			label->callbacks[MOUSE_ENTER] = winlist_mouse_enter;
			label->callbacks[MOUSE_LEAVE] = winlist_mouse_leave;
			
			// Get font metrics
			twidth = font_get_width(label->specific.label.layout.font, w->name, strlen(w->name));
			theight = font_get_height(label->specific.label.layout.font, w->name);

			if(twidth > *width)
				*width = twidth;

			label->height = theight + 4;

			(*n)++;
			(*height) += label->height + list->specific.listbox.inter_widget_space;
		}
	For_All_Mappable_Done(wm, w);

	*width += 4;
	for(i=0; i<list->specific.listbox.no_of_children; i++) {
		px_widget *l = list->specific.listbox.widget_list[i];

		l->set_size(l, *width, l->height);
		l->redraw(l);
	}
	list->set_size(list, *width, *height);

	return list;
}

// Destroy this menu
void winlist_window_destroy(window *w)
{
	Display *dpy = w->wm->display;

	XUnmapWindow(dpy, w->frame_window);
	XUngrabPointer(dpy, CurrentTime);
	window_set_state(w, WithdrawnState);

	w->wm->is_menu_mapped = False;

	XGrabServer(dpy);
	window_destroy(w);
	XUngrabServer(dpy);
}

// Click happens here: do not know what to do :)
Bool winlist_button_pressed(px_widget *pw, px_event *pxe)
{
	px_color pxc;

	pxc.type = SOLID;
	widget_xcolor_from_str(&pxc.first_color, "black");
	pw->set_background_color(pw, &pxc);
	pw->specific.label.set_normal_color(pw, "white");
	pw->redraw(pw);
	pw->show(pw);

	return True;
}

// Pointer released; must switch to the specified window
Bool winlist_button_released(px_widget *pw, px_event *pxe)
{
	window *w;

	w = (window *)pxe->private_data;

	winlist_switch_to(w->wm, pw->tag);
	w->flags.bits.needs_keyboard = True; // Set flag to be destroyed

	return True;
}

// Pointer enters the label
Bool winlist_mouse_enter(px_widget *pw, px_event *pxe)
{
	window *w;
	px_color pxc;

	w = (window *)pxe->private_data;
	pxc.type = SOLID;
	widget_xcolor_from_str(&pxc.first_color, "black");
	pw->set_background_color(pw, &pxc);
	pw->specific.label.set_normal_color(pw, "white");
	pw->redraw(pw);
	pw->show(pw);

	return True;
}

// Pointer leaves the label
Bool winlist_mouse_leave(px_widget *pw, px_event *pxe)
{
	window *w;
	px_color pxc;

	w = (window *)pxe->private_data;

	//widget_xcolor_from_str(&pxc.first_color, "white");
	if(menu_bg_type) {
		pw->set_background_color(pw, (px_color *) menu_background);
	} else {
		pxc.type = TRANSPARENT;
		pw->set_background_color(pw, &pxc);
	}
	pw->specific.label.set_normal_color(pw, "black");
	pw->redraw(pw);
	pw->show(pw);

	return True;
}

// Switch to the specified window
void winlist_switch_to(wm *wm, const char *tag)
{
	window *switch_win;

	// Get the window pointer from the tag
	switch_win = (window *)strtoul(tag, NULL, 16);

	// Switch to the window .. cool!
	if(switch_win) {
		window *old_mapped_window;

		old_mapped_window = wm->mapped_window;
		switch_win->show(switch_win);
		if(old_mapped_window && old_mapped_window != wm->mapped_window)
			old_mapped_window->hide(old_mapped_window);
	}
}
